type frag_t = (int * Nb.t) list

let cmp (i1,_) (i2,_) =
	match Pervasives.compare i1 i2 with
	| 0 -> invalid_arg "Cert.cmp"
	| n -> n

let sort a = List.sort cmp a

let prF: frag_t -> string
= function
	| [] -> "empty"
	| _::_ as l ->
		let pr (i, n) = Printf.sprintf "%s*(%i)" (Nb.pr n) i in
		String.concat " + " (List.map pr l)

let isEqF (f1: frag_t) (f2: frag_t): bool =
	List.length f1 = List.length f2 &&
	List.for_all2 (fun (i1, n1) (i2, n2) -> i1 = i2 && Nb.cmp n1 n2 = 0)
		(sort f1) (sort f2)

let mapF (fn: Nb.t -> Nb.t) (f : frag_t) =
	List.map (fun (i, n) -> (i, fn n)) f

type cstr_t =
| Direct of frag_t
| SplitEq of frag_t * frag_t

let prC =
	function
	| Direct f -> "Direct:\n\t" ^ (prF f)
	| SplitEq (f1, f2) -> "SplitEq:\n\t" ^ (prF f1) ^ "\n\t" ^ (prF f2)

let isEqC: cstr_t -> cstr_t -> bool
= fun ce ce' ->
	match ce, ce' with
	| Direct f, Direct f' -> isEqF f f'
	| SplitEq (f1, f2), SplitEq (f1', f2') -> isEqF f1 f1' && isEqF f2 f2'
	| Direct _, SplitEq (_, _)
	| SplitEq (_, _), Direct _ -> false

let cstrFromTriv: int -> frag_t -> cstr_t
= fun i f ->
	let (f1, f2) = List.partition (fun (i', _) -> i = i') f in
	match f1 with
	| [_, n] -> Direct (mapF (fun n' -> Nb.div n' (Nb.neg n)) f2)
	| _ -> Pervasives.failwith "Cert.cstrFromTriv"

let cstrInv: int -> frag_t -> cstr_t
= fun i f ->
	let (f1, f2) = List.partition (fun (i', _) -> i = i') f in
	match f1 with
	| [_, n] ->
		let invn = Nb.inv n in
		let op = fun n' -> Nb.mult (Nb.neg invn) n' in
		Direct ((i, invn)::(mapF op f2))
	| _ -> Pervasives.failwith "Cert.cstrInv"

type partialT
= (int * cstr_t) list

let partialPr: partialT -> string
= fun l ->
	let l' = List.map (fun (i, ce) -> Printf.sprintf "%i: %s" i (prC ce)) l in
	String.concat "\n" l'

let partialEq: partialT -> partialT -> bool
=	let cmp = fun (i, _) (i', _) -> Pervasives.compare i i' in
	let eq1 = fun (i, ce) (i', ce') -> i = i' && isEqC ce ce' in
	fun l1 l2 ->
		let l1' = List.sort cmp l1 in
		let l2' = List.sort cmp l2 in
		if List.length l1' = List.length l2'
		then List.for_all2  eq1 l1' l2'
		else false

type equivT
= partialT * partialT

type joinEltT = {
	id: int;
	arg1: cstr_t;
	arg2: cstr_t
}

let getId: joinEltT -> int
= fun j -> j.id

let getArg1: joinEltT -> cstr_t
= fun j -> j.arg1

let getArg2: joinEltT -> cstr_t
= fun j -> j.arg2

type joinT = joinEltT list

type t =
| Empty of frag_t
| Implies of (int * cstr_t) list
| Bind of (int * Cstr.t * cstr_t * t)

let rec isEq: t -> t -> bool
= fun ce ce' ->
	match ce, ce' with
	| Empty f, Empty f' -> isEqF f f'
	| Implies l, Implies l' -> partialEq l l'
	| Bind (i, _, ce1, ce2), Bind (i', _, ce1', ce2') ->
		i = i' && isEqC ce1 ce1' && isEq ce2 ce2'
	| _, _ -> false

type asgT
= AsgInv of (int * t) list * t | AsgNInv of t * (int * t) list

type cert_t = t

let mkF (init: (int * Nb.t) list) =
	if List.for_all (fun (_, a) -> Nb.cmpz a <> 0) init then
		try
			sort init
		with
		| Invalid_argument "Cert.cmp" -> invalid_arg "Cert.mkF"
	else
		invalid_arg "Cert.mkF"

let renameF: int -> int -> frag_t -> frag_t
= fun idOld idNew f ->
	let rename (i, n) =
		if idOld = i then
			(idNew, n)
		else
			(i, n)
	in
	List.map rename f

let mergeF (l0: (frag_t * Nb.t) list): frag_t =
	let l = List.map
		(fun (f, n) ->
			if Nb.cmpz n = 0 then []
			else
				List.map (fun (i, a) -> (i, Nb.mult n a)) f) l0
	in
	let rec merge2 f1 f2 =
		match f1, f2 with
		| [], t | t, [] -> t
		| (i1, n1)::t1, (i2, n2)::t2 ->
			if i1 < i2 then
				(i1, n1)::(merge2 t1 f2)
			else if i1 > i2 then
				(i2, n2)::(merge2 f1 t2)
			else
				let n = Nb.add n1 n2 in
                                let t = merge2 t1 t2 in
				if Nb.cmpz n = 0 then t else (i1, n)::t
	in
	List.fold_left merge2 [] (List.map sort l)

let rec pr =
	function
	| Empty f -> "Empty: " ^ (prF f)
	| Bind (id, _, ce1, ce2) -> Printf.sprintf "Bind %i: %s\nin\n%s" id (prC ce1) (pr ce2)
	| Implies ce0 ->
		match ce0 with
		| [] -> "Implies: empty"
		| h::t ->
			let p (i, c) = "(" ^ (string_of_int i) ^ "): " ^ (prC c) in
			List.fold_left (fun s c -> s ^ "\n" ^ (p c)) (p h) t
