(** Certificate types and operations on them. *)

(** All the certificates aim at proving inclusion properties,
which can be proved by exhibiting linear combinations of constraints.
[frag_t] specifies such a linear combination of constraints as a list of pairs.
Each pair gives the coefficient to be applied to a constraint identified by a number
(see {! Cons.t}). *)
type frag_t = (int * Nb.t) list

(** Pretty-print a value of type [frag_t]. *)
val prF: frag_t -> string

(** [isEqF f1 f2] returns [true] if [f1] and [f2] specify the same
linear combination.  The order of the coefficients is irrelevant. *)
val isEqF: frag_t -> frag_t -> bool

(** Build a [frag_t] for a coefficient list.
[mkF] checks that none of the coefficients is nil
and that each constraint does appear more than once.
[Invalid_argument "Cert.mkF"] is raised if one of these conditions is not met.
Well-formedness is relied on by partial certificate manipulations in {! Pol}. *)
val mkF: (int * Nb.t) list -> frag_t

(** Apply a given function to all the coefficients of a [frag_t]. *)
val mapF: (Nb.t -> Nb.t) -> frag_t -> frag_t

(** [rename idOld idNew f] replaces the coefficient of the contraint
identified by [idOld] by equal coefficients of the contraint identified
by [idNew] in [f]. *)
val renameF: int -> int -> frag_t -> frag_t

(** Make linear combinations of fragments. *)
val mergeF: (frag_t * Nb.t) list -> frag_t

(** This type expresses the arguments of a proof of inclusion of a set of
constraints in one constraint.  A proof can be [Direct], meaning that the
provided linear combinations yields the desired contraint.  When proving
inclusion in an equality, it might be useful to split it into two inequalities
(see {! Cstr.split}) and provide a proof for each of them ([SplitEq f1 f2]).
The implementation assumes that the equality being proved is the constraint
resulting from the linear combination described by [f1], with the comparison
sign changed to [Cstr.Eq]. *)
type cstr_t
=	| Direct of frag_t
	| SplitEq of frag_t * frag_t

(** Pretty-print a certificate for one constraint. *)
val prC: cstr_t -> string

(** [isEqC ce ce'] returns [true] if [ce] and [ce'] are the same proof argument.
Linear combination equality is check with [isEqF]. *)
val isEqC: cstr_t -> cstr_t -> bool

(** [cstrFromTriv i f] builds an inclusion certificate for constraint [i]
from a linear combination yielding a trivial constraint [f]. *)
val cstrFromTriv: int -> frag_t -> cstr_t

(** [cstrInv i f] takes a linear combination describing the rewriting
of a constraint [c], identified by [i] and yielding a constraint [c'];
it then builds the rewriting of [c'] (identified by [i]) yielding [c]. *)
val cstrInv: int -> frag_t -> cstr_t

(** A partial certificate as a list of proofs for some of the output constraints.
This type lacks integration in the module;
this is mostly of notational shortcut for now. *)
type partialT
= (int * cstr_t) list

(** Pretty-printer for values of type [partialT]. *)
val partialPr: partialT -> string

(** Equality test for values of type [partialT].
Equality on values of types [cstr_t] is tested with [isEqC].
Order is irrelevant. *)
val partialEq: partialT -> partialT -> bool

(** A on-going proof of equivalence between two polyhedra.
This type lacks integration in the module;
this is mostly of notational shortcut for now. *)
type equivT
= partialT * partialT

(** This type expresses a strategy for proving the inclusion of [p1] in [p2]. 
The two possibilities are showing that [p1] is [Empty] (i.e. bottom) or that
there exists, for each constraint [c] of [p2], a linear combination of the
constraints [p1] yielding [c]. The latter corresponds to an [Implies] certificate.
Additionally, the possibility is left to defined auxiliary constraints to
be used in a proof with [Bind]. When using it, care should be taken to use a
fresh identifier (i.e. no already used in [p1]).

{e Remarks.} The [Cstr.t] is not really useful and should be remove.
[Bind] is not a really good construction, judging by the amount of code it
takes to use it in {! Pol}. *)
type t
=	| Empty of frag_t
	| Implies of (int * cstr_t) list
	| Bind of (int * Cstr.t * cstr_t * t)

(** Pretty-print an inclusion certificate. *)
val pr: t -> string

(** Test equality of two certificates.
[frag_t] equality is checked using [isEqF].
Equality for [(int * cstr_t) list] is checked using [partialEq].
The order of the [Bind] constructs matters. *)
val isEq: t -> t -> bool

(** This type captures the arguments of correctness of one contraint [c] of
the result [p] of the convex hull of two polyhedra [p1] and [p2].
The correctness condition on [p] states that it should include in
both [p1] and [p2].  An element of type [joinEltT] provides a way to
prove that [p1] is included in [c] as its [arg1] field, and a way to
prove that [p2] is included in [c] as its [arg2] field.
Last, the constraint [c] of [p] is identified by [id]. *)
type joinEltT
= {id: int; arg1: cstr_t; arg2: cstr_t}

(** Get the contents of field [id] of a [joinEltT]. *)
val getId: joinEltT -> int

(** Get the contents of field [arg1] of a [joinEltT]. *)
val getArg1: joinEltT -> cstr_t

(** Get the contents of field [arg2] of a [joinEltT]. *)
val getArg2: joinEltT -> cstr_t

(** This type captures the arguments of correctness of the result [p]
of the convex hull of two polyhedra [p1] and [p2]. It is the collection
of correctness arguments for all of its constituting constraints. *)
type joinT = joinEltT list

(** The proof argument which justifies an assignment.
[AsgInv] contains the necessary information for the case of
an invertible operation (x := f(x)), while [AsgNInv] containts
inversion for a non-invertible operation (e.g. x := 3). *)
type asgT
=	| AsgInv of (int * t) list * t
	| AsgNInv of t * (int * t) list
