(** Linear constraints and operations on them. *)

type cmpT
= Eq | Le | Lt

(** The main type for linear contraints.
A constraint [c] of type [t] represents [c.v c.typ c.c]. *)
type t = {
	typ : cmpT; (** the comparison operator *)
	v: Vec.t; (** the linear term *)
	c: Nb.t (** the constant *) }

val get_typ : t -> cmpT 
val get_v : t -> Vec.t
val get_c : t -> Nb.t

(** The type of properties the can be told by examining a constraint on its own. *)
type prop_t =
| Trivial (** 0 = 0 or 0 < 1 are Trivial *)
| Contrad (** trivialy contradictory: e.g. 0 = 1 or 0 <  -1 *)
| Nothing (** neither trivial nor trivialy contradictory *)

(** See [mult]. *)
exception BadMult

(** See [elim]. *)
exception CannotElim

(** See [elim]. *)
exception NoElim

val cmpAdd: cmpT -> cmpT -> cmpT

(** [eval c x] checks that the given point [x] satisfies the constraint [c]. *)
val eval: t -> Vec.t -> bool

(** Add two constraints. *)
val add: t -> t -> t

(** Multiply a constraint by a constant.
If a negative constant is provided and the constraint is an inequality, then [BadMult] is raised. *)
val mult: Nb.t -> t -> t

(** [elim c1 c2 x] linearly combines [t1] and [t2] so as to eliminate [x] from the result.
The coefficients applied to [c1] and [c2] are returned, in order.
If [x] does not appear in any of the two, then exception [NoElim] is raised.
If elimination is impossible, due to [x] not appearing in one of the two constraints
or appearing with coefficients of the same sign in two inequalities, [CannotElim] is raised. *)
val elim: t -> t -> Var.t -> t * Nb.t * Nb.t

(** Compute the complement constraint to the given inequality.
If the supplied linear constraint is an equality, [Invalid_argument "Cstr.compl"] is raised. *)
val compl: t -> t

(** Split the given equality (e.g. x = 1) into the two equivalent inequalities (e.g. x <= 1 and -x <= -1).
If the supplied constraint is not an equality, [Invalid_argument "Cstr.split"] is raised. *)
val split: t -> t * t

(** [mk typ coefs cst] builds the linear constraint [coefs typ cst],
where each member of [coefs] gives the coefficient of a given variable.
Zero coefficients can be left implicit. *)
val mk: cmpT -> (Nb.t * Var.t) list -> Nb.t -> t

(** [eq lin cst] builds the equality [lin = cst],
where each member of [lin] is the coefficient of a variable.
Zero coefficients can be left implicit. *)
val eq: (Nb.t * Var.t) list -> Nb.t -> t

(** [le lin cst] builds the non-strict inequality [lin <= cst],
where each member of [lin] is the coefficient of a variable.
Zero coefficients can be left implicit. *)
val le: (Nb.t * Var.t) list -> Nb.t -> t

(** [lt lin cst] builds the strict inequality [lin < cst],
where each member of [lin] is the coefficient of a variable.
Zero coefficients can be left implicit. *)
val lt: (Nb.t * Var.t) list -> Nb.t -> t

(** [mkInt] is equivalent to [mk] but transforms [coefs] < [cst] into [coefs] <= ([cst] - 1). *)
val mkInt: cmpT -> (Nb.t * Var.t) list -> Nb.t -> t

(** Check the syntactical equality of two constraints.
x = 1 and x = 1 are syntactically equal, but not 2 * x = 2 and x = 1. *)
val isEqSyn: t -> t -> bool

(** Check the syntactical inclusion of one constraint in another.
The linear term of both constraints must be syntactically equal.
x = 0 is included in x < 1, but not is 2 * x < 2. *)
val isInclSyn: t -> t -> bool

(** [isEq c1 c2] checks whether the subspace described by [c1] is equal to that of [c2]. *)
val isEq: t -> t -> bool

(** [isIncl c1 c2] checks whether the subspace described by [c1] is included in that of [c2]. *)
val isIncl: t -> t -> bool

(** Tell what can be infered on a constraint on its own.  See [prop_t]. *)
val tellProp: t -> prop_t

(** Build the list of variables which appear with non-zero coefficients in the constraints.
There can be duplicates. *)
val listVars: t list -> Var.t list

(** [getCoefsFor mx l] gathers the "coefficient" of [mx] in the constraints
in [l]. If [mx] is [None] then the constants are gathered. If [mx] is
[Some x], the coefficients of variable [x] are gathered. The order of the
numbers in the returned list matches that of the constraints in [l]. *)
val getCoefsFor : Var.t option -> t list -> Nb.t list

(** [pr b c] pretty-prints contraint [c] using the variables and their names from [b]. *)
val pr: (Var.t -> string) -> t -> string
