type t = (Var.t * Cons.t) list

let pr: (Var.t -> string) -> t -> string
= fun varPr e ->
	List.fold_right (fun (_, c) s -> s ^ (Cons.pr varPr c) ^ "\n") e ""

let prExt: (Var.t -> string) -> t -> string
= fun varPr e ->
	let pr1 x c = Printf.sprintf "%s (def: %s)\n" (Cons.prExt varPr c) (varPr x) in
	List.fold_right (fun (x, c) s -> s ^ (pr1 x c)) e ""

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

let nil : t = []

let initc (s : t) = List.map (fun (x, c) -> (x, Cons.initc c)) s
let extractc (s : t) = List.map (fun (_, c) -> Cons.extractc c) s

let filter s c =
	let filter1 (x, c2) c1 =
		if Nb.cmpz (Vec.get (Cstr.get_v (Cons.get_c c1)) x) = 0 then
			c1
		else
			Cons.elimc (Cons.get_id c1) x c1 c2
 	in
	List.fold_right filter1 s c

let implies s c =
	match Cstr.tellProp (Cons.get_c (filter s c)) with
	| Cstr.Trivial -> true
	| Cstr.Contrad (* should tell something useful *)
	| Cstr.Nothing -> false

let isIncl (s1 : t) (s2 : t) =
	if List.length s1 < List.length s2 then
		NoIncl
	else

	let s = initc s1 in
	let rec _isIncl cert =
	function
	| [] -> Incl cert
	| (_, c)::t ->
		(* XXX: Could this be done without the -1 dance? *)
		let c1 = filter s (Cons.mk (-1) (Cons.get_c c)) in
		match Cstr.tellProp (Cons.get_c c1) with
		| Cstr.Contrad | Cstr.Nothing -> NoIncl
		| Cstr.Trivial ->
			let (d, w) = List.partition (fun (id, _) -> id = -1) (Cons.get_f c1) in
			match d with
			| [_, a] ->
				let norm (i, a1) = (i, Nb.div a1 (Nb.neg a)) in
				let f = Cert.mkF (List.map norm w) in
				_isIncl (((Cons.get_id c), Cert.Direct f)::cert) t

			| _ -> failwith "EqSet.isIncl"
	in
	_isIncl [] s2

let isEq (s1 : t) (s2 : t) =
	if List.length s1 = List.length s2 then
		List.for_all2 (fun (x1 ,c1) (x2, c2) ->
			x1 = x2 && Cstr.isInclSyn (Cons.get_c c1) (Cons.get_c c2)) s1 s2
	else
		false

let choose c =
	let rec find pred =
		function
		| Rtree.Nil -> None
		| Rtree.Sub (l, n, r) ->
			if pred n then
				Some (Var.XH, n)
			else
				match find pred l with
				| Some (x, n1) -> Some (Var.XO x, n1)
				| None ->
					match find pred r with
					| Some (x, n1) -> Some (Var.XI x, n1)
					| None -> None
	in
	match find (fun n -> Nb.cmpz n <> 0) (Cstr.get_v (Cons.get_c c)) with
	| None -> failwith "EqSet.choose"
	| Some (x, a) -> (x, a)

type meetT
=	| Added of t * Cert.equivT
	| Bot of Cert.frag_t

let meetPr: (Var.t -> string) -> meetT -> string
= fun varPr -> function
	| Added (e, (ce1, ce2)) ->
		Printf.sprintf "Added (s, (ce, ce')) with\ns:\n%s\nce:\n%s\nce':\n%s\n"
			(prExt varPr e) (Cert.partialPr ce1) (Cert.partialPr ce2)
	| Bot f -> Printf.sprintf "Bot: %s" (Cert.prF f)

let meetEq: meetT -> meetT -> bool
= fun ar ar' ->
	match ar, ar' with
	| Added (e, (ce1, ce2)), Added (e', (ce1', ce2')) ->
		isEq e e' && Cert.partialEq ce1 ce1' && Cert.partialEq ce2 ce2'
	| Bot f, Bot f' -> Cert.isEqF f f'
	| Added (_, _), Bot _
	| Bot _, Added (_, _) -> false

type addInternalT
=	| AddCont of t * Cert.partialT
	| AddBot of Cert.frag_t

let addM: t -> (int * Cstr.t) list -> meetT
=	let switch: int -> Cert.frag_t -> Nb.t * Cert.frag_t
	= fun i f ->
		let (f1, f2) = List.partition (fun (i', _) -> i = i') f in
		match f1 with
		| [_, n] ->
			let invN = Nb.inv n in
			(invN, Cert.mapF (fun n' -> Nb.mult (Nb.neg invN) n') f2)
		| _ -> failwith "EqSet.addM"
	in
	let add: addInternalT -> int * Cstr.t -> addInternalT
	= function
		| AddBot _ as r -> fun _ -> r
		| AddCont (s, red) ->
			fun (i, c) ->
				match Cstr.get_typ c with
				| Cstr.Le | Cstr.Lt -> failwith "EqSet.addM"
				| Cstr.Eq ->
					let c1 = filter s (Cons.mk i c) in
					match Cstr.tellProp (Cons.get_c c1) with
					| Cstr.Trivial ->
						let frag = Pervasives.snd (switch i (Cons.get_f c1)) in
						AddCont (s, (i, Cert.Direct frag)::red)
					| Cstr.Contrad -> AddBot (Cons.get_f c1)
					| Cstr.Nothing ->
						let (x, ax) = choose c1 in
						let c2 = Cstr.mult (Nb.inv ax) (Cons.get_c c1) in
						let f2: Cert.frag_t
						=	let f = fun n -> Nb.mult (Nb.inv ax) n in
							Cert.mapF f (Cons.get_f c1)
						in
						let newE = {Cons.id = i; Cons.c = c2; Cons.f = f2} in
						AddCont ((x, newE)::s, red)
	in
	fun s0 l ->
		let s0' = initc s0 in
		match List.fold_left add (AddCont (s0', [])) l with
		| AddBot f -> Bot f
		| AddCont (s, red) ->
			let fl = List.map (fun (_, c) -> (Cons.get_id c, Cons.get_f c)) s in
			let fwdCe = List.map (fun (i, f) -> (i, Cert.Direct f)) fl in
			let bwdCe =
				let bwd = fun (i, f) ->
					let (n, f') = switch i f in (i, Cert.Direct ((i, n)::f'))
				in
				List.map bwd fl in
			Added (s, (fwdCe, red @ bwdCe))

let add: t -> int -> Cstr.t -> meetT
= fun s i c -> addM s [i, c]

let meet: t -> t -> meetT
= fun s s' -> addM s (List.map (fun (_, c) -> (Cons.get_id c, Cons.get_c c)) s')

let rename s fromX toY =
	let rename1 (x, c) =
		((if x = fromX then toY else x), Cons.rename fromX toY c)
	in
	List.map rename1 s

let pick (msk : Var.t option Rtree.t) (e : Cons.t) =
	let rec find m v =
		match m, v with
		| Rtree.Nil, _ | _, Rtree.Nil -> None
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			if n1 = None || Nb.cmpz n2 = 0 then
				match find l1 l2 with
				| Some _ as v -> v
				| None -> find r1 r2
			else
				n1
	in
	find msk (Cstr.get_v (Cons.get_c e))

let rec subst x i e =
	function
	| [] -> (i, [])
	| (x1, e1)::l1 ->
		if x1 = x then
			failwith "EqSet.subst"
		else
			let e2 =
				if Nb.cmpz (Vec.get (Cstr.get_v (Cons.get_c e1)) x) = 0 then
					{ e1 with Cons.id = i }
				else
					Cons.elimc i x e e1
			in
			let (i1, l2) = subst x (i + 1) e l1 in
			(i1, (x1, e2)::l2)

let rec tryDefs (i : int) (msk : Var.t option Rtree.t) =
	function
	| [] -> (None, i, [])
	| (x, e)::l ->
		if Rtree.get None msk x = None then
			let (def, i1, l1) = tryDefs (i + 1) msk l in
			(def, i1, (x, { e with Cons.id = i })::l1)
		else
			let (i1, l1) = subst x i e l in
			(Some (e, x), i1, l1)

let trySubstM i msk l =
	let (def, i1, l1) = tryDefs i msk l in
	if def = None then
		let rec _trysubstm i msk =
			function
			| [] -> (None, i, [])
			| (x, e)::l ->

			match pick msk e with
			| None ->
				let e1 = { e with Cons.id = i } in
				let (def, i1, l1) = _trysubstm (i + 1) msk l in
				(def, i1, (x, e1)::l1)

			| Some x ->
				let (i1, l1) = subst x i e l in
				(Some (e, x), i1, l1)
		in
		_trysubstm i msk l
	else
		(def, i1, l1)

let trySubst x i l =
	let msk = Rtree.set None Rtree.Nil x (Some x) in
	let (optx, i1, s1) = trySubstM i msk l in
	match optx with
	| None -> (None, i1, s1)
	| Some (e, _) -> (Some e, i1, s1)

let joinSetup: bool -> Var.t -> Var.t option Rtree.t -> Var.t -> int -> t -> Var.t * Var.t option Rtree.t * t
= fun dup nxt relocTbl alpha idOffset s ->
	let apply (x, c) (nxt1, relocTbl1, s1) =
		let (nxt2, relocTbl2, c1) = Cons.joinSetup dup nxt1 relocTbl1 alpha idOffset c in
		let x1 =
			if dup then
				x
			else
				match Rtree.get None relocTbl2 x with
				| None -> failwith "EqSet.joinSetup2"
				| Some x1 -> x1
		in
		(nxt2, relocTbl2, (x1, c1)::s1)
	in
(* List.fold_right is necessary because order needs to be preserved (echelon form) *)
	List.fold_right apply s (nxt, relocTbl, nil)

let list (s : t) = List.map (fun (_, c) -> c) s
let isTop (s : t) = (s = [])
