(** {! Pol} treats equalities and inequalities differently.
This module is the data structure that hosts the equalities of a polyhedron. *)

type t = (Var.t * Cons.t) list

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

val nil: t

val filter: t -> Cons.t -> Cons.t

val choose: Cons.t -> Var.t * Nb.t

(** The result of an intersection.
Intersection is just concatenation and then minimization.
Either the result is an unsatisfiable equality set,
which can be shown by a linear combination of equalities
yielding a trivially contradictory equality.
This is the semantics of [Bot f].
Or the result is an equality set which is equivalent to
the union of a pre-existing set and new equalities.
This is the semantics of [Added (s, cert)]. *)
type meetT
=	| Added of t * Cert.equivT
	| Bot of Cert.frag_t

(** Pretty-print a value of type [meetT]. *)
val meetPr: (Var.t -> string) -> meetT -> string

(** Test whether two values of type [meetT] are equal.
Equality on [Cert.frag_t] is tested using [Cert.isEqF],
equality on [Cert.equivT] is tested using [Cert.partialEq] and
equality on t is tested using [isEq] *)
val meetEq: meetT -> meetT -> bool

(** [addM s l] computes the intersection of [s] with the constraints [l].
The integer provided with each constraint will identify it uniquely in the
set. It is considered different from all identifiers in [s], although this
is not checked. The linear combination which are returned as part of the
result value are expressed with respected to the input constraints. *)
val addM: t -> (int * Cstr.t) list -> meetT

(** [add s i c] is a short-hand notation for [addM s [i, c]]. *)
val add: t -> int -> Cstr.t -> meetT

(** [meet s s'] computes the intersection of [s] with [s'].
Constraint identifiers in both [s] and [s'] must be unique in the union
of the two sets. [meet] is implemented in terms of [addM]. *)
val meet: t -> t -> meetT

val implies: t -> Cons.t -> bool
val isIncl: t -> t -> rel_t
val isEq: t -> t -> bool

val rename: t -> Var.t -> Var.t -> t

val pick: Var.t option Rtree.t -> Cons.t -> Var.t option
val subst: Var.t -> int -> Cons.t -> t -> int * t

val tryDefs: int -> Var.t option Rtree.t -> t ->
	(Cons.t * Var.t) option * int * t

val trySubstM: int -> Var.t option Rtree.t -> t ->
	(Cons.t * Var.t) option * int * t

val trySubst: Var.t -> int -> t -> Cons.t option * int * t

val joinSetup: bool -> Var.t -> Var.t option Rtree.t -> Var.t -> int -> t -> Var.t * Var.t option Rtree.t * t

val extractc: t -> (int * Cert.cstr_t) list
val initc: t -> t

val list: t -> Cons.t list
val isTop: t -> bool

val pr: (Var.t -> string) -> t -> string
val prExt: (Var.t -> string) -> t -> string
