type t = Cons.t list

type prop_t =
| Empty of Cert.frag_t
| Trivial
| Implied of int * Cert.cstr_t
| Check of Cons.t

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

type simpl_t =
| SimplMin of t
| SimplBot of Cert.frag_t

module Stats
  = struct

  let fmeGenCnt : Nb.Z.t Pervasives.ref
    = Pervasives.ref Nb.Z.z

  let fmeGen : int -> unit
    = fun i -> fmeGenCnt := Nb.Z.add !fmeGenCnt (Nb.Z.mk i);;

  let fmeRedKohlerCnt : Nb.Z.t Pervasives.ref
    = Pervasives.ref Nb.Z.z

  let fmeRedKohler : int -> unit
    = fun i -> fmeRedKohlerCnt := Nb.Z.add !fmeRedKohlerCnt (Nb.Z.mk i);;

  let fmeRedSynCnt : Nb.Z.t Pervasives.ref
    = Pervasives.ref Nb.Z.z

  let fmeRedSyn : int -> unit
    = fun i -> fmeRedSynCnt := Nb.Z.add !fmeRedSynCnt (Nb.Z.mk i);;

  let fmeRedSxCnt : Nb.Z.t Pervasives.ref
    = Pervasives.ref Nb.Z.z

  let fmeRedSx : int -> unit
    = fun i -> fmeRedSxCnt := Nb.Z.add !fmeRedSxCnt (Nb.Z.mk i);;

  let reset : unit -> unit
    = fun () ->
    begin
      fmeGenCnt := Nb.Z.z;
      fmeRedKohlerCnt := Nb.Z.z;
      fmeRedSynCnt := Nb.Z.z;
      fmeRedSxCnt := Nb.Z.z;
    end

  let pr : unit -> string
    = fun () ->
    Printf.sprintf "Fourier-Motzkin elimination generated %s constraints\n%s constraints were redundant by Kohler's criterion\n%s constraints were syntactically redundant\n%s constraints were remove using inclusion tests\n"
		    (Nb.Z.pr !fmeGenCnt)
		    (Nb.Z.pr !fmeRedKohlerCnt)
		    (Nb.Z.pr !fmeRedSynCnt)
		    (Nb.Z.pr !fmeRedSxCnt)

end

let nil : t = []

let initc s = List.map Cons.initc s
let extractc s = List.map Cons.extractc s

let pr: (Var.t -> string) -> t -> string
= fun varPr s -> List.fold_left (fun str c -> str ^ (Cons.pr varPr c) ^ "\n") "" s

let prExt: (Var.t -> string) -> t -> string
= fun varPr s ->
	List.fold_left (fun str c -> str ^ (Cons.prExt varPr c) ^ "\n") "" s

let isEq s1 s2 =
	let incl l1 = List.for_all (fun c2 ->
		List.exists (fun c1 -> Cstr.isInclSyn (Cons.get_c c1) (Cons.get_c c2)) l1)
	in
	incl s1 s2 && incl s2 s1

(** [mkCert i pfi pfs] builds an inclusion certificate for the constraint [c]
identified by [i]. Substituting variables with their definition in [c] yield
a constraint [c'], which is the result of the linear combination [pfs]. Then,
[c'] is implied by the linear combination [pfi]. The function [mkCert] builds
the linear combination which directly implies [c]. *)
let mkCert: int -> Cert.frag_t -> Cert.frag_t -> Cert.cstr_t
= fun i pfi pfs ->
	let (d, w) = List.partition (fun (i', _) -> i = i') pfs in
	let wInv = List.map (fun (i', a) -> (i', Nb.neg a)) w in
	match d with
	| [_, a] ->
		let norm: Cert.frag_t -> Cert.frag_t
		=	let a1 =
				if Nb.cmpz a < 0
				then a
				else failwith "IneqSet.mkCert" (* XXX: useful check? *)
			in
			fun f -> List.map (fun (i', a2) -> (i', Nb.div a2 a1)) f
		in
		Cert.Direct (Cert.mkF (norm (List.append pfi wInv)))
	| _ -> failwith "IneqSet.mkCert"

let synIncl es s c =
	(* XXX: Is this check necessary? *)
	match Cstr.tellProp (Cons.get_c c) with
	| Cstr.Trivial -> Trivial
	| Cstr.Contrad -> Empty (Cert.mkF [(Cons.get_id c), Nb.u])
	| Cstr.Nothing ->

	(* XXX: Check that c.Cons.id is well-defined at every call site and get rid of the -1 dance. *)
	let c1 = EqSet.filter es (Cons.mk (-1) (Cons.get_c c)) in
	match Cstr.tellProp (Cons.get_c c1) with
	| Cstr.Trivial -> Implied (Cons.get_id c, mkCert (-1) [] (Cons.get_f c1))
	| Cstr.Contrad -> Empty (List.map (fun (id, n) -> if id = -1 then (Cons.get_id c, n) else (id, n)) (Cons.get_f c1))
	| Cstr.Nothing ->

	try
		let cI = List.find (fun c2 -> Cons.implies c2 c1) s in
		let ratio = (* XXX: should be done elsewhere *)
			match Vec.isomorph (Cstr.get_v (Cons.get_c c1)) (Cstr.get_v (Cons.get_c cI)) with
			| None -> failwith "IneqSet.synIncl"
			| Some r -> r
		in
		let ce = mkCert (-1) [(Cons.get_id cI), ratio] (Cons.get_f c1) in
		Implied (Cons.get_id c, ce)
	with
		Not_found ->
			let (d, w) = List.partition (fun (id, _) -> id = -1) (Cons.get_f c1) in
			match d with
			| [_, a] ->
				let f = Cert.mergeF [(Cons.get_f c), a; w,  Nb.u] in
				Check { c1 with Cons.id = Cons.get_id c; Cons.f = f }

			| _ -> failwith "IneqSet.synIncl"

let fromUnsat (w: Splx.witness_t) (id: int): Cert.frag_t =
	let (d, w1) = List.partition (fun (i, _) -> id = i) w in
	match d with
	| [] | _::_::_ -> invalid_arg "IneqSet.fromUnsat"
	| [_, a] ->
		let s = Nb.cmpz a in
		if s = 0 then
			invalid_arg "IneqSet.fromUnsat"
		else
			(* XXX: This should be checked further,
			but I believe that the test below is never true.
			The intuition is that a constraint cannot be multiplied
			by a negative scalar. *)
			let norm =
				if s > 0 then
					fun (i, a1) -> (i, Nb.div a1 (Nb.neg a))
				else
					fun (i, a1) -> (i, Nb.div a1 a)
			in
			Cert.mkF (List.map norm w1)

(* XXX: the -1 dance is necessary to handle the case of a constraint
identified by [i] involved in proving that a constraint identified by
the same [i] is implied. *)
let isIncl: Var.t -> EqSet.t -> t -> t -> rel_t
=	let chId: int -> Cons.t -> Cons.t
	= fun i c ->
		let f =
			let ch1 (i', n) =
				if Cons.get_id c = i'
				then (i, n)
				else (i', n)
			in
			List.map ch1 (Cons.get_f c)
		in
		{c with Cons.id = i; Cons.f = f}
	in
	fun nxt es s1 s2 ->
		let rec _isIncl cert optSx
		= function
		| [] -> Incl cert
		| c::t ->
			match synIncl es s1 (chId (-1) c) with
			| Empty _ -> NoIncl
			| Trivial -> failwith "IneqSet.isIncl"
			| Implied (id, ce) -> _isIncl ((Cons.get_id c, ce)::cert) optSx t
			| Check c1 ->
				let sx =
					match optSx with
					| Some sx -> sx
					| None ->
						let mkPair c = (Cons.get_id c, Cons.get_c c) in
						Splx.mk nxt (List.map mkPair s1)
				in
				let c1' = Cstr.compl (Cons.get_c c1) in
				match Splx.checkFromAdd (Splx.addAcc sx (-1, c1')) with
				| Splx.IsOk _ -> NoIncl
				| Splx.IsUnsat w ->
					let ce = mkCert (-1) (fromUnsat w (-1)) (Cons.get_f c1) in
					_isIncl ((Cons.get_id c, ce)::cert) (Some sx) t
	in
	_isIncl [] None s2

let synAdd es s c =
	match synIncl es s c with
	| Trivial | Implied _ -> ((Cons.get_id c), s)
	| Empty _ -> failwith "IneqSet.synAdd"
	| Check c1 ->
		let s1 = c1::(List.filter (fun c2 -> not (Cons.implies c1 c2)) s) in
		((Cons.get_id c1) + 1, s1)

let trim s0 sx0 =
	let check c (s, sx) =
		let sx1 = Splx.compl sx (Cons.get_id c) in
		match Splx.check sx1 with
		| Splx.IsOk _ -> c::s, sx
		| Splx.IsUnsat _ -> s, Splx.forget sx (Cons.get_id c)
	in
	(* XXX: Why is this fold_right? *)
	List.fold_right check s0 ([], sx0)

(** [searchEqWith sx c] looks for an implicit equality involving [c] in [sx],
where [sx] represents the subspace constrained by an inequality set.
If such an equality is found, then it is returned, along with a certificate
showing that it holds. Otherwise, [searchEqWith] returns [None].
XXX: what is assumed on [sx]? *)
let searchEqWith: Splx.t -> Cons.t -> (Cons.t * Cert.cstr_t) option
=	let build : Cons.t -> Splx.witness_t -> (Cons.t * Cert.cstr_t) option
	= fun c w ->
		let eq = {(Cons.get_c c) with Cstr.typ = Cstr.Eq} in
		let f1 = Cert.mkF [(Cons.get_id c), Nb.u] in
		let f2 = fromUnsat w (Cons.get_id c) in
		Some (Cons.mk (Cons.get_id c) eq, Cert.SplitEq (f1, f2))
	in
	fun sx c ->
		if (Cstr.get_typ (Cons.get_c c)) <> Cstr.Le
		then None
		else
			match Splx.stricten sx (Cons.get_id c) with
			| Splx.IsUnsat w -> build c w
			| Splx.IsOk sx1 ->
				match Splx.check sx1 with
				| Splx.IsUnsat w -> build c w
				| Splx.IsOk _ -> None

(** Notation for a list of newly-found equalities and a proof of them holding. *)
type eqListT = (Cons.t * Cert.cstr_t) list

(** [searchEqs sx s] looks for all implicit equalities in [s].
The function returns a list of implicit equalities [l], with accompanying proofs
and a set of inequalities [s'] with no implicit equalities, such that [s] is the
intersection of [s'] and [l].
The object [sx] must represent the subspace constrained by [s].
XXX: what is assumed on [sx]? *)
let searchEqs: Splx.t -> t -> t * eqListT
=	let search1: Splx.t -> t * eqListT -> Cons.t -> t * eqListT
	= fun sx (s, l) c ->
		match searchEqWith sx c with
		| None -> (c::s, l)
		| Some (e, pfe) -> (s, (e, pfe)::l)
	in
	fun sx s -> List.fold_left (search1 sx) ([], []) s

type meetT =
	| Added of (t * Cert.equivT)
	| ImplicitEqs of ((t * (Cstr.t * Cert.cstr_t) list) * Cert.equivT)
	| Contrad of Cert.frag_t

let meetPr: (Var.t -> string) -> meetT -> string
= fun varPr -> function
	| Added (s, (ce, ce')) -> Printf.sprintf
		"Added (s, (ce, ce')) with\ns:\n%s\nce:\n%s\nce':\n%s\n"
		(prExt varPr s) (Cert.partialPr ce) (Cert.partialPr ce')
	| ImplicitEqs ((s, l), (ce, ce')) ->
		let pr1: Cstr.t * Cert.cstr_t -> string
		= fun (c, ce) -> Printf.sprintf "%s proved with\n%s\n"
			(Cstr.pr varPr c) (Cert.prC ce)
		in
		let lstr = String.concat "\n" (List.map pr1 l) in
		Printf.sprintf
		"ImplicitEqs ((s, l), (ce, ce')) with\ns:\n%s\nl:\n%s\nce:\n%s\nce':\n%s\n"
		(prExt varPr s) lstr (Cert.partialPr ce) (Cert.partialPr ce')
	| Contrad f -> Printf.sprintf "Contrad: %s" (Cert.prF f)

let meetEq: meetT -> meetT -> bool
=	fun ar ar' ->
		match ar, ar' with
		| Added (s, (ce1, ce2)), Added (s', (ce1', ce2')) ->
			isEq s s' && Cert.partialEq ce1 ce1' && Cert.partialEq ce2 ce2'
		| ImplicitEqs ((s, l), (ce1, ce2)), ImplicitEqs ((s', l'), (ce1', ce2')) ->
			let hasEq: Cstr.t * Cert.cstr_t -> (Cstr.t * Cert.cstr_t) list -> bool
			= fun (c, ce) l ->
				List.exists (fun (c', ce') ->
					Cstr.isEqSyn c c' && Cert.isEqC ce ce') l
			in
			List.for_all (fun p -> hasEq p l) l' &&
			List.for_all (fun p -> hasEq p l') l &&
			isEq s s' && Cert.partialEq ce1 ce1' && Cert.partialEq ce2 ce2'
		| Contrad f, Contrad f' -> Cert.isEqF f f'
		| Added (_, _), ImplicitEqs (_, _)
		| Added (_, _), Contrad _
		| ImplicitEqs (_, _), Added (_, _)
		| ImplicitEqs (_, _), Contrad _
		| Contrad _, Added (_, _)
		| Contrad _, ImplicitEqs (_, _) -> false

type satChkT = Sat of Splx.t | Unsat of Cert.frag_t

(** [chkFeasibility nvar s] checks whether [s] is satisfiable and returns a
simplex object with a satisfied state if it is. If it is not satisfiable, then
a linear combination of the input constraints is returned. [nvar] is used for
fresh variable generation. *)
let chkFeasibility: Var.t -> t -> satChkT
= fun nvar s ->
	let cs = List.map (fun c -> (Cons.get_id c, Cons.get_c c)) s in
	match Splx.checkFromAdd (Splx.mk nvar cs) with
	| Splx.IsOk sx -> Sat sx
	| Splx.IsUnsat w -> Unsat w

(** [isRed sx i] checks whether the constraint identified by [i] is redundant in
the constraint set represented by [sx]. If it is, a certificate proving it is
returned. *)
let isRed: Splx.t -> int -> Cert.cstr_t option
= fun sx i ->
	let sx' = Splx.compl sx i in
	match Splx.check sx' with
	| Splx.IsOk _ -> None
	| Splx.IsUnsat w -> Some (Cert.Direct (fromUnsat w i))

(** [rmRed s sx] removes all the redundancy from [s]. The simplex object [sx] is
assumed to be in a statisfied state (it has been returned by [Splx.check]) and
to represent [s]. The returned simplex object is [sx] with all the redundant
bounds [Splx.forgot]ten. The returned certificate list contains a certificate
for all the removed constraints. *)
let rmRedAux: t -> Splx.t -> Splx.t * (t * (int * Cert.cstr_t) list)
=
	let rec rmRed1:
					Splx.t * (t * (int * Cert.cstr_t) list) ->
					Cons.t ->
					Splx.t * (t * (int * Cert.cstr_t) list)
	= fun (sx, (s, l)) c ->
		let i = Cons.get_id c in
		match isRed sx i with
		| None -> (sx, (c::s, l))
		| Some ce -> (Splx.forget sx i, (s, (i, ce)::l))
	in
	fun s sx -> List.fold_left rmRed1 (sx, ([], [])) s

(** [inclCerts l s s'] build the certificates for [s] implies [s'] and [s']
implies [s], where [s'] is [s] with the constraints described in [l] removed
for the given reason. *)
let inclCerts: (int * Cert.cstr_t) list -> t -> t -> Cert.equivT
= fun l s s' ->
	let ce =
		let subst: int * Cert.cstr_t -> int * Cert.cstr_t
		= fun (i , ce) ->
			if List.mem_assoc i l
			then (i, List.assoc i l)
			else (i, ce)
		in
		List.map subst (extractc s)
	in
	(extractc s', ce)

type redResT
= RedOk of t * Cert.equivT | RedBot of Cert.frag_t

(** Knowing that [c] syntactically implies [c'], [certSyn c c'] builds the
fragment which proves it. *)
let certSyn: Cons.t -> Cons.t -> int * Cert.cstr_t
= fun c c' ->
	let v = Cstr.get_v (Cons.get_c c) in
	let v' = Cstr.get_v (Cons.get_c c') in
	match Vec.isomorph v' v with
	| Some r -> (Cons.get_id c', Cert.Direct (Cert.mkF [Cons.get_id c, r]))
	| None -> failwith "IneqSet.addM.certSyn"

(** [isIncl s c] checks whether [c] is implied by [s] on syntactic grounds.
The optional certificate only refers to the input constraints. *)
let isInclSyn: t -> Cons.t -> (int * Cert.cstr_t) option
= fun s c ->
	match List.filter (fun c' -> Cons.implies c' c) s with
	| [] -> None
	| [c'] -> Some (certSyn c' c)
	| _ -> failwith "IneqSet.addM"

(** [rmSynRed s l] removes syntactically redundant constraints from the union
of [s] and [l], assuming that there are not such constraints when considering
[s] alone.  For each removed constraint, a certificate fragment is given. *)
let rmRedSyn: t -> Cons.t list -> t * (int * Cert.cstr_t) list
=	let rm1: t * Cert.partialT -> Cons.t -> t * Cert.partialT
	= fun (s, cel) c ->
		match isInclSyn s c with
		| Some ce -> (s, ce::cel)
		| None ->
			let (s1, s2) = List.partition (fun c' -> Cons.implies c c') s in
			(c::s2, (List.map (fun c' -> certSyn c c') s1) @ cel)
	in
	fun s l -> List.fold_left rm1 (s, []) l

let rmRed: Var.t -> t -> redResT
= fun nvar s ->
	let (s1, l1) = rmRedSyn [] s in
	match chkFeasibility nvar s1 with
	| Unsat f -> RedBot f
	| Sat sx ->
		let (_, (s', l)) = rmRedAux s1 sx in
		RedOk (s', inclCerts (l1 @ l) s s')

let addM: Var.t -> t -> (int * Cstr.t) list -> meetT
=
(** [searchEqs s sx] finds all the implicit equalities in the inequality
system [s]. To do so, it uses [sx], which is expected to be in a satisfiable
state and to represent [s].
Note that this version of [searchEqs] has the same shortcomings as the top-level
one but does not remove the inequalities that yield an implicit equality. *)
	let searchEqs: t -> Splx.t -> eqListT
	=	let search1: Splx.t -> eqListT -> Cons.t -> eqListT
		= fun sx l c ->
			match searchEqWith sx c with
			| None -> l
			| Some (e, pfe) -> (e, pfe)::l
		in
		fun s sx -> List.fold_left (search1 sx) [] s
	in
(** Check that all the constraints in the list are inequalities. *)
	let allIneqs: (int * Cstr.t) list -> bool
	=	let isIneq: int * Cstr.t -> bool
		= fun (_, c) ->
			match Cstr.get_typ c with
			| Cstr.Eq -> false
			| Cstr.Le | Cstr.Lt -> true
		in
		fun l -> List.for_all isIneq l
	in
(** Check whether there are trivially contradictory constraints in the given list.
If there are, the identifier of one of them is returned. *)
	let hasTrivContrad: (int * Cstr.t) list -> int option
	=	let isContrad: int * Cstr.t -> bool
		= fun (_, c) -> Cstr.tellProp c = Cstr.Contrad
		in
		fun l ->
			try Some (Pervasives.fst (List.find isContrad l))
			with Not_found -> None
	in
(** Remove trivial constraints from a given list of constraints.
A certificate fragment is returned for each of the removed constraints. *)
(* XXX: an elegant way of merging [rmTriv] and [hasTrivContrad] should be found. *)
	let rmTriv: Cons.t list -> Cons.t list * Cert.partialT
	=	let isTriv: Cons.t -> bool
		= fun c -> Cstr.tellProp (Cons.get_c c) = Cstr.Trivial
		in
		fun l ->
			let (l1, l2) = List.partition isTriv l in
			(l2, List.map (fun c -> (Cons.get_id c, Cert.Direct [])) l1)
	in
	fun nvar s l -> 
		let s1 = initc s in
		if not (allIneqs l)
		then failwith "IneqSet.addM"
		else
			match hasTrivContrad l with
			| Some i -> Contrad (Cert.mkF [i, Nb.u])
			| None ->
				let cl = List.map (fun (i, c) -> Cons.mk i c) l in
				let (cl', trivl) = rmTriv cl in
				let (s2, synredl) = rmRedSyn s1 cl' in
				match chkFeasibility nvar s2 with
				| Unsat f -> Contrad f
				| Sat sx ->
					let (sx', (s3, l2)) = rmRedAux s2 sx in
					let certs = inclCerts (trivl @ synredl @ l2) (s1 @ cl) s3 in
					let el =
						let l = searchEqs s3 sx' in
						List.map (fun (c, ce) -> (Cons.get_c c, ce)) l
					in
					if List.length el = 0
					then Added (s3, certs)
					else ImplicitEqs ((s3, el), certs)

let rename s fromX toY = List.map (Cons.rename fromX toY) s

let pick : Var.t option Rtree.t -> t -> Var.t option = fun msk s ->
	let update v n p m =
		match Nb.cmpz n with
		| 0 -> (v, p, m)
		| n when n < 0 -> (v, p + 1, m)
		| _ -> (v, p, m + 1)
	in
	let rec build m cnt ve =
		match m, cnt, ve with
		| Rtree.Nil, _, _ -> Rtree.Nil
		| _, rt, Rtree.Nil -> rt
		| Rtree.Sub (l1, n1, r1), Rtree.Nil, Rtree.Sub (l2, n2, r2) ->
			let l = build l1 Rtree.Nil l2 in
			let r = build r1 Rtree.Nil r2 in
			let n =
				if n1 = None then
					(None, 0, 0)
				else
					update n1 n2 0 0
			in
			Rtree.Sub (l, n, r)

		| Rtree.Sub (l1,n1,r1), Rtree.Sub (l2,n2,r2), Rtree.Sub (l3,n3,r3) ->
			let l = build l1 l2 l3 in
			let r = build r1 r2 r3 in
			let n =
				if n1 = None then
					n2
				else
					let (_, p, m) = n2 in
					update n1 n3 p m
			in
			Rtree.Sub (l, n, r)
	in
	let scores = List.fold_left (fun a c -> build msk a (Cstr.get_v (Cons.get_c c)))
		Rtree.Nil s
	in
	let (optv, _) =
		let choose (best, sc) (v, p, m) =
			if v = None || p = 0 && m = 0 then
				(best, sc)
			else
				let nsc = p * m - (p + m) in
				if best = None then
					(v, nsc)
				else
					if nsc < sc then
						(v, nsc)
					else
						(best, sc)
		in
		Rtree.fold choose (None, -1) scores
	in
	optv

let add: Var.t -> t -> int -> Cstr.t -> meetT
= fun nvar s i c -> addM nvar s [i, c]

let meet: Var.t -> t -> t -> meetT
= fun nvar s s' ->
	let slist' = List.map (fun c -> (Cons.get_id c, Cons.get_c c)) s' in
	addM nvar s slist'

let trimSet nxt (id, s) =
	let cl = List.map (fun c -> Cons.get_id c, Cons.get_c c) s in
	match Splx.checkFromAdd (Splx.mk nxt cl) with
	| Splx.IsUnsat _ -> Pervasives.failwith "IneqSet.trimSet"
	| Splx.IsOk sx -> (id, Pervasives.fst (trim s sx))

let simpl: Var.t -> EqSet.t -> t -> simpl_t
= fun nxt es s ->
	let rec filter s1 = function
	| [] ->
		let cl = List.map (fun c -> Cons.get_id c, Cons.get_c c) s1 in
		(match Splx.checkFromAdd (Splx.mk nxt cl) with
		| Splx.IsOk sx -> SimplMin (Pervasives.fst (trim s1 sx))
		| Splx.IsUnsat w -> SimplBot w)
	| h::t ->
		match synIncl es s1 h with
		| Trivial | Implied _ -> filter s1 t
		| Empty f -> SimplBot f
		| Check h1 ->

		filter (h1::(List.filter (fun c -> not (Cons.implies h1 c)) s1)) t
	in
	filter [] (List.rev s)

let subst: Var.t -> EqSet.t -> Var.t -> int -> Cons.t -> t -> int * t
= fun nxt es x id e s ->
	let gen (id, s) c =
		let c1 =
			if Nb.cmpz (Vec.get (Cstr.get_v (Cons.get_c c)) x) = 0 then
				{ c with Cons.id = id }
			else
				Cons.elimc id x e c
		in
		synAdd es s c1
	in
	trimSet nxt (List.fold_left gen (id, nil) s)

let fmElim: Var.t -> int -> EqSet.t -> Var.t -> int -> t -> int * t
= fun nxt nth es x id s ->
	let (pos, z, neg) = List.fold_left (fun (p, z, n) c ->
		match Nb.cmpz (Vec.get (Cstr.get_v (Cons.get_c c)) x) with
		| 0 -> (p, c::z, n)
		| a when a < 0 -> (c::p, z, n)
		| _ -> (p, z, c::n)) ([], [], []) s
	in
	let (id1, zs) = List.fold_left
		(fun (id, s) c -> (id + 1, { c with Cons.id = id }::s)) (id, nil) z
	in
	let add : t -> Cons.t -> int * t
	  = fun s c ->
	  begin
	    Stats.fmeGen 1;
	    if List.length (Cons.get_f c) > nth + 1
	    then
	      begin
		Stats.fmeRedKohler 1;
		(Cons.get_id c, s);
	      end
	    else
	      let (i, s') = synAdd es s c in
	      begin
		Stats.fmeRedSyn (List.length s + 1 - List.length s');
		(i, s');
	      end
	  end
	in
	let apply (id0, s0) c = List.fold_left
		(fun (id, s) c1 -> add s (Cons.elimc id x c c1)) (id0, s0) neg
	in
	let (i', s') = List.fold_left apply (id1, zs) pos in
	let (i'', s'') = trimSet nxt (i', s') in
	begin
	  Stats.fmeGen (List.length zs);
	  Stats.fmeRedSx (List.length s' - List.length s'');
	  (i'', s'');
	end

let fmElimM: Var.t -> int -> EqSet.t -> int -> Var.t option Rtree.t -> t -> int * t
= fun nxt nth0 es i msk s ->
	let rec elim nth cnt i1 s1 =
		match pick msk s1 with
		| None ->
			if cnt = 0 then (* renumber even if nothing to project *)
				List.fold_right (fun c (id, s2) ->
					(id + 1, { c with Cons.id = id }::s2)) s1 (i, nil)
			else
				(i1, s1)

		| Some x ->
			let (i2, s2) = fmElim nxt nth es x i s1 in
			elim (nth + 1) (cnt + 1) i2 s2
	in
	elim nth0 0 i s

let joinSetup: bool -> Var.t -> Var.t option Rtree.t -> Var.t -> int -> t -> Var.t * Var.t option Rtree.t * t
= fun dup nxt relocTbl alpha idOffset s ->
	let apply (nxt1, relocTbl1, s1) c =
		let (nxt2, relocTbl2, c1) = Cons.joinSetup dup nxt1 relocTbl1 alpha idOffset c in
		(nxt2, relocTbl2, c1::s1)
	in
	List.fold_left apply (nxt, relocTbl, nil) s

let isTop : t -> bool = fun s -> (s = [])
let list : t -> Cons.t list = fun s -> s
