(** {! Pol} treats equalities and inequalities differently.
This module is the data structure that hosts the inequalities of a polyhedron. *)

type t = Cons.t list

type prop_t =
| Empty of Cert.frag_t
| Trivial
| Implied of int * Cert.cstr_t
| Check of Cons.t

type rel_t =
| NoIncl
| Incl of (int * Cert.cstr_t) list

type simpl_t =
| SimplMin of t
| SimplBot of Cert.frag_t

val nil: t

(** Result type for the intersection of an inequality set [s]
with one new inequality [c]. A result [Added (s', (ce1, ce2)] indicates that [c]
has successful been added to [s]. The certificate [ce1] proves the inclusion of
[c::s] in [s'] and [ce2] proves the inclusion of [s'] in [c::s].
A result [ImplicitEqs ((s', l), (ce1, ce2))] is identical to
[Added (s', (ce1, ce2))], with the extra information that there are implicit
equalities in [s']. A result [Redundant ce] indicates that [c] is implied by [s],
which is shown by [ce]. Last, a result [Contrad f] indicates that the resulting
set has no solution, which is evidenced by the linear combination [f] yielding
a trivially contradictory constraint. *)
(* XXX [ImplicitEqs] breaks the invariant on a value of type [t]: the returned set
contains implicit equalities. *)
type meetT =
	| Added of (t * Cert.equivT)
	| ImplicitEqs of ((t * (Cstr.t * Cert.cstr_t) list) * Cert.equivT)
	| Contrad of Cert.frag_t

(** Pretty-printer for values of type [meetT]. *)
val meetPr: (Var.t -> string) -> meetT -> string

(** Equality test for values of type [meetT].
Constraint equality is checked with [Cstr.isEqSyn],
Certificate equality is checked with [Cert.isEqC],
Fragment equalitiy is checked with [Cert.isEqF],
[Cert.equivT] equality is checked with [Cert.partialEq]. *)
val meetEq: meetT -> meetT -> bool

(** [addM nvar s l] adds the inequalities [l] to the set [s].
If any [c] in [l] is an equality, then [Failure "IneqSet.addM"] is raised.
All returned linear combinations are expressed with respect to the input
constraints (all certificate fragments are re-initialised).
[nvar] is used to generate fresh variables. *)
val addM: Var.t -> t -> (int * Cstr.t) list -> meetT

(** [add nvar s i c] adds one constraint [c], uniquely identified by [i] to [s].
This is a short-hand notation for [addM nvar s [i, c]]. *)
val add: Var.t -> t -> int -> Cstr.t -> meetT

(** [meet nvar s s'] compute the intersection between two inequality sets.
It is implemented using [addM]. *)
val meet: Var.t -> t -> t -> meetT

(** The result of a representation reduction operation on a set [s].
	A result [RedOk (s', (ce1, ce2))] indicates that [s'] is a reduced
representation for [s] and [ce1] and [ce2] show that [s] implied [s']
and that [s'] implies [s], respectively.
	A result [RedBot f] indiquates that the set has no solution and
gives a linear combination yielding a trivially contradictory constraint
to show it. *)
type redResT
= RedOk of t * Cert.equivT | RedBot of Cert.frag_t

(** [rmRed nvar s] eliminates redundancy from [s].
In its current form, it does not look for implicit equalities, though.
[nvar] is used to generate fresh variables. *)
(* XXX: add a syntatic elimination pass? *)
val rmRed: Var.t -> t -> redResT

val isIncl: Var.t -> EqSet.t -> t -> t -> rel_t
val isEq: t -> t -> bool
val rename: t -> Var.t -> Var.t -> t

val pick: Var.t option Rtree.t -> t -> Var.t option
val subst: Var.t -> EqSet.t -> Var.t -> int -> Cons.t -> t -> int * t
val simpl: Var.t -> EqSet.t -> t -> simpl_t
val fmElim: Var.t -> int -> EqSet.t -> Var.t -> int -> t -> int * t
val fmElimM: Var.t -> int -> EqSet.t -> int -> Var.t option Rtree.t -> t -> int * t

val joinSetup: bool -> Var.t -> Var.t option Rtree.t -> Var.t -> int -> t -> Var.t * Var.t option Rtree.t * t

(** [fromUnsat w i] turns an unsatisfiability witness [w] from the simplex algorithm
into a linear combination yielding the constraint identified by [i].
The simplex algorithm was given the complement (identified by [i]) of constraint [i].
The identifier [i] must appear exactly once in [w] and must have a non-zero coefficient;
exception [Invalid_argument "IneqSet.fromUnsat"] is raised otherwise. *)
val fromUnsat: Splx.witness_t -> int -> Cert.frag_t

val synIncl: EqSet.t -> t -> Cons.t -> prop_t

(** All the linear combinations that lead to the current inequality set were recorded, starting from when [initc] was called on it.
[extractc] builds the certificate that shows that the resulting (current) inequality set is included in the starting point. *)
val extractc: t -> (int * Cert.cstr_t) list

(** The linear combinations that constraints within an inequality set are subjected to are recorded,
from a given starting point. Calling [initc] on an inequality set defines such a starting point. *)
val initc: t -> t

val list: t -> Cons.t list
val isTop: t -> bool

val pr: (Var.t -> string) -> t -> string
val prExt: (Var.t -> string) -> t -> string

module Stats : sig
  val reset : unit -> unit
  val pr : unit -> string

end
