(** This module offers an interface to the number manipulated throughout the VPL.
The implementation uses the ZArith front-end to GMP. *)

(** The type of rational numbers. The internal representation is canonical. *)
type t

(** Zero value. *)
val z: t

(** Value 1. *)
val u: t

(** Value -1. *)
val negU: t

(** [cmp n1 n2] compares two rational numbers.
The result is negative if [n1] < [n2], nil if [n1] = [n2] and positive if [n1] > [n2].
This is compatible with [Pervasives.compare]. *)
val cmp: t -> t -> int

(** [cmpz n] is equivalent to [cmp z n]. *)
val cmpz: t -> int

(** [isZ n] returns true is [n] is zero ([z]) and false otherwise. *)
val isZ : t -> bool

(** Some operations can yield infinity (e.g. 1/0) or undefined results (0/0).
[isReal n] returns true if [n] is neither infinite nor undefined. *)
val isReal: t -> bool

(** [isInt n] returns true if [n] is a relative integer
that is if its denominator is equal to [u]. *)
val isInt: t -> bool

(** Compute the absolute value. *)
val abs: t -> t

(** Compute the opposite of a number. *)
val neg: t -> t

(** Compute the inverse of a number. *)
val inv: t -> t

(** Add two rationals. *)
val add: t -> t -> t

(** [sub n1 n2] yields [n1] - [n2]. *)
val sub: t -> t -> t

(** Multiply two rationals. *)
val mult: t -> t -> t

(** [div n1 n2] divides [n1] by [n2]. *)
val div: t -> t -> t

(** [mk i] builds an element of type [t] and of value [i].
[mk ~den:i1 i2] builds an element of type [t] and of value [i2]/[i1]. *)
val mk: int -> int -> t
val mk1 : int -> t

(** Build a rational from a floating-point number. *)
(* XXX: precision? *)
val of_float: float -> t

(** Build a rational from a ZArith rational. *)
val ofQ : Q.t -> t

(** Convert a rational to a ZArith rational. *)
val toQ : t -> Q.t

(** Pretty-print a rationnal number. *)
val pr: t -> string

(** The type of relative integers that rationals are made out of, and operations on them.
This mainly used in {! Glue}. *)
module Z:
sig
	type t

	val z: t
	val u: t

	(** addition *)
	val add : t -> t -> t

	(** [sub a b] yields [a] - [b] *)
	val sub : t -> t -> t

	(** multiplication *)
	val mul : t -> t -> t

	(** [div a b] yields the result of integer division [a]/[b].
	The result obeys the rule of signs.  Division by zero has
	undefined behaviour. *) 
	val div : t -> t -> t

	(** [rem a b] yields the remainder of integer division [a]/[b].
	The result has the sign of [a].  [rem a 0] has undefined behaviour. *)
	val rem : t -> t -> t

	val cmp: t -> t -> int
	val neg: t -> t
	val lAnd: t -> t -> t
	val orL: t -> t -> t
	val shiftL: t -> int -> t
	val shiftR: t -> int -> t

	val mk: int -> t

	(** build a value of type [t] from its string representation in base 10 *)
	val ofString : string -> t

	val toInt : t -> int
	val pr: t -> string
end

(** [toZ n] returns the numerator and denominator of [n], in that order. *)
val toZ: t -> Z.t * Z.t

(** [ofZ n d] builds a rational of numerator [n] and denominator [d]. *)
val ofZ: Z.t -> Z.t -> t

(** [gcd nGcd dGcd n] returns the greatest common divisor of nGcd and the numerator of [n]
and that of [dGcd] and the denominator of [n]. *)
val gcd: Z.t * Z.t -> t -> Z.t * Z.t
