type t = {
	eqs: EqSet.t;
	ineqs: IneqSet.t;
	nxt: int }

let get_eqs (x : t) = x.eqs
let get_ineqs (x : t) = x.ineqs
let get_nxt (x : t) = x.nxt

let pr: (Var.t -> string) -> t -> string
= fun varPr p ->
	if EqSet.isTop p.eqs && IneqSet.isTop p.ineqs
	then "top"
	else (EqSet.pr varPr p.eqs) ^ (IneqSet.pr varPr p.ineqs)

let prRaw: t -> string
= fun p -> pr Var.pr p

let prExt: (Var.t -> string) -> t -> string
= fun varPr p ->
	let estr =
		if EqSet.isTop p.eqs
		then " top"
		else "\n" ^ (EqSet.prExt varPr p.eqs)
	in
	let istr =
		if IneqSet.isTop p.ineqs
		then " top"
		else "\n" ^ (IneqSet.prExt varPr p.ineqs)
	in
	Printf.sprintf "{\n\tnxt = %i;\n\teqs =%s;\n\tineqs =%s\n}" p.nxt estr istr

let prExtRaw: t -> string
= fun p -> prExt Var.pr p

type rel_t =
| NoIncl
| Incl of Cert.t

type bndT =
	| Infty
	| Open of Nb.t
	| Closed of Nb.t

type itvT = { low: bndT; up: bndT }

let get_up (i: itvT) = i.up
let get_low (i: itvT) = i.low

type cstT = S of Nb.t | I of itvT

(** The description of an assignment operation.
[var] is given the value [lin] + [cst]. *)
type assign_t = {
	var: Var.t;
	lin: (Nb.t * Var.t) list;
	cst: cstT }

let getVar: assign_t -> Var.t
= fun a -> a.var

let getLin: assign_t -> (Nb.t * Var.t) list
= fun a -> a.lin

let getCst: assign_t -> cstT
= fun a -> a.cst

let prItv: (Var.t -> string) -> Vec.t -> itvT -> string
= fun varPr v itv ->
	let prLow = function
		| Infty -> "-infty < "
		| Open n -> (Nb.pr n) ^ " < "
		| Closed n -> (Nb.pr n) ^ " <= "
	in
	let prUp = function
		| Infty -> " < +infty"
		| Open n -> " < " ^ (Nb.pr n)
		| Closed n -> " <= " ^ (Nb.pr n)
	in
	(prLow itv.low) ^ (Vec.pr varPr v) ^ (prUp itv.up)

let prItvRaw: Vec.t -> itvT -> string
= fun v itv -> prItv Var.pr v itv

(** Initialize the certificate fragment associated with every constraint,
so that a subsequently built certificate will refer to the input polyhedron. *)
let fragInit: t -> t
= fun p -> {p with eqs = EqSet.initc p.eqs; ineqs = IneqSet.initc p.ineqs}

let isEqSyn p1 p2: bool =
	let incl l1 l2 = List.for_all
		(fun c2 -> List.exists
			(fun c1 -> Cstr.isInclSyn (Cons.get_c c1) (Cons.get_c c2)) l1) l2
	in
	let eq l1 l2 = incl l1 l2 && incl l2 l1 in
	EqSet.isEq p1.eqs p2.eqs && eq p1.ineqs p2.ineqs

(* XXX: optimize away trivial bindings? inline everything? *)
let certCombineFwd: (int * Cert.cstr_t) list list -> Cert.t -> Cert.t
=
	let maxId: (int * Cert.cstr_t) list -> int
	=	let maxIdF: int -> Cert.frag_t -> int
		= fun m f -> List.fold_left (fun m (i, _) -> Pervasives.max m i) m f
		in
		let maxIdC: int -> int * Cert.cstr_t -> int
		= fun m (_, c) ->
			match c with
			| Cert.SplitEq (f1, f2) -> maxIdF (maxIdF m f1) f2
			| Cert.Direct f -> maxIdF m f
		in
		fun l -> List.fold_left maxIdC 0 l
	in
	let module IntOrd
	= struct
		type t = int
		let compare: t -> t -> int = Pervasives.compare
		end
	in
	let module Alloc = Map.Make (IntOrd) in
	let idRealloc: int -> (int * Cert.cstr_t) list -> int * int Alloc.t
	=	let realloc1: int * int Alloc.t -> int * Cert.cstr_t -> int * int Alloc.t
		= fun (i, tbl) (i', c) ->
			if Alloc.mem i' tbl
			then invalid_arg "Pol.certCombine"
			else (i+1, Alloc.add i' i tbl)
		in
		fun m l -> List.fold_left realloc1 (m, Alloc.empty) l
	in
	let chgIds: int Alloc.t -> (int * Cert.cstr_t) list -> (int * Cert.cstr_t) list
	=	let f tbl
		= fun (i, c) ->
			if Alloc.mem i tbl
			then (Alloc.find i tbl, c)
			else failwith "Pol.certCombineFwd"
		in
		fun tbl l -> List.map (f tbl) l
	in
	let chgF: int Alloc.t -> Cert.frag_t -> Cert.frag_t
	=	let f tbl
		= fun (i, n) ->
			if Alloc.mem i tbl
			then (Alloc.find i tbl, n)
			else (i, n)
		in
		fun tbl frag -> List.map (f tbl) frag
	in
	let chgIdsF: int Alloc.t -> (int * Cert.cstr_t) list -> (int * Cert.cstr_t) list
	=	let chg1: int Alloc.t -> int * Cert.cstr_t -> int * Cert.cstr_t
		= fun tbl (i, c) ->
			match c with
			| Cert.Direct f -> i, Cert.Direct (chgF tbl f)
			| Cert.SplitEq (f, f') -> i, Cert.SplitEq (chgF tbl f, chgF tbl f')
		in
		fun tbl l -> List.map (fun c -> chg1 tbl c) l
	in
	let bind: (int * Cert.cstr_t) list -> Cert.t -> Cert.t
	=	let bind1: Cert.t -> int * Cert.cstr_t -> Cert.t
		=	let triv = Cstr.eq [] Nb.z in
			fun ce (i, iCe) ->
				if iCe = Cert.Direct [i, Nb.u]
				then ce
				else Cert.Bind (i, triv, iCe, ce)
		in
		fun l ce -> List.fold_left bind1 ce (List.rev l)
	in
	(* XXX: perhaps no need to List.rev before List.rev'ing in bind. *)
	function
		| [] -> fun ce -> ce
		| _::_ as l -> fun ce ->
			let m = maxId (List.hd l) in
			let f: int * int Alloc.t * (int * Cert.cstr_t) list list ->
					(int * Cert.cstr_t) list ->
					int * int Alloc.t * (int * Cert.cstr_t) list list
			= fun (m, tbl, ll) l ->
				let (m', tbl') = idRealloc (m + 1) l in
				(m', tbl', (chgIds tbl' (chgIdsF tbl l))::ll)
			in
			let (_, tbl, l') = List.fold_left f (m, Alloc.empty, []) l in
			let ce' =
				match ce with
				| Cert.Implies l -> Cert.Implies (chgIdsF tbl l)
				| Cert.Empty f -> Cert.Empty (chgF tbl f)
				| Cert.Bind (_, _, _, _) -> failwith "Pol.certCombineFwd"
			in
			bind (List.concat (List.rev l')) ce'

let certCombine: Cert.equivT list -> Cert.t * Cert.t
= function
	| [] -> invalid_arg "Pol.certCombine"
	| _::_ as l ->
		let getLast: (int * Cert.cstr_t) list list ->
			Cert.t * (int * Cert.cstr_t) list list
		= fun l ->
			match List.rev l with
			| [] -> invalid_arg "Pol.certCombine"
			| h::l' -> (Cert.Implies h, List.rev l')
		in
		let (lfwd, lbwd) = List.split l in
		let (cefwd, lfwd') = getLast lfwd in
		let (cebwd, lbwd') = getLast (List.rev lbwd) in
		(certCombineFwd lfwd' cefwd, certCombineFwd lbwd' cebwd)

type rewriteT
= Rewritten of (IneqSet.t * Cert.equivT) | RewriteBot of Cert.frag_t

(** [rewriteIneqs es s] projects from [s] all variables defined in [es].
The produced certificates prove that the result [s'] is implied by [s],
and vice versa. No redundancy elimination is performed on [s']. *)
let rewriteIneqs: EqSet.t -> IneqSet.t -> rewriteT
=	let rewrite: EqSet.t -> rewriteT -> Cons.t -> rewriteT
	= fun es -> function
		| RewriteBot _ as r -> fun _ -> r
		| Rewritten (s, (ce, ce')) -> fun c ->
			let c' = EqSet.filter es c in
			match Cstr.tellProp (Cons.get_c c') with
			| Cstr.Trivial ->
				let i = Cons.get_id c' in
				let bwd = (i, Cert.cstrFromTriv i (Cons.get_f c')) in
				Rewritten (s, (ce, bwd::ce'))
			| Cstr.Contrad -> RewriteBot (Cons.get_f c')
			| Cstr.Nothing ->
				let i = Cons.get_id c' in
				let f = Cons.get_f c' in
				let fwd = (i, Cert.Direct f) in
				let bwd = (i, Cert.cstrInv i f) in
				Rewritten (c'::s, (fwd::ce, bwd::ce'))
	in
	fun es s ->
		List.fold_left (rewrite es) (Rewritten ([], ([], []))) (IneqSet.initc s)

type rewriteBisT
= Rewritten' of ((int * Cstr.t) list * Cert.equivT) | RewriteBot' of Cert.frag_t

let listRw: EqSet.t -> (int * Cstr.t) list -> rewriteBisT
=	let rw1: EqSet.t -> rewriteBisT -> int * Cstr.t -> rewriteBisT
	= fun eset -> function
		| RewriteBot' _ as rw -> fun _ -> rw
		| Rewritten' (ilist, (fwd, bwd)) -> fun (i, c) ->
			let cons = EqSet.filter eset (Cons.mk i c) in
			let (c', f) = (Cons.get_c cons, Cons.get_f cons) in
			match Cstr.tellProp c' with
			| Cstr.Contrad -> RewriteBot' f
			| Cstr.Trivial ->
				let bwd' = (i, Cert.cstrFromTriv i f)::bwd in
				Rewritten' (ilist, (fwd, bwd'))
			| Cstr.Nothing ->
				let fwd' = (i, Cert.Direct f)::fwd in
				let bwd' = (i, Cert.cstrInv i f)::bwd in
				Rewritten' ((i, c')::ilist, (fwd', bwd'))
	in
	fun eset ilist ->
		let rwZ = Rewritten' ([], ([], [])) in
		List.fold_left (rw1 eset) rwZ ilist

type 'a mayBotT
= Ok of 'a | Bot of Cert.t

let cextractc: int * Cstr.t -> int * Cert.cstr_t
= fun (i, _) -> i, Cert.Direct (Cert.mkF [i, Nb.u])

type ependingT
= EMin of EqSet.t | EAdd of (int * Cstr.t) list * ependingT

let rec einitc: ependingT -> ependingT
= function
	| EMin eset -> EMin (EqSet.initc eset)
	| EAdd (l, eset) -> EAdd (l, einitc eset)

let rec eextractc: ependingT -> Cert.partialT
= function
	| EMin eset -> EqSet.extractc eset
	| EAdd (l, eset) -> List.map cextractc l @ eextractc eset

type ipendingT
=	| IMin of IneqSet.t
	| IAdd of (int * Cstr.t) list * ipendingT
	| IAddRw of (int * Cstr.t) list * ipendingT
	| IRed of IneqSet.t
	| IRw of IneqSet.t

let rec needsRewrite: ipendingT -> ipendingT
= function
	| IMin iset -> IRw iset
	| IAdd (l, ip) -> IAddRw (l, needsRewrite ip)
	| IAddRw (l, ip) -> IAddRw (l, needsRewrite ip)
	| IRed iset -> IRw iset
	| IRw _ as ip -> ip

let rec iinitc: ipendingT -> ipendingT
= function
	| IAdd (l, iset) -> IAdd (l, iinitc iset)
	| IAddRw (l, iset) -> IAddRw (l, iinitc iset)
	| IRed iset -> IRed (IneqSet.initc iset)
	| IRw iset -> IRw (IneqSet.initc iset)
	| IMin iset -> IMin (IneqSet.initc iset)

let rec iextractc: ipendingT -> Cert.partialT
= function
	| IAdd (l, iset) | IAddRw (l, iset) -> List.map cextractc l @ iextractc iset
	| IRed iset -> IneqSet.extractc iset
	| IRw iset -> IneqSet.extractc iset
	| IMin iset -> IneqSet.extractc iset

type logT
= {nx: int; eset: ependingT; iset: ipendingT; log: Cert.equivT list}

let logEset: logT -> ependingT -> Cert.equivT -> logT
= fun lp eset (fwd, bwd) ->
	let sCe = iextractc lp.iset in
	{lp with
		eset = einitc eset;
		iset = needsRewrite lp.iset;
		log = (fwd @ sCe, bwd @ sCe)::lp.log
	}

let logIset: logT -> ipendingT -> Cert.equivT -> logT
= fun lp iset (fwd, bwd) ->
	let esCe = eextractc lp.eset in
	{lp with
		iset = iinitc iset;
		log = (fwd @ esCe, bwd @ esCe)::lp.log
	}

type meetT
= Added of t * (Cert.t * Cert.t) | Contrad of Cert.t

let meetPr: (Var.t -> string) -> meetT -> string
= fun varPr -> function
	| Contrad ce -> Printf.sprintf "Contrad ce, with ce:\n%s" (Cert.pr ce)
	| Added (p, (fwd, bwd)) -> Printf.sprintf
		"Added (p, (fwd, bwd)), with p:\n%s\nfwd:\n%s\nbwd:\n%s"
		(prExt varPr p) (Cert.pr fwd) (Cert.pr bwd)

let meetEq: meetT -> meetT -> bool
= fun m1 m2 ->
	match m1, m2 with
	| Contrad ce1, Contrad ce2 -> Cert.isEq ce1 ce2
	| Added (p1, (fwd1, bwd1)), Added (p2, (fwd2, bwd2)) ->
		isEqSyn p1 p2 && Cert.isEq fwd1 fwd2 && Cert.isEq bwd1 bwd2
	| Contrad _, Added (_, _)
	| Added (_, _), Contrad _ -> false

let mkPfBot: logT -> Cert.frag_t -> Cert.t
= fun p f ->
	let fwd = List.map Pervasives.fst p.log in
	certCombineFwd (List.rev fwd) (Cert.Empty f)

let logOut: logT mayBotT -> meetT
= function
	| Bot ce -> Contrad ce
	| Ok lp ->
		match lp.iset with
		| IRed _ | IRw _ | IAdd (_, _) | IAddRw (_, _) -> assert false
		| IMin iset ->
			match lp.eset with
			| EAdd (_, _) -> assert false
			| EMin eset ->
				match lp.log with
				| [] -> assert false
				| _::_ as log ->
					let p = {
						nxt = lp.nx;
						eqs = eset;
						ineqs = iset
					} in
					Added (p, certCombine (List.rev log))

let logAddEqs: logT -> (Cstr.t * Cert.cstr_t) list -> logT
= fun p l ->
	if List.length l = 0
	then p
	else
		let alloc = fun (nxt, l) (c, ce) -> (nxt + 1, (nxt, c, ce)::l) in
		let (nxt', l') = List.fold_left alloc (p.nx, []) l in
		let eqs = List.map (fun (i, c, _) -> (i, c)) l' in
		let eCe = List.map (fun (i, _, ce) -> (i, ce)) l' in
		let esCe = eextractc p.eset in
		logEset {p with nx = nxt'} (EAdd (eqs, p.eset)) (eCe @ esCe, esCe)

let logEqSetAddM: logT -> logT mayBotT
= fun lp1 ->
	let rec flatten: (int * Cstr.t) list -> ependingT -> logT mayBotT
	= fun l -> function
		| EAdd (l', lp) -> flatten (l @ l') lp
		| EMin eset as ep ->
			if List.length l = 0
			then Ok {lp1 with eset = ep}
			else
				match EqSet.addM eset l with
				| EqSet.Bot f -> Bot (mkPfBot lp1 f)
				| EqSet.Added (eset', ce) -> Ok (logEset lp1 (EMin eset') ce)
	in
	flatten [] lp1.eset

(* XXX: If there is nothing to rewrite, the log should not be changed. *)
let logRewriteIneqs: logT -> logT mayBotT
=	let idCert =
		fun l -> List.map (fun (i, c) -> (i, Cert.Direct (Cert.mkF [i, Nb.u]))) l
	in
	let rec rw: logT -> EqSet.t -> ipendingT -> (ipendingT * Cert.equivT) mayBotT
	= fun lp eset ip ->
		match ip with
		| IAdd (l, ip1) ->
			(match rw lp eset ip1 with
			| Bot _ as r -> r
			| Ok (ip2, (fwd, bwd)) ->
				let cel = idCert l in
				Ok (IAdd (l, ip2), (cel @ fwd, cel @ bwd)))
		| IAddRw (l, ip1) ->
			(match listRw eset l with
			| RewriteBot' f -> Bot (mkPfBot lp f)
			| Rewritten' (l', (fwdl, bwdl)) ->
				match rw lp eset ip1 with
				| Bot _ as r -> r
				| Ok (ip2, (fwd, bwd)) ->
					if List.length l' = 0
					then Ok (ip2, (fwdl @ fwd, bwdl @ bwd))
					else Ok (IAdd (l', ip2), (fwdl @ fwd, bwdl @ bwd)))
		| IMin iset | IRed iset ->
			let ce = IneqSet.extractc iset in
			Ok (ip, (ce, ce))
		| IRw iset ->
			match rewriteIneqs eset iset with
			| RewriteBot f -> Bot (mkPfBot lp f)
			| Rewritten (iset', ce) -> Ok (IRed iset', ce)
	in
	fun lp ->
		match lp.eset with
		| EAdd (_, _) -> assert false
	 	| EMin eset ->
			match rw lp eset lp.iset with
			| Bot ce -> Bot ce
			| Ok (ip, ce) -> Ok (logIset lp ip ce)

let logIneqRmRed: Var.t -> logT -> logT mayBotT
= fun nvar lp1 ->
	match lp1.iset with
	| IAdd (_, _) | IAddRw (_, _) | IRw _ -> assert false
	| IMin _ -> Ok lp1
	| IRed iset ->
		match IneqSet.rmRed nvar iset with
		| IneqSet.RedBot f -> Bot (mkPfBot lp1 f)
		| IneqSet.RedOk (iset', ce) -> Ok (logIset lp1 (IMin iset') ce)

let logIneqSetAddM: Var.t -> logT -> logT mayBotT
=	let doAdd: Var.t -> logT -> (int * Cstr.t) list -> IneqSet.t -> logT mayBotT
	= fun nvar lp l iset ->
		match IneqSet.addM nvar iset l with
		| IneqSet.Contrad f -> Bot (mkPfBot lp f)
		| IneqSet.Added (iset', ce) -> Ok (logIset lp (IMin iset') ce)
		| IneqSet.ImplicitEqs ((iset', elist), ce) ->
			let lp' = logIset lp (IMin iset') ce in
			Ok (logAddEqs lp' elist)
	in
	fun nvar lp ->
		let rec flatten: (int * Cstr.t) list -> ipendingT -> logT mayBotT
		= fun l -> function
			| IRw _
			| IAddRw (_, _) -> assert false
			| IAdd (l', ip) -> flatten (l @ l') ip
			| IRed iset as ip ->
				if List.length l = 0
				then logIneqRmRed nvar {lp with iset = ip}
				else doAdd nvar lp l iset
			| IMin iset as ip ->
				if List.length l = 0
				then Ok {lp with iset = ip}
				else doAdd nvar lp l iset
		in
		flatten [] lp.iset

let logBind: logT mayBotT -> (logT -> logT mayBotT) -> logT mayBotT
= fun mlp f ->
	match mlp with
	| Bot _ as b -> b
	| Ok lp -> f lp

let (<*>) = logBind

let logInit: t -> (int * Cstr.t) list -> logT mayBotT
=	let split: (int * Cstr.t) list -> (int * Cstr.t) list * (int * Cstr.t) list
	= fun l -> List.partition (fun (_, c) -> Cstr.get_typ c = Cstr.Eq) l
	in
	fun p l ->
		let (elist, ilist) = split l in
		Ok {
			nx = p.nxt;
			eset = EAdd (elist, EMin (EqSet.initc p.eqs));
			iset = IAddRw (ilist, IMin (IneqSet.initc p.ineqs));
			log = []
		}

(* XXX: it is a bit awkward that [p.nxt] is supposed to take into account
the identifiers used in [l]. *)
let addAux: Var.t -> t -> (int * Cstr.t) list -> meetT
= fun nvar p l -> logOut
	(logInit p l
	<*> (fun lp -> logEqSetAddM lp)
	<*> (fun lp -> logRewriteIneqs lp)
	<*> (fun lp -> logIneqSetAddM nvar lp)
	<*> (fun lp -> logEqSetAddM lp) (* handle implicit equalities *)
	<*> (fun lp -> logRewriteIneqs lp)
	<*> (fun lp -> logIneqRmRed nvar lp))

let addMSub: Var.t -> t -> Cstr.t list -> meetT
=	let alloc: int -> Cstr.t list -> int * (int * Cstr.t) list
	= fun nxt l -> List.fold_left (fun (i, l) c -> (i+1, (i, c)::l)) (nxt, []) l in
	fun nvar p l ->
		let (nxt', l') = alloc p.nxt l in
		addAux nvar {p with nxt = nxt'} l'

let addSub: Var.t -> t -> Cstr.t -> meetT
= fun nvar p c -> addMSub nvar p [c]

let meetSub: Var.t -> t -> t -> meetT
=	let shiftIds: int -> t -> t
	=	let shiftId: int -> Cons.t -> Cons.t
		= fun i c -> {c with Cons.id = Cons.get_id c + i}
		in
		fun i p ->
			let ineqs' = List.map (shiftId i) p.ineqs in
			let eqs' = List.map (fun (x, e) -> (x, shiftId i e)) p.eqs in
			{nxt = p.nxt + i;  ineqs = ineqs'; eqs = eqs'}
	in
	let cList: t -> (int * Cstr.t) list
	= fun p ->
		let consl = IneqSet.list p.ineqs @ EqSet.list p.eqs in
		List.map (fun c -> (Cons.get_id c, Cons.get_c c)) consl
	in
	fun nvar p1 p2 ->
		let p2' = shiftIds p1.nxt p2 in
		addAux nvar {p1 with nxt = p1.nxt + p2.nxt} (cList p2')

let top: t
= {
	eqs = EqSet.nil;
	ineqs = IneqSet.nil;
	nxt = 0
}

let mkSub: Var.t -> Cstr.t list -> t option
=	let add1: Var.t -> t option -> Cstr.t -> t option
	= fun nvar -> function
		| None -> fun _ -> None
		| Some p -> fun c ->
			match addSub nvar p c with
			| Added (p', _) -> Some p'
			| Contrad _ -> None
	in
	fun nvar l -> List.fold_left (add1 nvar) (Some top) l

let projectSub: Var.t -> t -> Var.t -> t * Cert.t
= fun nxtVar p x ->
	let (opte, i, eqs) = EqSet.trySubst x 0 (EqSet.initc p.eqs) in
	let (nxt, ineqs) =
		match opte with
		| Some e -> IneqSet.subst nxtVar eqs x i e (IneqSet.initc p.ineqs)
		| None -> IneqSet.fmElim nxtVar 1 eqs x i (IneqSet.initc p.ineqs)
	in
	let p1 = {
		ineqs = IneqSet.initc ineqs;
		eqs = EqSet.initc eqs;
		nxt = nxt }
	in
	let cert =
		Cert.Implies (List.append (EqSet.extractc eqs) (IneqSet.extractc ineqs))
	in
	(p1, cert)

let projectMSub: Var.t -> t -> Var.t list -> t * Cert.t
= fun nxtVar p l ->
	let msk =
		List.fold_left (fun m x -> Rtree.set None m x (Some x)) Rtree.Nil l
	in
	let rec project1 : int -> EqSet.t -> IneqSet.t -> t
	= fun nth eqs ineqs ->
		let (opte, i, eqs1) = EqSet.trySubstM 0 msk eqs in
		match opte with
		| Some (e, x) ->
			let (_, ineqs1) = IneqSet.subst nxtVar eqs1 x i e ineqs in
			project1 (nth + 1) eqs1 ineqs1

		| None ->
			let (i1, ineqs1) = IneqSet.fmElimM nxtVar nth eqs1 i msk ineqs in
			{ineqs = ineqs1; eqs = eqs1; nxt = i1}
	in
	let p1 = project1 1 (EqSet.initc p.eqs) (IneqSet.initc p.ineqs) in
	let cert = Cert.Implies (List.append (EqSet.extractc p1.eqs) (IneqSet.extractc p1.ineqs)) in
	let p2 = {
		ineqs = IneqSet.initc p1.ineqs;
		eqs = EqSet.initc p1.eqs;
		nxt = p1.nxt }
	in
	(p2, cert)

type discr_t = Op1 of int | Op2 of int | Other

(* note: x = y1 + y2 and alpha1 + alpha2 = 1 are substituted on the fly *)
let joinSetup: Var.t -> t -> t -> Var.t * (int -> discr_t) * t * Var.t list
= fun varNxt p1 p2 ->
	let idOff1 = 2 in
	let idOff2 = 2 + p1.nxt in
	let idEnd = 2 + p1.nxt + p2.nxt in
	let alpha = varNxt in
	let aIneqs =
		[Cons.mk 0 (Cstr.mk Cstr.Le [Nb.u, alpha] Nb.u);
		Cons.mk 1 (Cstr.mk Cstr.Le [Nb.negU, alpha] Nb.z)]
	in
	let (varNxt1, r, eqs1) = EqSet.joinSetup false (Var.next varNxt) Rtree.Nil alpha idOff1 p1.eqs in
	let (varNxt2, r, eqs2) = EqSet.joinSetup true varNxt1 r alpha idOff2 p2.eqs in
	let (varNxt3, r, ineqs1) = IneqSet.joinSetup false varNxt2 r alpha idOff1 p1.ineqs in
	let (varNxt4, r, ineqs2) = IneqSet.joinSetup true varNxt3 r alpha idOff2 p2.ineqs in
	let eqs = List.append eqs1 eqs2 in
	let ineqs = List.concat [ineqs1; ineqs2; aIneqs] in
	let discr n =
		if idOff1 <= n && n < idOff2 then
			Op1 idOff1
		else
			if idOff2 <= n && n < idEnd then
				Op2 idOff2
			else
				Other
	in
	let vars =
		Rtree.fold (fun a -> function None -> a | Some x -> x::a) [alpha] r
	in
	(varNxt4, discr, {ineqs = ineqs; eqs = eqs; nxt = idEnd}, vars)

let joinSub: Var.t -> t -> t -> t * Cert.joinT
= fun nxtVar p1 p2 ->
	let postProcess (p, cert) =
(* XXX: This whole function is an ugly hackish quickfix for an issue with strict inequality support.
See notClosed0 in the join test suite. *)
		let ltList: t -> int list
		= fun p ->
			let ltCons = List.filter (fun c -> Cstr.get_typ (Cons.get_c c) = Cstr.Lt) (IneqSet.list p.ineqs) in
			List.map Cons.get_id ltCons
		in
		let isLt: int list -> int -> bool
		= fun l i0 -> List.exists (fun i -> i = i0) l
		in
		let hasLt: int list -> Cert.cstr_t -> bool
		= fun l -> function
			| Cert.SplitEq _ -> failwith "Pol.join"
			| Cert.Direct f -> List.exists (fun (i, _) -> isLt l i) f
		in
		let ltP = ltList p in
		let ltP1 = ltList p1 in
		let ltP2 = ltList p2 in
		let elect j =
			if isLt ltP (Cert.getId j) then
				if hasLt ltP1 (Cert.getArg1 j) && hasLt ltP2 (Cert.getArg2 j) then
					false
				else
					true
			else
				false
		in
		let promote: t -> int list -> t
		= fun p les ->
			let promote1 c =
				if List.exists (fun i -> i = Cons.get_id c) les then
					{c with Cons.c = {(Cons.get_c c) with Cstr.typ = Cstr.Le }}
				else
					c
			in
			{p with ineqs = List.map promote1 p.ineqs}
		in
		(promote p (List.map Cert.getId (List.filter elect cert)), cert)
	in
	let (nxtVar1, discr, p0, vars) = joinSetup nxtVar p1 p2 in
	let split (i0, c) =
		let tell (f1, f2) (i, n) =
			match discr i with
			| Op1 orig -> ((i - orig, n)::f1, f2)
			| Op2 orig -> (f1, (i - orig, n)::f2)
			| Other -> (f1, f2)
		in
		match c with
		| Cert.SplitEq _ -> failwith "Pol.join"
		| Cert.Direct f ->
			let (f1, f2) = List.fold_left tell ([], []) f in
			(i0, Cert.Direct f1, Cert.Direct f2)
	in
	let (p, cert) = projectMSub nxtVar1 p0 vars in
	match cert with
	| Cert.Empty _ | Cert.Bind _ -> failwith "Pol.join"
	| Cert.Implies l ->
		let l0 = List.map split l in
		let l1 = List.map (fun (i, a1, a2) -> {Cert.id = i; Cert.arg1 = a1; Cert.arg2 = a2}) l0 in
		postProcess (p, l1)

type relVarT = OldSt of Var.t | NewSt of Var.t
type relLinT = (Nb.t * relVarT) list
type relCstrT = {cmpSign: Cstr.cmpT; linear: relLinT; constant: Nb.t}

let assignRelSub: Var.t -> t -> relCstrT list -> t option
=	let asgVars: Var.t -> relCstrT list -> (Var.t * Var.t) list * Var.t
	=	let asgVarsLin:
					(Var.t * Var.t) list * Var.t ->
					relCstrT ->
					(Var.t * Var.t) list * Var.t
		=	let asgVars1:
					(Var.t * Var.t) list * Var.t ->
					Nb.t * relVarT ->
					(Var.t * Var.t) list * Var.t
			= fun (alist, nvar) (_, v) ->
				match v with
				| OldSt _ -> (alist, nvar)
				| NewSt x ->
					if List.mem_assoc x alist
					then (alist, nvar)
					else ((x, nvar)::alist, Var.next nvar)
			in
			fun acc c -> List.fold_left asgVars1 acc c.linear
		in
		fun nvar l -> List.fold_left asgVarsLin ([], nvar) l
	in
	let rewrite: (Var.t * Var.t) list -> relCstrT list -> Cstr.t list
	=	let rewriteCstr: (Var.t * Var.t) list -> relCstrT -> Cstr.t
		=	let rewriteLin: (Var.t * Var.t) list -> relLinT -> (Nb.t * Var.t) list
			=	let rewrite1
				= fun alist (n, x) ->
					match x with
					| OldSt x' -> (n, x')
					| NewSt x' ->
						if List.mem_assoc x' alist
						then (n, List.assoc x' alist)
						else (n, x')
				in
				fun alist l -> List.map (rewrite1 alist) l
			in
			fun alist c -> Cstr.mk c.cmpSign (rewriteLin alist c.linear) c.constant
		in
		fun alist l -> List.map (rewriteCstr alist) l
	in
	let rename: (Var.t * Var.t) list -> t -> t
	=	let rename1: t -> Var.t * Var.t -> t
		= fun p (from_var, to_var) ->
			{p with
				eqs = EqSet.rename p.eqs from_var to_var;
				ineqs = IneqSet.rename p.ineqs from_var to_var}
		in
		fun alist p -> List.fold_left rename1 p alist
	in
	(* XXX: Normalization is done in intersection anyway, isn't there
	a more elegant way to gather accurate information on the assignment
	being performed? *)
	let norm: relCstrT list -> relCstrT list
	=	let normCstr: relCstrT -> relCstrT
		=	let cmp: relVarT -> relVarT -> int
			= fun x1 x2 ->
				match x1, x2 with
				| OldSt _, NewSt _ -> -1
				| NewSt _, OldSt _ -> 1
				| OldSt x1', OldSt x2'
				| NewSt x1', NewSt x2' -> Var.cmp x1' x2'
			in
			let rec rmDup: relLinT -> relLinT
			= function
				| (n, x)::((n', x')::l' as l) ->
					if cmp x x' = 0
					then
						let n1 = Nb.add n n' in
						if Nb.cmpz n1 = 0
						then rmDup l'
						else rmDup ((n1, x')::l')
					else
						if Nb.cmpz n = 0
						then rmDup ((n', x')::l')
						else (n, x)::(rmDup l)
				| (n, x)::nil as l ->
					if Nb.cmpz n = 0
					then nil
					else l
				| [] -> []
			in
			fun c ->
				let lin = List.sort (fun (_, x1) (_, x2) -> cmp x1 x2) c.linear in
				{c with linear = rmDup lin}
		in
		fun l -> List.map normCstr l
	in
	(* XXX: Should alist = [] really imply that assignRelSub == identity?
	There are lots of other ways to build a meaningless transformation,
	from an assignment point of view and we don't check for them. *)
	fun nvar p l ->
		let l' = norm l in
		let (alist, nvar') = asgVars nvar l' in
		if List.length alist = 0
		then Some p
		else
			match mkSub nvar' (rewrite alist l') with
			| None -> None
			| Some p1 ->
				match meetSub nvar' p p1 with
				| Contrad _ -> None
				| Added (p2, _) ->
					let (p3, _) = projectMSub nvar' p2 (List.map fst alist) in
					Some (rename (List.map (fun (a, b) -> (b, a)) alist) p3)
				
let assignSub: Var.t -> t -> assign_t -> t * Cert.asgT
=	let addM: Var.t -> t -> Cstr.t list -> t * (int * Cert.t) list
	= fun nxtV ->
		let add1: t * (int * Cert.t) list -> Cstr.t -> t * (int * Cert.t) list
		= fun (p, l) c ->
			match addSub nxtV p c with
			| Contrad _ -> failwith "Pol.assign: should not happen"
			| Added (p', (ce, _)) -> (p', (p.nxt, ce)::l)
		in
		fun p l -> List.fold_left add1 (p, []) l
	in
	let genCs: Var.t -> bool -> assign_t -> Cstr.t list
	= fun x' inv a ->
		match a.cst with
		| I i ->
			let lin = List.map (fun (n, x) -> (Nb.neg n, x)) a.lin in
			if inv
			then List.append
				(match get_up i with
				| Infty -> []
				| Open n -> [Cstr.lt ((Nb.u, x')::lin) n]
				| Closed n -> [Cstr.le ((Nb.u, x')::lin) n])

				(match get_low i with
				| Infty -> []
				| Open n -> [Cstr.lt ((Nb.negU, x')::a.lin) (Nb.neg n)]
				| Closed n -> [Cstr.le ((Nb.negU, x')::a.lin) (Nb.neg n)])

			else List.append
				(match get_up i with
				| Infty -> []
				| Open n -> [Cstr.lt ((Nb.u, a.var)::lin) n]
				| Closed n -> [Cstr.le ((Nb.u, a.var)::lin) n])

				(match get_low i with
				| Infty -> []
				| Open n -> [Cstr.lt ((Nb.negU, a.var)::a.lin) (Nb.neg n)]
				| Closed n -> [Cstr.le ((Nb.negU, a.var)::a.lin) (Nb.neg n)])
		| S n ->
			let lin = List.map (fun (n, x) -> (Nb.neg n, x)) a.lin in
			if inv
			then [Cstr.eq ((Nb.u, x')::lin) n]
			else [Cstr.eq ((Nb.u, a.var)::lin) n]
	in
	let hasX: Var.t -> (Nb.t * Var.t) list -> bool
	= fun x l -> List.exists (fun (n, y) -> Var.cmp x y = 0 && Nb.cmpz n <> 0) l
	in
	fun nx p asg ->
		if hasX asg.var asg.lin (* Is the transformation invertible? *)
		then
			let (p1, cert1) = addM (Var.next nx) p (genCs nx true asg) in
			let (p2, cert2) = projectSub (Var.next nx) p1 asg.var in
			let res = { p2 with
				eqs = EqSet.rename p2.eqs nx asg.var;
				ineqs = IneqSet.rename p2.ineqs nx asg.var }
			in
			(res, Cert.AsgInv (cert1, cert2))
		else
			let (p1, cert1) = projectSub nx p asg.var in
			(* XXX: resetting of p1.nxt is necessary for the Coq part
			to give the right index to c1 *)
			let (p2, cert2) = addM nx {p1 with nxt = p.nxt} (genCs nx false asg) in
			(p2, Cert.AsgNInv (cert1, cert2))

let assignMSub:  Var.t -> t -> assign_t list -> t
= fun nxtVar p0 assigns ->
	let vars = List.map (fun a -> a.var) assigns in
	let nlin a = List.map (fun (ax, x) -> (Nb.neg ax, x)) a.lin in
	let chk a =
		List.for_all (fun v1 ->
			List.for_all (fun (_, v2) -> v1 <> v2) a.lin) vars
	in
	if List.for_all chk assigns then
		let (p1, _) = projectMSub nxtVar p0 vars in
		let put p a =
			match a.cst with
			| I _ -> failwith "Pol.assignMSub: not implemented"
			| S n ->

			match addSub nxtVar p (Cstr.mk Cstr.Eq ((Nb.u, a.var)::nlin a) n) with
			| Contrad _ -> failwith "Pol.assignM"
			| Added (pN, _) -> pN
		in
		List.fold_left put p1 assigns
	else
		let put (r, p, nxtVar1) a =
			match a.cst with
			| I _ -> failwith "Pol.assignMSub: not implemented"
			| S n ->

			let x = nxtVar1 in
			let c = Cstr.mk Cstr.Eq ((Nb.u, x)::nlin a) n in
			match addSub nxtVar1 p c with
			| Contrad _ -> failwith "Pol.assignM"
			| Added (pN, _) -> ((x, a.var)::r, pN, Var.next nxtVar1)
		in
		let (r, p1, nxtVar1) = List.fold_left put ([], p0, nxtVar) assigns in
		let (p2, _) = projectMSub nxtVar1 p1 vars in
		let eRename s (x1, x2) = EqSet.rename s x1 x2 in
		let iRename s (x1, x2) = IneqSet.rename s x1 x2 in
		{p2 with
			eqs = List.fold_left eRename p2.eqs r;
			ineqs = List.fold_left iRename p2.ineqs r}

let widen p1 p2 =
	let elect s c =
		let c1 = EqSet.filter p1.eqs c in
		match Cstr.tellProp (Cons.get_c c1) with
		| Cstr.Trivial -> c::s (* XXX: c shouldn't get in, right? *)
		| Cstr.Nothing ->
			if List.exists (fun c2 -> Cons.isEq c2 c1) p1.ineqs then
				c::s
			else
				s

		| Cstr.Contrad -> failwith "Pol.widen"
	in
	{ p2 with ineqs = List.fold_left elect [] p2.ineqs }

let isInclSub: Var.t -> t -> t -> rel_t
= fun nxtVar p1 p2 ->
	let p1 = fragInit p1 in
	let p2 = fragInit p2 in
	match EqSet.isIncl p1.eqs p2.eqs with
	| EqSet.NoIncl -> NoIncl
	| EqSet.Incl certE ->

	match IneqSet.isIncl nxtVar p1.eqs p1.ineqs p2.ineqs with
	| IneqSet.NoIncl -> NoIncl
	| IneqSet.Incl certI -> Incl (Cert.Implies (List.append certI certE))

let mkplist p =
	let e = EqSet.list p.eqs in
	let i = IneqSet.list p.ineqs in
	List.map (fun c -> ((Cons.get_id c), (Cons.get_c c))) (List.append e i)

let opt2itv (f: Nb.t -> Nb.t): Splx.Opt.optT -> bndT * Cert.frag_t option
  = function
  | Splx.Opt.Finite (_, n, w) -> (Closed (f n), Some (Cert.mkF w))
  | Splx.Opt.Sup (_, n, w) -> (Open (f n), Some (Cert.mkF w))
  | Splx.Opt.Infty -> (Infty, None)

let getUpperBoundImpl : Vec.t -> Splx.t Splx.mayUnsatT -> bndT * Cert.frag_t option
  = fun v sx ->
  match Splx.Opt.max' sx v with
  | Splx.IsUnsat _ -> Pervasives.failwith "Pol.getUpperBoundSub"
  | Splx.IsOk u -> opt2itv (fun n -> n) u

let getUpperBoundSub : Var.t -> t -> Vec.t -> bndT * Cert.frag_t option
  = fun x p v -> mkplist p |> Splx.mk x |> getUpperBoundImpl v

let getLowerBoundImpl : Vec.t -> Splx.t Splx.mayUnsatT -> bndT * Cert.frag_t option
  = fun v sx ->
  match Splx.Opt.max' sx (Vec.neg v) with
  | Splx.IsUnsat _ -> Pervasives.failwith "Pol.getLowerBoundSub"
  | Splx.IsOk l -> opt2itv Nb.neg l

let getLowerBoundSub : Var.t -> t -> Vec.t -> bndT * Cert.frag_t option
  = fun x p v -> mkplist p |> Splx.mk x |> getLowerBoundImpl v

let itvizeSub: Var.t -> t -> Vec.t -> itvT * Cert.frag_t option * Cert.frag_t option
= fun nxtVar p v ->
	let sx = Splx.mk nxtVar (mkplist p) in
	let (up, upf) = getUpperBoundImpl v sx in
	let (low, lpf) = getLowerBoundImpl v sx in
	{low = low; up = up}, lpf, upf

let varList: t -> Var.t list
= fun p -> Cstr.listVars (List.map Pervasives.snd (mkplist p))

let add: t -> Cstr.t -> meetT
= fun p c ->
	let nxt = Var.horizon (List.append (varList p) (Cstr.listVars [c])) in
	addSub nxt p c

let addM: t -> Cstr.t list -> meetT
= fun p l ->
	let nxt = Var.horizon (List.append (varList p) (Cstr.listVars l)) in
	addMSub nxt p l

let meet: t -> t -> meetT
= fun p1 p2 ->
	let nxt = Var.horizon (List.append (varList p1) (varList p2)) in
	meetSub nxt p1 p2

let mk: Cstr.t list -> t option
= fun l -> mkSub (Var.horizon (Cstr.listVars l)) l

let project: t -> Var.t -> t * Cert.t
= fun p x -> projectSub (Var.horizon (x::(varList p))) p x

let projectM: t -> Var.t list -> t * Cert.t
= fun p l -> projectMSub (Var.horizon (List.append l (varList p))) p l

let join: t -> t -> t * Cert.joinT
= fun p1 p2 ->
	let nxt = Var.horizon (List.append (varList p1) (varList p2)) in
	joinSub nxt p1 p2

let assign: t -> assign_t -> t * Cert.asgT
= fun p asg ->
	let vl = asg.var::(List.map Pervasives.snd asg.lin) in
	let nxt = Var.horizon (List.append (varList p) vl) in
	assignSub nxt p asg

let assignM: t -> assign_t list -> t
= fun p0 assigns ->
	let varListA: assign_t -> Var.t list
	= fun a -> a.var::(List.map Pervasives.snd a.lin)
	in
	let nxt = Var.horizon (List.append (varList p0) (List.concat (List.map varListA assigns))) in
	assignMSub nxt p0 assigns

let isIncl: t -> t -> rel_t
= fun p1 p2 ->
	let nxt = Var.horizon (List.append (varList p1) (varList p2)) in
	isInclSub nxt p1 p2

let itvize: t -> Vec.t -> itvT * Cert.frag_t option * Cert.frag_t option
= fun p v ->
	let nxt = Var.horizon (List.append (Vec.listVars [v]) (varList p)) in
	itvizeSub nxt p v

let getUpperBound : t -> Vec.t -> bndT * Cert.frag_t option
= fun p v ->
	let nxt = Var.horizon (List.append (Vec.listVars [v]) (varList p)) in
	getUpperBoundSub nxt p v

let getLowerBound : t -> Vec.t -> bndT * Cert.frag_t option
= fun p v ->
	let nxt = Var.horizon (List.append (Vec.listVars [v]) (varList p)) in
	getLowerBoundSub nxt p v

let rename: Var.t -> Var.t -> t -> t
= fun x y p -> {p with
	eqs = EqSet.rename p.eqs x y;
	ineqs = IneqSet.rename p.ineqs x y}

(* DM: TODO-> Alexis: relire *)
let get_cstr p : Cstr.t list =
  (List.map (fun (_,c) -> Cons.get_c c) (get_eqs p)) @
  (List.map Cons.get_c (get_ineqs p))

let inter p1 p2 =
  mk ((get_cstr p1) @ (get_cstr p2))

let invChk: t -> bool * string
=	let chkNxt: t -> bool
	= fun p ->
		let isLtI: Cons.t -> bool = fun c -> Cons.get_id c < p.nxt in
		let isLtE: Var.t * Cons.t -> bool = fun (_, c) -> Cons.get_id c < p.nxt in
		List.for_all isLtI p.ineqs && List.for_all isLtE p.eqs
	in
	let chkId: t -> bool
	=	let rec hasNoDup: int list -> bool
		= function
		| [] | _::[] -> true
		| i1::i2::t -> i1 <> i2 && hasNoDup (i2::t)
		in
		fun p ->
			let idsI = List.map Cons.get_id p.ineqs in
			let idsE = List.map (fun (_, c) -> Cons.get_id c) p.eqs in
			hasNoDup (List.sort Pervasives.compare (List.append idsI idsE))
	in
	let chkTriv: t -> bool
	=	let chk1: Cons.t -> bool
		= fun c -> Cstr.tellProp (Cons.get_c c) = Cstr.Nothing
		in
		fun p ->
			List.for_all chk1 p.ineqs && List.for_all (fun (_, c) -> chk1 c) p.eqs
	in
	let chkDefU: t -> bool
	=	let chk1: Var.t * Cons.t -> bool
		= fun (x, c) ->
			let lin = Cstr.get_v (Cons.get_c c) in
			Nb.cmp Nb.u (Vec.get lin x) = 0
		in
		fun p -> List.for_all chk1 p.eqs
	in
	let chkDefE: t -> bool
	=	let chk1: Var.t * Cons.t -> Var.t list * bool -> Var.t list * bool
		= fun (x, c) (l, r) -> 
			let lin = Cstr.get_v (Cons.get_c c) in
			(x::l, r && List.for_all (fun x' -> Nb.cmpz (Vec.get lin x') = 0) l)
		in
		fun p -> snd (List.fold_right chk1 p.eqs ([], true))
	in
	let chkDefI: t -> bool
	=	let chk1: Var.t list -> Cons.t -> bool
		= fun l c ->
			let lin = Cstr.get_v (Cons.get_c c) in
			List.for_all (fun x -> Nb.cmpz (Vec.get lin x) = 0) l
		in
		fun p -> List.for_all (chk1 (List.map fst p.eqs)) p.ineqs
	in
	let chkBot: t -> bool
	=	let cs: t -> (int * Cstr.t) list
		= fun p ->
			List.map (fun c -> (Cons.get_id c, Cons.get_c c))
				(List.append p.ineqs (List.map snd p.eqs))
		in
		fun p ->
			let h = Var.horizon (varList p) in
			match Splx.checkFromAdd (Splx.mk h (cs p)) with
			| Splx.IsOk _ -> true
			| Splx.IsUnsat _ -> false
	in
	let chkRed: t -> bool
	=	let chkE: EqSet.t -> bool
		=	let add: EqSet.t option -> Var.t * Cons.t -> EqSet.t option
			= function
			| None -> fun _ -> None
			| Some es -> fun (_, c) ->
				match EqSet.addM es [Cons.get_id c, Cons.get_c c] with
				| EqSet.Added (es', _) -> Some es'
				| EqSet.Bot _ -> None
			in
			fun es ->
				match List.fold_left add (Some EqSet.nil) es with
				| None -> false
				| Some es' -> List.length es = List.length es'
		in
		let chkI: t -> bool
		= fun p ->
			let nxt = Var.horizon (varList p) in
			match IneqSet.simpl nxt p.eqs p.ineqs with
			| IneqSet.SimplBot _ -> false
			| IneqSet.SimplMin is -> List.length is = List.length p.ineqs
		in
		fun p -> chkI p && chkE p.eqs
	in
	fun p ->
		let tr: bool -> string -> bool * string
		= fun r m -> (r, if r then "" else m)
		in
		let log = [
			tr (chkNxt p) "nxt is not greater than all other identifiers";
			tr (chkId p) "there are duplicate identifiers";
			tr (chkTriv p) "some constraints are trivial or contradictory";
			tr (chkDefU p) "non-unit coefficient for a defined variable";
			tr (chkDefE p) "a defined variable appears in an equality";
			tr (chkDefI p) "a defined variable appears in an inequality";
			tr (chkBot p) "the polyhedron is empty";
			tr (chkRed p) "there are redundant constraints"
		] in
		(List.for_all (fun (r, _) -> r) log,
			let errs = List.map snd log in
			String.concat "\n" (List.filter (fun s -> String.length s > 0) errs))

