(** This is the main module of the VPL.
It defines what a polyhedron is and what operations can be done on it.
Not that there is no handling of the empty polyhedron (i.e. the bottom value):
a value of type [t] is assumed to be non-bottom and operators that may yield
an empty polyhedron have an [option] type as return value. *)

(** The type of an abstract value.
The internal representation splits the linear constraints
into a set of equalities and a set of inequalities *)
type t = {
	eqs: EqSet.t;
	ineqs: IneqSet.t;
	nxt: int }

val get_eqs : t -> EqSet.t
val get_ineqs : t -> IneqSet.t
val get_nxt : t -> int

(** The return type of the inclusion test [isIncl].
If inclusion holds, the certificate can be checked using function [check]. *)
type rel_t =
| NoIncl
| Incl of Cert.t

type bndT =
	| Infty
	| Open of Nb.t
	| Closed of Nb.t

type itvT = { low: bndT; up: bndT }

val get_low: itvT -> bndT
val get_up: itvT -> bndT

type cstT = S of Nb.t | I of itvT

(** The description of an assignment operation.
[var] is given the value [lin] + [cst]. *)
type assign_t = {
	var: Var.t;
	lin: (Nb.t * Var.t) list;
	cst: cstT }

val getVar: assign_t -> Var.t
val getLin: assign_t -> (Nb.t * Var.t) list
val getCst: assign_t -> cstT

(** The polyhedron with no constraints. *)
val top: t

(** [isEqSyn p1 p2] returns true if [p1] and [p2] are made of exactly the same constraints.
The comparison is sensitive to insertion order for equalities, but not for inequalities. *)
val isEqSyn: t -> t -> bool

(** [certCombineFwd [l1; l2; ...; ln] ce] transforms a certificate [ce] proving
something about polyhedron [pn] into a certificate [ce'] proving something about
polyhedron [p0], knowing that [p0] is included in [p1] by [l1], that [l2] proves
the inclusion of [p1] in [p2] and so on.
Note that certificate [ce] must not use the [Cert.Bind] construct. *)
val certCombineFwd: (int * Cert.cstr_t) list list -> Cert.t -> Cert.t

(** [mk l] builds the polyhedron representing the conjunction of constraints in [l].
If the resulting polyhedron is empty (i.e. the conjunction is unsatisfiable),
[None] is returned. *)
val mk: Cstr.t list -> t option
val mkSub: Var.t -> Cstr.t list -> t option

(** A value of type [meetT] represents the result of an intersection of two
sets of constraints [s1] and [s2]. Either this intersection is empty, or it is
some new set of constraints [s]. The constructor [Contrad] represents the former
possibility, where the certificate gives a proof of the intersection being empty.
A value [Added (s, (fwd, bwd))] states that [fwd] is the proof of the intersection
of [s1] and [s2] being included in [s], while [bwd] is a proof of [s] being
included in the intersection of [s1] and [s2]. *)
type meetT
= Added of t * (Cert.t * Cert.t) | Contrad of Cert.t

(** Pretty-printer for values of type meetPr. *)
val meetPr: (Var.t -> string) -> meetT -> string

(** Equality test for values of type meetT.
Certificate equality is tested using [Cert.isEq] and
polyhedron equality is test with [isEqSyn]. *)
val meetEq: meetT -> meetT -> bool

(** Intersect a polyhedron with a new constraint. *)
val add: t -> Cstr.t -> meetT
val addSub: Var.t -> t -> Cstr.t -> meetT

(** Intersect a polyhedron with a list of new constraints. *)
val addM: t -> Cstr.t list -> meetT
val addMSub: Var.t -> t -> Cstr.t list -> meetT

(** Compute the intersection of two polyhedra. *)
val meet: t -> t -> meetT
val meetSub: Var.t -> t -> t -> meetT

(** Forget all information about a given variable [v] in a polyhedron [p].
The certificate proves that the result includes the geometric projection of the [p] with respect to [v].
The certificate can be checked using function [check]. *)
val project: t -> Var.t -> t * Cert.t
val projectSub: Var.t -> t -> Var.t -> t * Cert.t

(** Forget all information about a set of variables.
Although the result is identical to projecting the variables one by one,
heuristic choices on the projection order make [projectM] more efficient. *)
val projectM: t -> Var.t list -> t * Cert.t
val projectMSub: Var.t -> t -> Var.t list -> t * Cert.t

(** [join p1 p2] computes the convex hull [p] of [p1] and [p2].
The certificate provides the necessary arguments to show inclusion of [p1] and [p2] in [p]. *)
val join: t -> t -> t * Cert.joinT
val joinSub: Var.t -> t -> t -> t * Cert.joinT

(** [assign p x v c = p' cert1 cert2] represents the effect of the assignment [x = v + c] on [p].
The meaning of [cert1] and [cert2] depend on the characteristics of the assignment as follows.
There are two ways to implement this operation, depending on whether the transformation
represented by the assignment is invertible (i.e. whether [x] appears on both sides of the [=]).
If the transformation is not invertible, [x] gets projected from [p] (yielding [cert1])
and the constraint [x = v + c] is added to the result (yielding [cert2]).
Otherwise, the constraint [x' = v + c] is added to [p] (yielding [cert1]),
where [x'] is a temporary fresh variable, then [x] is projected from the result (yielding [cert2])
and [x'] is renamed into [x]. *)
val assign: t -> assign_t -> t * Cert.asgT
val assignSub: Var.t -> t -> assign_t -> t * Cert.asgT

(** [assignM p l] represents the parallel assignements [l] on [p]. *)
val assignM: t -> assign_t list -> t
val assignMSub: Var.t -> t -> assign_t list -> t

type relVarT = OldSt of Var.t | NewSt of Var.t
type relLinT = (Nb.t * relVarT) list
type relCstrT = {cmpSign: Cstr.cmpT; linear: relLinT; constant: Nb.t}

val assignRelSub: Var.t -> t -> relCstrT list -> t option

(** Widening operator.
The result includes the two operands, although no certificate is created.
Note that [widen p1 p2] relies on [p1] being included in [p2]. *)
val widen: t -> t -> t

(** [isIncl p1 p2] checks the (geometrical) inclusion of [p1] in [p2]. *)
val isIncl: t -> t -> rel_t
val isInclSub: Var.t -> t -> t -> rel_t

(** [let (itv, lProof, uProof) = itvize p v] computes the bounds imposed on the linear form [v]
by the constraints of the polyhedron [p].
For each bound (upper and lower), an inclusion certificate is returned if it is finite,
allowing the prove that the returned interval is indeed correct.
[lProof] and [uProof] are the certificates for the lower and upper bounds, respectively. *)
val itvize: t -> Vec.t -> itvT * Cert.frag_t option * Cert.frag_t option
val itvizeSub: Var.t -> t -> Vec.t -> itvT * Cert.frag_t option * Cert.frag_t option

(** [getUpperBound p v] computes the upper bound on the linear form [v]
implied by polyhedron [p].  If the bound is finite, a certificate is
provided. *)
val getUpperBound : t -> Vec.t -> bndT * Cert.frag_t option
val getUpperBoundSub : Var.t -> t -> Vec.t -> bndT * Cert.frag_t option

(** [getLowerBound p v] computes the lower bound on the linear form [v]
implied by polyhedron [p].  If the bound is finite, a certificate is
provided. *)
val getLowerBound : t -> Vec.t -> bndT * Cert.frag_t option
val getLowerBoundSub : Var.t -> t -> Vec.t -> bndT * Cert.frag_t option

(** [rename x y p] renames variable [x] to variable [y] in polyhedron [p].
[y] must be fresh in [p]. *)
val rename: Var.t -> Var.t -> t -> t

(** This function checks that the given polyhedron satisfies the following
invariants. It returns [(true, "")] if it does and [(false, reason)] if not,
where [reason] lists the checks that have failed. 
- not constraint has an index greater than [nxt]
- no two constraints have the same identifier
- there is no trivial constraint
- the polyhedron is not empty
- there is no redundant constraint
- invariants on the set of equalities {ul
 {li a defined variable appears with unit coefficient}
 {li a defined variable does not appear in the equalities before it}}
- invariants on the set of inequalities {ul
 {li none of the variables defined by an equality appears in an inequality}} *)
val invChk: t -> bool * string

(** Pretty-printer for a polyhedron. *)
val pr: (Var.t -> string) -> t -> string
val prExt: (Var.t -> string) -> t -> string
val prItv: (Var.t -> string) -> Vec.t -> itvT -> string
val prRaw: t -> string
val prExtRaw: t -> string
val prItvRaw: Vec.t -> itvT -> string


(**/**)
type discr_t

val joinSetup: Var.t -> t -> t -> Var.t * (int -> discr_t) * t * Var.t list

(* XXX: get the list of variables bound by a polyhedron,
with duplicates. *)
val varList : t -> Var.t list

(* DM extension *)
val get_cstr : t -> Cstr.t list
val inter : t -> t -> t option
