module
Val =
	struct
	type t = { v: Nb.t; d: Nb.t }
	let get_d (x : t) = x.d
        let get_v (x : t) = x.v

	let ofNb n = { v = n; d = Nb.z }

	let pdelta n = { v = n; d = Nb.u }

	let ndelta n = { v = n; d = Nb.negU }

	let z = { v = Nb.z; d = Nb.z }

	let cmp v1 v2 =
		match Nb.cmp v1.v v2.v with
		| 0 -> Nb.cmp v1.d v2.d
		| n -> n

	let hasDelta v = (Nb.cmpz v.d <> 0)

	let muln v n = {
		v = Nb.mult v.v n;
		d = Nb.mult v.d n
	}

	let divn v n = {
		v = Nb.div v.v n;
		d = Nb.div v.d n
	}

	let add v1 v2 = {
		v = Nb.add v1.v v2.v;
		d = Nb.add v1.d v2.d
	}

	let sub v1 v2 = {
		v = Nb.sub v1.v v2.v;
		d = Nb.sub v1.d v2.d
	}

	let adddelta v = { v with d = Nb.add v.d Nb.u }

	let subdelta v = { v with d = Nb.sub v.d Nb.u }

	let pr v = Printf.sprintf "%s + %s.delta" (Nb.pr v.v) (Nb.pr v.d)

	end

type bnd_t = { id: int; scale: Nb.t; bv: Val.t }
let get_id x = x.id
let get_scale x = x.scale
let get_bv x = x.bv

type state_t = { v: Val.t; low: bnd_t option; up: bnd_t option }
let get_low x = x.low
let get_up x = x.up
let get_v x = x.v

type bck_t =
| Ok
| KoL of Val.t
| KoU of Val.t

type witness_t = (int * Nb.t) list

type whatnext_t = NonBasic of Var.t | UnsatSplx of witness_t

type 'a mayUnsatT
= IsOk of 'a | IsUnsat of witness_t

type t = {
	mat: (Vec.t option) Rtree.t;
	state: state_t Rtree.t;
	nxt: Var.t
}
let get_mat x = x.mat
let get_state x = x.state
let get_nxt x = x.nxt

let prWitness w = Cert.prF w

let prBnd = function
	| None -> "None"
	| Some b -> Printf.sprintf "Some {id = %i; scale = %s; bv = %s}" b.id (Nb.pr b.scale) (Val.pr b.bv)

let prSt st = Printf.sprintf "{v = %s; low = %s; up = %s}" (Val.pr st.v) (prBnd st.low) (prBnd st.up)

let prState: (Var.t -> string) -> state_t Rtree.t -> string
= fun varPr state ->
	let pr st x = Printf.sprintf "%s: %s" x (prSt st) in
	Rtree.pr "\n" pr varPr state

let prMat: (Var.t -> string) -> Vec.t option Rtree.t -> string
= fun varPr mat ->
	let pr c x =
		match c with
		| None -> ""
		| Some v -> Printf.sprintf "%s: %s = 0" x (Vec.pr varPr v)
	in
	Rtree.pr "\n" pr varPr mat

let pr: (Var.t -> string) -> t -> string
= fun varPr s -> Printf.sprintf "matrix:\n%s\nstate:\n%s\n" (prMat varPr s.mat) (prState varPr s.state)

let bndMax : bnd_t -> bnd_t -> bnd_t
= fun b b' -> if Val.cmp b.bv b'.bv < 0 then b' else b

let bndMin : bnd_t -> bnd_t -> bnd_t
= fun b b' -> if Val.cmp b.bv b'.bv > 0 then b' else b

let mergeB : state_t -> state_t -> state_t
= fun st1 st2 ->
	let l =
		match st1.low, st2.low with
		| l, None | None, l -> l
		| Some l1, Some l2 -> Some (bndMax l1 l2)
	in
	let u =
		match st1.up, st2.up with
		| u, None | None, u -> u
		| Some u1, Some u2 -> Some (bndMin u1 u2)
	in
	{st1 with low = l; up = u}

let setVal ve st =
	let rec acc a = function
		| Rtree.Nil, _ | _, Rtree.Nil -> a
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, st2, r2) ->
			let a1 = Val.sub a (Val.muln st2.v n1) in
			acc (acc a1 (l1, l2)) (r1, r2)
	in
	acc Val.z (ve, st)

type fitBndT
= FitBnd of Val.t * state_t | BndUnsat of witness_t

let fitBnd : state_t -> fitBndT
= fun st ->
	match st.low, st.up with
	| None, None -> FitBnd (Val.z, st)
	| Some l, None ->
		if Val.cmp st.v l.bv < 0
		then FitBnd (Val.sub l.bv st.v, {st with v = l.bv})
		else FitBnd (Val.z, st)
	| None, Some u ->
		if Val.cmp u.bv st.v < 0
		then FitBnd (Val.sub u.bv st.v, {st with v = u.bv})
		else FitBnd (Val.z, st)
	| Some l, Some u ->
		if Val.cmp l.bv u.bv > 0
		then BndUnsat [l.id, l.scale; u.id, u.scale]
		else
			if Val.cmp st.v l.bv < 0
			then FitBnd (Val.sub l.bv st.v, {st with v = l.bv})
			else
				if Val.cmp u.bv st.v < 0
				then FitBnd (Val.sub u.bv st.v, {st with v = u.bv})
				else FitBnd (Val.z, st)

let rec incrupdate d xN m st =
	match m, st with
	| Rtree.Nil, rt -> rt
	| Rtree.Sub _, Rtree.Nil -> failwith "Splx.incrupdate"

	| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
		let l = incrupdate d xN l1 l2 in
		let r = incrupdate d xN r1 r2 in
		let n =
			match n1 with
			| None -> n2
			| Some cons ->
				let n = Vec.get cons xN in
				{n2 with v = Val.sub n2.v (Val.muln d n)}
		in
		Rtree.Sub (l, n, r)

let mkBnd id (c: Cstr.t) (a: Nb.t): state_t =
	let b: Nb.t = Nb.div (Cstr.get_c c) a in
	let aZ = Nb.cmpz a in
	if aZ = 0 then
		invalid_arg "Splx.mkBnd"
	else
		match Cstr.get_typ c with
		| Cstr.Eq ->
			let l = {id = id; scale = Nb.inv (Nb.neg a); bv = Val.ofNb b} in
			let u = {id = id; scale = Nb.inv a; bv = Val.ofNb b} in
			{v = Val.z; low = Some l; up = Some u}
		| Cstr.Le ->
			if aZ < 0 then
				let u = {id = id; scale = Nb.inv a; bv = Val.ofNb b} in
				{v = Val.z; low = None; up = Some u}
			else
				let l = {id = id; scale = Nb.inv (Nb.neg a); bv = Val.ofNb b} in
				{v = Val.z; low = Some l; up = None}
		| Cstr.Lt ->
			if aZ < 0 then
				let u = {id = id; scale = Nb.inv a; bv = Val.ndelta b} in
				{v = Val.z; low = None; up = Some u}
			else
				let l = {id = id; scale = Nb.inv (Nb.neg a); bv = Val.pdelta b} in
				{v = Val.z; low = Some l; up = None}

let add : t -> int * Cstr.t -> t mayUnsatT
=	let boundVar : t -> int -> Cstr.t -> Var.t -> t mayUnsatT
	= fun sx id c x ->
		let a = Vec.get (Cstr.get_v c) x in
		let oBnd = Rtree.get {v = Val.z; low = None; up = None} sx.state x in
		match fitBnd (mergeB oBnd (mkBnd id c a)) with
		| BndUnsat w -> IsUnsat w
		| FitBnd (delta, bnd') ->
			let st = Rtree.set {v = Val.z; low = None; up = None} sx.state x bnd' in
			IsOk {sx with state = incrupdate delta x sx.mat st}
	in
	let constrain : t -> int -> Cstr.t -> t
	= fun s id c ->
		let nCons = Vec.set (Vec.neg (Cstr.get_v c)) s.nxt Nb.u in
		let nVState = {(mkBnd id c Nb.u) with v = setVal nCons s.state} in
		let nMat = Rtree.set None s.mat s.nxt (Some nCons) in
		let nState: state_t Rtree.t
		=	let stNil: state_t = {v = Val.z; low = None; up = None} in
			let rec fill: state_t Rtree.t -> Vec.t -> state_t Rtree.t
			= fun st v ->
				match v, st with
				| Rtree.Nil, (_ as st') -> st'
				| Rtree.Sub (l, _, r), Rtree.Nil ->
					Rtree.Sub (fill Rtree.Nil l, stNil, fill Rtree.Nil r)
				| Rtree.Sub (lv, _, rv), Rtree.Sub (ls, ns, rs) ->
					Rtree.Sub (fill ls lv, ns, fill rs rv)
			in
			fill (Rtree.set stNil s.state s.nxt nVState) (Cstr.get_v c)
		in
		{mat = nMat; state = nState; nxt = Var.next s.nxt}
	in
	let handleTriv : int -> Cstr.t -> t -> t mayUnsatT
	= fun i c sx ->
		match Cstr.tellProp c with
		| Cstr.Trivial -> IsOk sx
		| Cstr.Contrad -> IsUnsat [i, Nb.u]
		| Cstr.Nothing -> Pervasives.failwith "Splx.add"
	in
	fun s (id, c) ->
		match Cstr.listVars [c] with
		| [] -> handleTriv id c s
		| x :: [] -> boundVar s id c x
		| _ -> IsOk (constrain s id c)

let addAcc : t mayUnsatT -> int * Cstr.t -> t mayUnsatT
= function
	| IsOk sx -> add sx
	| IsUnsat _ as r -> fun _ -> r

let mk: Var.t -> (int * Cstr.t) list -> t mayUnsatT
= fun lim l ->
	let empty = {mat = Rtree.Nil; state = Rtree.Nil; nxt = lim} in
	List.fold_left (fun s c -> addAcc s c) (IsOk empty) l

let update m0 st0 =
	let updateval ve0 st0 n =
		let rec _upv a ve st =
			match ve, st with
			| Rtree.Nil, _ | _, Rtree.Nil -> a
			| Rtree.Sub (l1, _, r1), Rtree.Sub (l2, (None, _), r2) ->
				_upv (_upv a l1 l2) r1 r2

			| Rtree.Sub (l1, n, r1), Rtree.Sub (l2, (Some d, _), r2) ->
				_upv (_upv (Val.add a (Val.muln d n)) l1 l2) r1 r2
		in
		{ n with v = _upv n.v ve0 st0 }
	in
	let rec _up m st =
		match m, st with
		| _, Rtree.Nil -> Rtree.Nil
		| Rtree.Nil, Rtree.Sub (l, (_, n), r) ->
			Rtree.Sub (_up Rtree.Nil l, n, _up Rtree.Nil r)

		| Rtree.Sub (l1, None, r1), Rtree.Sub (l2, (_, n), r2) ->
			Rtree.Sub (_up l1 l2, n, _up r1 r2)

		| Rtree.Sub (l1, Some ve, r1), Rtree.Sub (l2, (_, n), r2) ->
			Rtree.Sub (_up l1 l2, updateval ve st0 n, _up r1 r2)
	in
	_up m0 st0

let pivot m xB xN =
	let xNCons =
		let xBCons =
			match Rtree.get None m xB with
			| None -> invalid_arg "Splx.pivot"
			| Some cons -> cons
		in
		Vec.div (Vec.get xBCons xN) xBCons
	in
	let elim = function
		| None -> None
		| Some cons -> Some (Vec.elim xN xNCons cons)
	in
	let m1 = Rtree.map elim (Rtree.set None m xB None) in
	Rtree.set None m1 xN (Some xNCons)

let _in_bound (b: state_t): bck_t =
	match b.low with
	| Some l when Val.cmp l.bv b.v > 0 -> KoL l.bv
	| Some _ | None ->
		match b.up with
		| Some u when Val.cmp b.v u.bv > 0 -> KoU u.bv
		| Some _ | None -> Ok

type chgDirT = HasToIncr | HasToDecr

type choiceT
= {var : Var.t; vec : Vec.t; st : state_t}

let fromLeft : choiceT option -> choiceT option
= function None -> None | Some c -> Some {c with var = Var.XO c.var}

let fromRight : choiceT option -> choiceT option
= function None -> None | Some c -> Some {c with var = Var.XI c.var}

module type Strategy
= sig

	val pickBasic : t -> choiceT option

	val pickNBasic : t -> chgDirT -> Vec.t -> whatnext_t

end

module Bland : Strategy
= struct

	let pickBasic : t -> choiceT option
	= fun s ->
		let elect : Vec.t option -> state_t -> choiceT option
		= fun mc st ->
			match mc with
			| None -> None
			| Some c ->
				match _in_bound st with
				| Ok -> None
				| KoL _
				| KoU _ -> Some {var = Var.XH; vec = c; st = st}
		in
		let rec find : Vec.t option Rtree.t -> state_t Rtree.t -> choiceT option
		= fun m stree ->
			match m, stree with
			| Rtree.Nil, _
			| _, Rtree.Nil -> None
			| Rtree.Sub (l1, mc, r1), Rtree.Sub (l2, st, r2) ->
				match elect mc st with
				| Some _ as r -> r
				| None ->
					match find l1 l2 with
					| Some _ as r -> fromLeft r
					| None -> fromRight (find r1 r2)
		in
		find s.mat s.state

	let pickNBasic s d xiv =
		let tryincr w a st =
			match st.up with
			| None -> NonBasic Var.XH
			| Some u when Val.cmp u.bv st.v > 0 -> NonBasic Var.XH
			| Some u -> UnsatSplx ((u.id, Nb.mult u.scale a)::w)
		in
		let trydecr w a st =
			match st.low with
			| None -> NonBasic Var.XH
			| Some l when Val.cmp l.bv st.v < 0 -> NonBasic Var.XH
			| Some l -> UnsatSplx ((l.id, Nb.mult l.scale a)::w)
		in
		let (fpos, fneg) =
			match d with
			| HasToDecr -> (tryincr, trydecr)
			| HasToIncr -> (trydecr, tryincr)
		in
		let f w a st =
			match Nb.cmpz a with
			| 0 -> UnsatSplx w
			| n when n < 0 -> fpos w a st
			| _ -> fneg w (Nb.neg a) st
		in
		let rec find a0 ve st =
			match ve, st with
			| Rtree.Nil, _ -> UnsatSplx a0
			| Rtree.Sub _, Rtree.Nil -> NonBasic Var.XH
			| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
				match f a0 n1 n2 with
				| NonBasic _ as r -> r
				| UnsatSplx a1 ->
					match find a1 l1 l2 with
					| NonBasic v -> NonBasic (Var.XO v)
					| UnsatSplx a2 ->
						match find a2 r1 r2 with
						| NonBasic v -> NonBasic (Var.XI v)
						| UnsatSplx _ as r -> r
		in
		find [] xiv s.state

end

module Steep : Strategy
= struct

	type indepChoiceT
	= Bounded of Var.t * Val.t | Unbounded of Var.t

	let pickBasic : t -> choiceT option
	= fun s ->
		let candidate : Vec.t option -> state_t -> choiceT option
		= function
			| None -> fun _ -> None
			| Some ve -> fun st ->
				match _in_bound st with
				| Ok -> None
				| KoL _ | KoU _ -> Some {var = Var.XH; vec = ve; st = st}
		in
		let score: choiceT -> Val.t
		= fun c ->
			match _in_bound c.st with
			| KoL l -> Val.sub l c.st.v
			| KoU u -> Val.sub c.st.v u
			| Ok -> Pervasives.failwith "Splx.pickbasic_steep.score"
		in
		let choose : choiceT option -> choiceT option -> choiceT option
		= function
			| None -> fun oc -> oc
			| Some c as oc -> function
				| None -> oc
				| Some c' as oc'-> if Val.cmp (score c) (score c') < 0 then oc' else oc
		in
		let rec find: Vec.t option Rtree.t -> state_t Rtree.t -> choiceT option
		= fun m st ->
			match m, st with
			| Rtree.Nil, _ | _, Rtree.Nil -> None
			| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
				let ocl = fromLeft (find l1 l2) in
				let ocr = fromRight (find r1 r2) in
				choose (choose ocl ocr) (candidate n1 n2)
		in
		find s.mat s.state

	let pickNBasic : t -> chgDirT -> Vec.t -> whatnext_t
	=	let choose : indepChoiceT option -> indepChoiceT option -> indepChoiceT option
		= function
			| None -> fun c -> c
			| Some (Unbounded _) as oc -> fun _ -> oc
			| Some (Bounded (_, sc)) as oc -> function
				| None -> oc
				| Some (Unbounded _) as oc' -> oc'
				| Some (Bounded (_, sc')) as oc' ->
					if Val.cmp sc sc' < 0 then oc' else oc
		in
		let tryDecr : state_t -> indepChoiceT option
		= fun st ->
			match st.low with
			| None -> Some (Unbounded Var.XH)
			| Some l ->
				if Val.cmp l.bv st.v < 0
				then Some (Bounded (Var.XH, Val.sub st.v l.bv)) (* XXX: coefficient *)
				else None
		in
		let tryIncr : state_t -> indepChoiceT option
		= fun st ->
			match st.up with
			| None -> Some (Unbounded Var.XH)
			| Some u ->
				if Val.cmp st.v u.bv < 0
				then Some (Bounded (Var.XH, Val.sub u.bv st.v)) (* XXX: coefficient *)
				else None
		in
		let candidate : chgDirT -> Nb.t -> state_t -> indepChoiceT option
		= function
			| HasToIncr -> fun a st ->
				let asign = Nb.cmpz a in
				if asign = 0 then None
				else
					if asign < 0 then tryDecr st
					else tryIncr st
			| HasToDecr -> fun a st ->
				let asign = Nb.cmpz a in
				if asign = 0 then None
				else
					if asign < 0 then tryIncr st
					else tryDecr st
		in
		let fromLeft : indepChoiceT option -> indepChoiceT option
		= function
			| None -> None
			| Some (Unbounded x) -> Some (Unbounded (Var.XO x))
			| Some (Bounded (x, sc)) -> Some (Bounded (Var.XO x, sc))
		in
		let fromRight : indepChoiceT option -> indepChoiceT option
		= function
			| None -> None
			| Some (Unbounded x) -> Some (Unbounded (Var.XI x))
			| Some (Bounded (x, sc)) -> Some (Bounded (Var.XI x, sc))
		in
		let rec witness : chgDirT -> Vec.t -> state_t Rtree.t -> witness_t
		=	let bound : Nb.t -> chgDirT -> state_t -> bnd_t option
			= fun a ->
				let asign = Nb.cmpz a in
				if asign = 0 then fun _ _ -> None
				else if asign < 0 then function
					| HasToIncr -> fun s -> s.low
					| HasToDecr -> fun s -> s.up
				else function
					| HasToIncr -> fun s -> s.up
					| HasToDecr -> fun s -> s.low
			in
			fun d v st ->
				match v, st with
				| Rtree.Nil, _ -> []
				| _, Rtree.Nil -> Pervasives.invalid_arg "Splx.picknbasic_steep"
				| Rtree.Sub (l1, a, r1), Rtree.Sub (l2, s, r2) ->
					let wl : witness_t = witness d l1 l2 in
					let wr : witness_t = witness d r1 r2 in
					match bound a d s with
					| None -> wl @ wr
					| Some b -> (b.id, Nb.mult b.scale (Nb.abs a)) :: (wl @ wr)
		in
		let rec find : chgDirT -> Vec.t -> state_t Rtree.t -> indepChoiceT option
		= fun d v st ->
			match v, st with
			| Rtree.Nil, _ -> None
			| Rtree.Sub _, Rtree.Nil -> Some (Unbounded Var.XH)
			| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
				let l = fromLeft (find d l1 l2) in
				let r = fromRight (find d r1 r2) in
				choose (choose l r) (candidate d n1 n2)
		in
		fun s d v ->
			match find d v s.state with
			| None -> UnsatSplx (witness d v s.state)
			| Some (Unbounded x) -> NonBasic x
			| Some (Bounded (x, _)) -> NonBasic x

end

type strategyT
= Bland | Steep

let pickBasic : strategyT -> t -> choiceT option
= function
	| Bland -> Bland.pickBasic
	| Steep -> Steep.pickBasic

let pickNBasic : strategyT -> t -> chgDirT -> Vec.t -> whatnext_t
= function
	| Bland -> Bland.pickNBasic
	| Steep -> Steep.pickNBasic

type stepT
= StepCont of t | StepSat of t | StepUnsat of witness_t

let step : strategyT -> t -> stepT
= fun strgy sx ->
	match pickBasic strgy sx with
	| None -> StepSat sx
	| Some ch ->
		let (whatNext, xival) =
			match _in_bound ch.st with
			| Ok -> Pervasives.failwith "step"
			| KoL l -> pickNBasic strgy sx HasToIncr ch.vec, l
			| KoU u -> pickNBasic strgy sx HasToDecr ch.vec, u
		in
		match whatNext with
		| UnsatSplx w -> StepUnsat w
		| NonBasic xj ->
			let xi = ch.var in
			let m = pivot sx.mat xi xj in
			let xiSt = Rtree.get {v = Val.z; low = None; up = None} sx.state xi in
			let st = Rtree.set {v=Val.z; low = None; up = None} sx.state xi {xiSt with v = xival} in
			StepCont {sx with mat = m; state = incrupdate (Val.sub xival xiSt.v) xi m st}

let presimpl s0 =
	let rec find ve st =
		let walk l1 l2 r1 r2 =
			match find l1 l2 with
			| Some (a, v) -> Some (a, Var.XO v)
			| None ->
				match find r1 r2 with
				| None -> None
				| Some (a, v) -> Some (a, Var.XI v)
		in
		match ve, st with
		| Rtree.Nil, _ -> None
		| Rtree.Sub (l, n, r), Rtree.Nil ->
			if Nb.cmpz n = 0 then
				walk l Rtree.Nil r Rtree.Nil
			else
				Some (n, Var.XH)
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			if Nb.cmpz n1 <> 0 && n2.low = None && n2.up = None then
				Some (n1, Var.XH)
			else
				walk l1 l2 r1 r2
	in
	let (_, newm) =
		let rec simpl fl m =
			let recurse n fl l r =
				let (fll, nl) = simpl fl l in
				let (flr, nr) = simpl fll r in
				(flr, Rtree.Sub (nl, n, nr))
			in
			match m with
			| Rtree.Nil -> (fl, Rtree.Nil)
			| Rtree.Sub (l, None, r) -> recurse None fl l r
			| Rtree.Sub (l, Some ve, r) ->
				let nve = List.fold_right (fun f v -> f v) fl ve in
				match find nve s0.state with
				| None -> recurse (Some nve) fl l r
				| Some (a, va) ->
					recurse None ((Vec.elim va (Vec.div a nve))::fl) l r
		in
		simpl [] s0.mat
	in
	let nst = (* basic variables will be set by update *)
		Rtree.map (fun n ->
			match _in_bound n with
			| Ok -> (None, n)
			| KoL v | KoU v ->
				(Some (Val.sub n.v v), { n with v = v })) s0.state
	in
	{ s0 with mat = newm; state = update newm nst }

let check_loop : t -> t mayUnsatT
=	let chooseStrategy : int -> strategyT -> strategyT
	= fun cnt -> function
		| Bland -> Bland
		| Steep -> if cnt < 20 then Steep else Bland
	in
	let rec loop : int -> strategyT -> t -> t mayUnsatT
	= fun cnt strgy sx ->
		let strgy' = chooseStrategy cnt strgy in
		match step strgy' sx with
		| StepCont sx' -> loop (cnt+1) strgy' sx'
		| StepSat sx' -> IsOk sx'
		| StepUnsat w -> IsUnsat w
	in
	fun sx -> loop 1 Steep sx

let check : t -> t mayUnsatT
= fun sx -> check_loop (presimpl sx)

let checkFromAdd : t mayUnsatT -> t mayUnsatT
= function
	| IsOk sx -> check sx
	| IsUnsat _ as u -> u

let compl s id =
	let f s =
		match s.low, s.up with
		| Some l, _ when l.id = id ->
			let u1 = { l with bv = Val.subdelta l.bv } in
			Some (Val.sub u1.bv s.v, { v = u1.bv; low = None; up = Some u1 })

		| _, Some u when u.id = id ->
			let l1 = { u with bv = Val.adddelta u.bv } in
			Some (Val.sub l1.bv s.v, { v = l1.bv; low = Some l1; up = None })

		| _, _ -> None
	in
	match Rtree.find f s.state with
	| None -> invalid_arg "Splx.compl"
	| Some (v, (delta, b)) ->
		let st1 = Rtree.set {v = Val.z; low = None; up = None} s.state v b in
		{s with state = incrupdate delta v s.mat st1}

let stricten : t -> int -> t mayUnsatT
=	let predUp : int -> state_t -> state_t option
	= fun id st ->
		match st.up with
		| None -> None
		| Some u -> if u.id = id && not (Val.hasDelta u.bv)
			then Some st
			else None
	in
	let pred : int -> state_t -> state_t option
	= fun id st ->
		match st.low with
		| None -> predUp id st
		| Some l -> if l.id = id && not (Val.hasDelta l.bv)
			then Some st
			else predUp id st
	in
	let tryStrictenLow : int -> bnd_t option -> bnd_t option
	= fun id -> function
		| None -> None
		| Some b as mb -> if b.id = id
			then Some {b with bv = Val.adddelta b.bv}
			else mb
	in
	let tryStrictenUp : int -> bnd_t option -> bnd_t option
	= fun id -> function
		| None -> None
		| Some b as mb -> if b.id = id
			then Some {b with bv = Val.subdelta b.bv}
			else mb
	in
	fun sx id ->
		match Rtree.find (pred id) sx.state with
		| None -> Pervasives.invalid_arg "Splx.stricten"
		| Some (x, st) ->
			let st' = {st with
				low = tryStrictenLow id st.low;
				up = tryStrictenUp id st.up
			} in
			match fitBnd st' with
			| BndUnsat w -> IsUnsat w
			| FitBnd (delta, st') ->
				let sTree = Rtree.set {v = Val.z; low = None; up = None}
					sx.state x st'
				in
				IsOk {sx with state = incrupdate delta x sx.mat sTree}

let forget s id =
	let f bnd =
		match bnd.low, bnd.up with
		| Some b, None | None, Some b when b.id = id ->
			Some (true, { bnd with low = None; up = None })

		| Some l, Some _ when l.id = id ->
			Some (false, { bnd with low = None })

		| Some _, Some u when u.id = id ->
			Some (false, { bnd with up = None })

		| _, _ -> None
	in
	match Rtree.find f s.state with
	| None -> invalid_arg "Splx.forget"
	| Some (v, (p, b)) ->
		let s1 = {s with state = Rtree.set {v = Val.z; low = None; up = None} s.state v b} in
		if p then
			presimpl s1 (* sledge hammer *)
		else
			s1

module Opt = struct
	type progressT = Unbnd | UpTo of Val.t | NoChange

	type dirT = Incr | Decr

	type actionT = Done | Move of (Var.t * dirT * progressT)

	type nextT = OptFinite of Val.t | OptUnbnd | GoOn of t

	type optT
	  = Finite of (t * Nb.t * witness_t)
	  | Sup of (t * Nb.t * witness_t)
	  | Infty

	let prProgress = function
		| Unbnd -> "Unbnd"
		| UpTo v -> Printf.sprintf "UpTo %s" (Val.pr v)
		| NoChange -> "NoChange"
	
	let prDir = function
		| Incr -> "Incr"
		| Decr -> "Decr"

	let prAction: (Var.t -> string) -> actionT -> string
	= fun varPr -> function
		| Done -> "Done"
		| Move (v, d, bnd) -> Printf.sprintf "Move (%s, %s, %s)" (varPr v) (prDir d) (prProgress bnd)

	let prNext: (Var.t -> string) -> nextT -> string
	= fun varPr -> function
		| GoOn s1 -> Printf.sprintf "GoOn\n%s\n" (pr varPr s1)
		| OptUnbnd -> "OptUnbnd"
		| OptFinite v1 -> Printf.sprintf "OptFinite %s" (Val.pr v1)

	let prOpt = function
		| Finite (_, n, w) -> Printf.sprintf "Finite %s, cert: %s" (Nb.pr n) (prWitness w)
		| Sup (_, n, w) -> Printf.sprintf "Sup %s, cert: %s" (Nb.pr n) (prWitness w)
		| Infty -> "Infty"

	let tryDecr st =
		match st.low with
		| None -> Unbnd
		| Some l when Val.cmp l.bv st.v < 0 -> UpTo (Val.sub st.v l.bv)
		| Some _ -> NoChange

	let tryIncr st =
		match st.up with
		| None -> Unbnd
		| Some u when Val.cmp u.bv st.v > 0 -> UpTo (Val.sub u.bv st.v)
		| Some _ -> NoChange

	let pickNBasic z s =
		let rec find vec state =
			let walk l1 l2 r1 r2 =
				match find l1 l2 with
				| Move (x, dir, bnd) -> Move (Var.XO x, dir, bnd)
				| Done ->
					match find r1 r2 with
					| Move (x, dir, bnd) -> Move (Var.XI x, dir, bnd)
					| Done -> Done
			in
			match vec, state with
			| Rtree.Nil, _ -> Done
			| Rtree.Sub (l, n, r), Rtree.Nil ->
				let nSign = Nb.cmpz n in
				if nSign = 0 then
					walk l Rtree.Nil r Rtree.Nil
				else
					Move (Var.XH, (if nSign < 0 then Decr else Incr), Unbnd)
			| Rtree.Sub (l1, n, r1), Rtree.Sub (l2, st, r2) ->
				let nSign = Nb.cmpz n in
				if nSign = 0 then
					walk l1 l2 r1 r2
				else
					let (dir, tryFn) = if nSign < 0 then (Decr, tryDecr) else (Incr, tryIncr) in
					match tryFn st with
					| Unbnd | UpTo _ as b -> Move (Var.XH, dir, b)
					| NoChange -> walk l1 l2 r1 r2
		in
		match Rtree.get None s.mat z with
		| None -> invalid_arg "Splx.Opt.pickNBasic"
		| Some vec -> find (Vec.set vec z Nb.z) s.state

	let pickBasic s xN xNBnd dir =
		let choose x1 bnd1 x2 bnd2 =
			match bnd1, bnd2 with
			| NoChange, NoChange | Unbnd, Unbnd -> (x1, bnd1)
			| NoChange, Unbnd | NoChange, UpTo _ | UpTo _, Unbnd -> (x1, bnd1)
			| Unbnd, UpTo _ | Unbnd, NoChange | UpTo _, NoChange -> (x2, bnd2)
			| UpTo d1, UpTo d2 ->
				if Val.cmp d1 d2 <= 0 then
					(x1, bnd1)
				else
					(x2, bnd2)
		in
		let (tryPos, tryNeg) =
			match dir with
			| Incr -> (tryDecr, tryIncr)
			| Decr -> (tryIncr, tryDecr)
		in
		let rec find m st =
			match m, st with
			| Rtree.Nil, _ -> None
			| _, Rtree.Nil -> Some (Var.XH, Unbnd)
			| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
				let nN =
					match n1 with
					| None -> Unbnd
					| Some cons ->
						let aN = Vec.get cons xN in
						match Nb.cmpz aN with
						| 0 -> Unbnd
						| aSign ->
							match (if aSign < 0 then tryPos else tryNeg) n2 with
							| Unbnd | NoChange as bnd -> bnd
							| UpTo d -> UpTo (Val.divn d (Nb.abs aN))
				in
				let maybeL = find l1 l2 in
				let maybeR = find r1 r2 in
				match maybeL, maybeR with
				| None, None -> Some (Var.XH, nN)
				| Some (xL, nL), None -> Some (choose (Var.XO xL) nL Var.XH nN)
				| None, Some (xR, nR) -> Some (choose (Var.XI xR) nR Var.XH nN)
				| Some (xL, nL), Some (xR, nR) ->
					let (xS, nS) = choose (Var.XO xL) nL (Var.XI xR) nR in
					Some (choose xS nS Var.XH nN)
		in
		match find s.mat s.state with
		| None -> (xN, xNBnd)
		| Some (x, xBnd) -> choose xN xNBnd x xBnd

	let step z s =
		match pickNBasic z s with
		| Done ->
			let zSt = Rtree.get {v = Val.z; low = None; up = None} s.state z in
			OptFinite zSt.v
		| Move (xN, dir, b) ->
			let (xB, xNBnd) = pickBasic s xN b dir in
			match xNBnd with
			| Unbnd -> OptUnbnd
			| NoChange -> GoOn {s with mat = pivot s.mat xB xN}
			| UpTo delta ->
				let xNSt = Rtree.get {v = Val.z; low = None; up = None} s.state xN in
				let nSt =
					let upVal = match dir with Incr -> Val.add | Decr -> Val.sub in
					Rtree.set {v = Val.z; low = None; up = None} s.state xN {xNSt with v = upVal xNSt.v delta}
				in
				let m =
					if xB = xN then
						s.mat
					else
						pivot s.mat xB xN
				in
				let d = function Incr -> delta | Decr -> Val.sub Val.z delta in
				GoOn {s with mat = m; state = incrupdate (d dir) xN s.mat nSt}

	let setObj s obj =
		let cons = Vec.set (Vec.neg obj) s.nxt Nb.u in
		let zSt = {v = setVal cons s.state; low = None; up = None} in
		let nMat = Rtree.set None s.mat s.nxt (Some cons) in
		let nState = Rtree.set {v = Val.z; low = None; up = None} s.state s.nxt zSt in
		(s.nxt, {mat = nMat; state = nState; nxt = Var.next s.nxt})

	let mkCert z objCons state0 =
		let rec build cert vec state =
			match vec, state with
			| Rtree.Nil, _ -> cert
			| Rtree.Sub (_, _, _), Rtree.Nil -> invalid_arg "Splx.Opt.mkCert"
			| Rtree.Sub (l1, n, r1), Rtree.Sub (l2, st, r2) ->
				let cert1 =
					(* XXX: what happens when fiddling with the sign of the coefficient when equations sneak in? *)
					let nSign = Nb.cmpz n in
					if nSign = 0 then
						cert
					else if nSign < 0 then
						match st.low with
						| None -> invalid_arg "Splx.Opt.mkCert"
						| Some l -> (l.id, Nb.mult l.scale n)::cert
					else
						match st.up with
						| None -> invalid_arg "Splx.Opt.mkCert"
						| Some u -> (u.id, Nb.mult u.scale (Nb.neg n))::cert
				in
				let cert2 = build cert1 l1 l2 in
				build cert2 r1 r2
		in
		build [] (Vec.set objCons z Nb.z) state0

	let max : t -> Vec.t -> optT mayUnsatT
	= fun s0 obj ->
		let (z, s1) = setObj s0 obj in
		let rec optimize s =
			match step z s with
			| GoOn s1 -> optimize s1
			| OptUnbnd -> Infty
			| OptFinite o ->
				let objCons =
					match Rtree.get None s.mat z with
					| None -> failwith "Splx.Opt.max"
					| Some c -> c
				in
				match Nb.cmpz (Val.get_d o) with
				| 0 -> Finite (s, Val.get_v o, mkCert z objCons s.state)
				| n when n > 0 -> Sup (s, Val.get_v o, mkCert z objCons s.state)
				| _ -> failwith "Splx.Opt.max"
		in
		match check_loop s1 with
		| IsOk sx -> IsOk (optimize sx)
		| IsUnsat _ as u -> u

	let max' : t mayUnsatT -> Vec.t -> optT mayUnsatT
	= function
		| IsUnsat _ as u -> fun _ -> u
		| IsOk sx -> fun obj -> max sx obj

end
