(** This module implements a feasibility-only simplex algorithm.
The basis for the algorithm comes from the article "Integrating simplex with DPLL(T)"
published by Bruno Dutertre and Leonardo de Moura in 2006 as
an SRI International technical report.
The interface of this module needs a fair amount of user-friendliness. *)

module Val:
sig
	type t

	val get_v: t -> Nb.t
	val ofNb: Nb.t -> t
	val pdelta: Nb.t -> t
	val ndelta: Nb.t -> t
	val z: t
	val cmp: t -> t -> int
	val hasDelta: t -> bool
	val muln: t -> Nb.t -> t
	val divn: t -> Nb.t -> t
	val add: t -> t -> t
	val sub: t -> t -> t
	val adddelta: t -> t
	val subdelta: t -> t
	val pr: t -> string
end

(**
data type for the description of bound

The [id] field records which input constraint the bound is the result of.
The [scale] field is used to generate correct certificates in case the
input constraint was of the form [2.x <= 2].
The [bv] field is the actual value of the bound of the variable.
*)
type bnd_t
= {id: int; scale: Nb.t; bv: Val.t}

(** accessor for the [id] field of a value of type [bnd_t] *)
val get_id : bnd_t -> int

(** accessor for the [scale] field of a value of type [bnd_t] *)
val get_scale : bnd_t -> Nb.t

(** accessor for the [bv] field of a value of type [bnd_t] *)
val get_bv : bnd_t -> Val.t

type state_t = { v: Val.t; low: bnd_t option; up: bnd_t option }
val get_v : state_t -> Val.t
val get_low : state_t -> bnd_t option
val get_up : state_t -> bnd_t option

type t = {
	mat: (Vec.t option) Rtree.t;
	state: state_t Rtree.t;
	nxt: Var.t }
val get_mat : t ->  (Vec.t option) Rtree.t
val get_state: t -> state_t Rtree.t
val get_nxt: t -> Var.t

(*	The type of a contradiction witness.  This is the list of the non-zero
	coefficients to apply to the linear constraints of the instance, identified
	by an index number starting from zero and incremented for each newly added
	contraint.
	It describes a linear combination of the constraints of the problem yielding a
	trivially contradictory constraint.
*)
type witness_t = (int * Nb.t) list

type 'a mayUnsatT
= IsOk of 'a | IsUnsat of witness_t

(**
[add s (id, c)] adds contraints [c], identified by [id],
to the simplex problem [s].

[add] assumes, but does not check, that [id] uniquely identifies [c].
If the addition of [c] to the simplex problem makes it unsatisfiable,
[add] may or may not return [IsUnsat w], where [w] describes a linear
combination of the input constraints yielding a contradiction,
otherwise, the updated problem is returned.

In the current implementation, [IsUnsat] is only returned when
specifying incompatible bounds on single variables, e.g. [x < 1] and
[x > 1].
*)
val add: t -> int * Cstr.t -> t mayUnsatT

(**
[addAcc] is a convenience wrapper around [add] which propagates
[IsUnsat] or performs the addition using [add] as necessary.
*)
val addAcc : t mayUnsatT -> int * Cstr.t -> t mayUnsatT

(**
[mk lim l] builds a new simplex problem, composed of the constraints in [l].

Variable [lim] specifies a frame beyond which fresh variables can be
allocated, using [Var.next]. Note that [lim] itself may be used as a
fresh variable. There is no possibility to change the frame after a
simplex data structure has been created using [mk]: all the constraints
ever added to the result of [mk] should have their variables "smaller"
than [lim].

[mk] is implemented in terms of [add] and has the same behaviour
regarding [IsUnsat] return values.
*)
val mk: Var.t -> (int * Cstr.t) list -> t mayUnsatT

(**
look for an assignment of the variables of the problem which satisfies
all the constraints

If such an assignment is found, the last simplex tableau is returned.
If no such assignment exist, an unsatifiability witness is returned
which describes a linear combination of the input constraints yielding
a contradiction.

Due to a simplification step which occurs before the actual search
of an assignment, it is not possible to extract a complete assignment
of values to the variables of the problem. For the same reason, adding
new constraints to a tableau returned by [check] and calling [check]
again will not do anything useful. However, constraint manipulation
using [compl], [stricten] and [forget], followed by a new [check], will
work as expected.
*)
val check : t -> t mayUnsatT

(**
convenience wrapper around [check] which propagate unsatifiability
or checks satisfiability as required
*)
val checkFromAdd : t mayUnsatT -> t mayUnsatT

(*	Complement the bound with given integer identifier.
	If multiple bounds have this identifier, one is arbitrarily chosen.
	Assumes the current point satisfies all constraints and bounds.
*)
val compl: t -> int -> t

(*	Make the bound with given integer identifier strict. *)
val stricten: t -> int -> t mayUnsatT

(*	Remove the bound with the given integer identifier.
	If this is the only bound on the corresponding variable, project said
	variable from the constraints.
*)
val forget: t -> int -> t

(** Pretty-print an element of type [t].
The function [pr] wraps functions [prMat] and [prState]. *)
val pr: (Var.t -> string) -> t -> string

(** [prMat s] pretty-prints the constraint matrix (field [mat]) of [s]. *)
val prMat: (Var.t -> string) -> Vec.t option Rtree.t -> string

(** [prState s] pretty-prints the bounds and values on variables (field [state]) of [s]. *)
val prState: (Var.t -> string) -> state_t Rtree.t -> string

(** Pretty-print an element of type [bnd_t option]. *)
val prBnd: bnd_t option -> string

(** Pretty-print an element of type [state_t]. *)
val prSt: state_t -> string

(** Pretty-print a value of type [witness_t]. *)
val prWitness: witness_t -> string

(**/**)
(*	The following types and functions are private, hence the leading ``_''.
	They are used for diagnostic purposes.
*)

type bck_t =
| Ok
| KoL of Val.t
| KoU of Val.t

type whatnext_t = NonBasic of Var.t | UnsatSplx of witness_t

(** [mergeB st1 st2] computes the intersection of the bounds described by [st1] and [st2].
The [v] field of the result is equal to [st1.v].
If the interval resulting from the intersection is empty,
exception [Unsat w] is raised with [w] describing the offending bounds. *)

val mergeB: state_t -> state_t -> state_t

(** [setVal ve st] computes the amount by which the value of one of the variables involved in [ve]
(assuming it has a unit coefficient) should be increased so that the state [st] satisfies the constraint [ve]. *)
val setVal: Vec.t -> state_t Rtree.t -> Val.t

(** [fitBnd st = (d, st')] adjusts the value of the variable which state is described by [st], so that it fits its bounds.
It is assumed (but not enforced), that there is at least one possible value which fits the bounds.
The [fitBnd] function returns the new state [st'] with updated value and [d] = [st'.v - st.v]. *)

type fitBndT
= FitBnd of Val.t * state_t | BndUnsat of witness_t

val fitBnd: state_t -> fitBndT

(** [incrupdate v x m st] adjusts the value of the basic variables in [st] so that
the constraints [m] are still satisfied after a change [d] on the value of the non-basic variable [x] (new value is [x] + [d]). *)
val incrupdate: Val.t -> Var.t -> (Vec.t option) Rtree.t -> state_t Rtree.t -> state_t Rtree.t

(** [pivot m xB xN] transforms that contraint matrix [m] so that
the basic variable [xB] becomes non-basic and
the non-basic variable [xN] becomes basic.
The handling of variable values to preserve the invariant is left to the caller. *)
val pivot: Vec.t option Rtree.t -> Var.t -> Var.t -> Vec.t option Rtree.t

val _in_bound: state_t -> bck_t

type choiceT
= {var : Var.t; vec : Vec.t; st : state_t}

type strategyT
= Bland | Steep

type chgDirT
= HasToIncr | HasToDecr

module type Strategy
= sig

	val pickBasic : t -> choiceT option

	val pickNBasic : t -> chgDirT -> Vec.t -> whatnext_t

end

module Bland : Strategy
module Steep : Strategy

type stepT
= StepCont of t | StepSat of t | StepUnsat of witness_t

val step : strategyT -> t -> stepT
val presimpl: t -> t

(**/**)
module Opt: sig
(** This module builds on {! Splx} to offer an optimizing simplex. *)

(** A value of type [progressT] describes the change that can be operated on a given quantity.
The change does not specify whether this is an increase or a decrease.
In the case of [UpTo d], the change [d] is always positive. *)
	type progressT = Unbnd | UpTo of Val.t | NoChange

(** Describe the direction of a value change: increasing or decreasing. *)
	type dirT = Incr | Decr

(** During the optimization, a value of type [actionT] decribes the choice on the course of actions,
based on the contents of the objective line.
[Done] means that the optimal value for the objective function is reached.
[Move (v, d, p)] means that adjusting the value of the variable [v], in direction [d],
will increase the value of the objective function.
In the latter case, the bounds on [v] limit its change according to [p]. *)
	type actionT = Done | Move of (Var.t * dirT * progressT)

(** At the end of each pivoting step, decide whether the optimum is reached or
progress still needs to be made. *)
	type nextT = OptFinite of Val.t | OptUnbnd | GoOn of t

(** This is the result type of the optimization. *)
	type optT =
	| Finite of (t * Nb.t * witness_t) (** The optimized form can take values up to and including the given value. *)
	| Sup of (t * Nb.t * witness_t) (** The optimized form can take values up to and excluding the given value. *)
	| Infty (** The optimized form is unbounded. *)

	(**
	maximize a linear form under a set of constraints

	[max] returns either a description of the optimum for the given
	linear form or an unsatifiability witness if the constraint system
	is not satisfiable.

	The simplex tableau may be built using functions [add] or [mk], but
	[check] must not be called on it prior to calling [max] as [check]
	performs simplifications possible for satifiability checking, but not
	for optimization.
	*)
	val max: t -> Vec.t -> optT mayUnsatT

	(**
	convenience wrapper around [max] for direct use on the result of
	functions [add] or [mk]
	*)
	val max' : t mayUnsatT -> Vec.t -> optT mayUnsatT

(** Pretty-print a value of type [progressT]. *)
	val prProgress: progressT -> string

(** Pretty-print a value of type [dirT]. *)
	val prDir: dirT -> string

(** Pretty-print a value of type [actionT].
The provided basis must be the one in which the non-basic variable is defined. *)
	val prAction: (Var.t -> string) -> actionT -> string

(** Pretty-print a value of type [nextT]. *)
	val prNext: (Var.t -> string) -> nextT -> string

(** Pretty-print a value of type [optT]. *)
	val prOpt: optT -> string

(** The following function are internal functions which are exported for testing purposes only.
They maintain the invariant on the simplex structure,
which is that the value of all the variables should satisfy both their bounds and the constraints. *)

(** [pickNBasic z s] observes the objective line of the simplex tableau,
represented by the contraint of basic variable [z]
and decides what should be done. *)
	val pickNBasic: Var.t -> t -> actionT

(** [pickBasic s xN xNBnd dir] looks for the basic variable with which to pivot [xN].
The selected basic variable [xB] is the one which constrains the most the change of the value of [xN] in direction [dir].
The constraint on this change, [xNBnd1] is returned along with [xB].
The [xNBnd] argument is the constraint set on the change of the value of [xN] by the bounds of [xN].
If [xNBnd] is the most restrictive of all of the constraints imposed on the value of [xN], then [xB] = [xN]. *)
	val pickBasic: t -> Var.t -> progressT -> dirT -> Var.t * progressT

(** [step z s] does one optimization step:
it chooses a non-basic variable which can be used to increase the value of the goal,
it chooses a basic variable with which to pivot and finally pivots.
It notices that the optimum is reached only if it is reached in [s]. *)
	val step: Var.t -> t -> nextT

(** [setObj s obj] sets [obj] as the linear form to optimize in [s] and
returns the corresponding problem,
along with the basic variable associated with [obj]. *)
	val setObj: t -> Vec.t -> Var.t * t

end
