let coqPosToZ: BinNums.positive -> Nb.Z.t
= fun p0 -> CoqPr.posTr p0

let zToCoqPos: Nb.Z.t -> BinNums.positive
= fun z0 ->
	if Nb.Z.cmp Nb.Z.z z0 >= 0 then
		invalid_arg "Support.zToCoqPos"
	else
		let rec f (z: Nb.Z.t): BinNums.positive =
			if Nb.Z.cmp z Nb.Z.u = 0 then
				BinNums.Coq_xH
			else
				if Nb.Z.cmp (Nb.Z.lAnd z Nb.Z.u) Nb.Z.u = 0 then
					BinNums.Coq_xI (f (Nb.Z.shiftR z 1))
				else
					BinNums.Coq_xO (f (Nb.Z.shiftR z 1))
		in
		f z0

let coqZToZ: BinNums.coq_Z -> Nb.Z.t
= fun z -> CoqPr.zTr z

let zToCoqZ: Nb.Z.t -> BinNums.coq_Z
= fun z ->
	let sign = Nb.Z.cmp Nb.Z.z z in
	if sign = 0 then
		BinNums.Z0
	else
		if sign < 0 then
			BinNums.Zpos (zToCoqPos z)
		else
			BinNums.Zneg (zToCoqPos (Nb.Z.neg z))

let nToNb: NumC.QNum.t -> Nb.t
= fun q ->
	let num = coqZToZ (QArith_base.coq_Qnum q) in
	let den = coqPosToZ (QArith_base.coq_Qden q) in
	Nb.ofZ num den

let nToNumC: Nb.t -> NumC.QNum.t
= fun n ->
	let (num, den) = Nb.toZ n in
	let q = {
		QArith_base.coq_Qnum = zToCoqZ num;
		QArith_base.coq_Qden = zToCoqPos den }
	in
	Qcanon.coq_Q2Qc q

let ltNorm: LinTerm.LinQ.exportT -> LinTerm.LinQ.exportT
= fun lt ->
	let rec rmZeroes: LinTerm.LinQ.exportT -> LinTerm.LinQ.exportT
	= function
		| [] -> []
		| (x, n)::lt1 ->
			let lt2 = rmZeroes lt1 in
			if NumC.QNum.eqDec NumC.QNum.z n
				then lt2
				else (x, n)::lt2
	in
	let rec rmDup: LinTerm.LinQ.exportT -> LinTerm.LinQ.exportT
	= function (* rmDup assumes that the input is sorted. *)
		| [] -> []
		| (x1, n1)::lt1 ->
			match lt1 with
			| [] -> [x1, n1]
			| (x2, n2)::lt2 ->
				if ProgVar.PVar.isEq x1 x2
					then rmDup ((x1, NumC.QNum.add n1 n2)::lt2)
					else (x1, n1)::(rmDup lt1)
	in
	let pairCmp: ProgVar.PVar.t * NumC.QNum.t -> ProgVar.PVar.t * NumC.QNum.t -> int
	= fun (x1, _) (x2, _) ->
		if ProgVar.PVar.isLt x1 x2
			then -1
			else if ProgVar.PVar.isEq x1 x2
				then 0
				else 1
	in
	rmZeroes (rmDup (List.sort pairCmp lt))

let progVarToVar: ProgVar.PVar.t -> Var.t
= fun x ->
	let rec posToVar: BinNums.positive -> Var.t
	= function
		| BinNums.Coq_xH -> Var.XH
		| BinNums.Coq_xO p -> Var.XO (posToVar p)
		| BinNums.Coq_xI p -> Var.XI (posToVar p)
	in
	posToVar (ProgVar.PVar.export x)

let varToProgVar: Var.t -> ProgVar.PVar.t
= fun x ->
	let rec varToPos: Var.t -> BinNums.positive
	= function
		| Var.XH -> BinNums.Coq_xH
		| Var.XI p -> BinNums.Coq_xI (varToPos p)
		| Var.XO p -> BinNums.Coq_xO (varToPos p)
	in
	ProgVar.PVar.import (varToPos x)

let ltToVec: LinTerm.LinQ.t -> Vec.t
= fun lt ->
	let convert: ProgVar.PVar.t * NumC.QNum.t -> Nb.t * Var.t
	= fun (x, n) -> (nToNb n, progVarToVar x)
	in
	Vec.mk (List.map convert (ltNorm (LinTerm.LinQ.export lt)))

let vecToLt: Vec.t -> LinTerm.LinQ.t
= fun v ->
	let rmZeroes: (Nb.t * Var.t) list -> (Nb.t * Var.t) list
	= fun l -> List.filter (fun (n, _) -> Nb.cmpz n <> 0) l
	in
	let convert: Nb.t * Var.t -> ProgVar.PVar.t * NumC.QNum.t 
	= fun (n, x) -> (varToProgVar x, nToNumC n)
	in
	LinTerm.LinQ.import (List.map convert (rmZeroes (Rtree.toList v)))

let cToCmpT: NumC.cmpT -> Cstr.cmpT
= function 
	| NumC.EqT -> Cstr.Eq
	| NumC.LeT -> Cstr.Le
	| NumC.LtT -> Cstr.Lt

let cToCmp: Cstr.cmpT -> NumC.cmpT
= function
	| Cstr.Eq -> NumC.EqT
	| Cstr.Le -> NumC.LeT
	| Cstr.Lt -> NumC.LtT

let cToCstr: CstrC.Cstr.t -> Cstr.t
= fun c -> {
	Cstr.v = ltToVec (CstrC.Cstr.coefs c);
	Cstr.typ = cToCmpT (CstrC.Cstr.typ c);
	Cstr.c = nToNb (CstrC.Cstr.cst c)
}

let cToCstrC: Cstr.t -> CstrC.Cstr.t
= fun c -> {
	CstrC.Cstr.typ = cToCmp (Cstr.get_typ c);
	CstrC.Cstr.coefs = vecToLt (Cstr.get_v c);
	CstrC.Cstr.cst = nToNumC (Cstr.get_c c)
}

let cstrCPr: CstrC.Cstr.t -> string
= fun c -> CoqPr.charListTr (CstrC.Cstr.pr c)

let cToConsC: Cons.t -> ConsSet.Cs.cT
= fun c -> {ConsSet.Cs.id = Cons.get_id c; ConsSet.Cs.cstr = cToCstrC (Cons.get_c c)}

let pToConsSet: Pol.t -> ConsSet.Cs.t
= fun p ->
	let e = EqSet.list (Pol.get_eqs p) in
	let i = IneqSet.list (Pol.get_ineqs p) in
	List.map cToConsC (List.append e i)

(** This function is meant to normalize an object of type [ConsSet.Cs.t] that is,
two objects which would be equal up to constraint coefficients order and
constraint order become syntactically equal. However, no renumbering of constraints
take place. *)
let csNorm: ConsSet.Cs.t -> ConsSet.Cs.t
=	let sort: ConsSet.Cs.t -> ConsSet.Cs.t
	=	let cmp: ConsSet.Cs.cT -> ConsSet.Cs.cT -> int
		= fun c1 c2 -> Pervasives.compare (ConsSet.Cs.id c1) (ConsSet.Cs.id c2)
		in
		fun cs -> List.sort cmp cs
	in
        (* Sylvain: NOT USED IN THE NEW IMPLEMENTATION OF LINTERMS 
	let normLt: ConsSet.Cs.t -> ConsSet.Cs.t
	=       let norm1: ConsSet.Cs.cT -> ConsSet.Cs.cT
		= fun c ->
			let c1 = ConsSet.Cs.cstr c in
			let lt = CstrC.Cstr.coefs c1 in
 			let c2 = {c1 with CstrC.Cstr.coefs = LinTerm.LinQ.normDep lt} in
			{c with ConsSet.Cs.cstr = c2}
		in
		fun cs -> List.map norm1 cs 
	in *)
	fun cs -> sort cs (* Sylvain: (normLt cs) *)

let chkSync: ConsSet.Cs.t -> Pol.t -> bool
= fun cs1 p ->
	let cs2 = pToConsSet p in
	(csNorm cs1) = (csNorm cs2)

let sort: ('a*'b) list -> ('a*'b) list
= fun a ->
	let cmp (i1, _) (i2, _) =
		match Pervasives.compare i1 i2 with
		| 0 -> invalid_arg "Support.sort"
		| n -> n
	in
	List.sort cmp a

let fToFrag: Cert.frag_t -> CertC.Cert.frag_t
= fun f -> List.map (fun (i, n) -> (i, nToNumC n)) f

let fragPr: CertC.Cert.frag_t -> string
= fun f ->
	let pairPr (i, n) =
		Printf.sprintf "(%i * %s)" i (CoqPr.charListTr (NumC.QNum.pr n))
	in
	String.concat " " (List.map pairPr f)

let cToCertCstr: Cert.cstr_t -> CertC.Cert.cstrT
= function
	| Cert.Direct f -> CertC.Cert.Direct (fToFrag f)
	| Cert.SplitEq (f1, f2) -> CertC.Cert.SplitEq (fToFrag f1, fToFrag f2)

let cstrPr: CertC.Cert.cstrT -> string
= function
	| CertC.Cert.Direct f -> "Direct: " ^ (fragPr f)
	| CertC.Cert.SplitEq (f1, f2) ->
		Printf.sprintf "SplitEq:\n\t%s\n\t%s" (fragPr f1) (fragPr f2)

let rec cToCertC: Cert.t -> CertC.Cert.t
= fun c ->
	let convL (l: (int * Cert.cstr_t) list) =
		sort (List.map (fun (i, c) -> (i, cToCertCstr c)) l)
	in
	match c with
	| Cert.Implies l -> CertC.Cert.Final (CertC.Cert.Implies (convL l))
	| Cert.Empty f -> CertC.Cert.Final (CertC.Cert.Empty (fToFrag f))
	| Cert.Bind (i, c, ce1, ce2) -> CertC.Cert.Bind (i, cToCertCstr ce1, cToCertC ce2)

let finalPr: CertC.Cert.finalT -> string
= function
	| CertC.Cert.Implies l ->
		let pr1 = fun (i, c) -> Printf.sprintf "%i: %s" i (cstrPr c) in
		"Implies\n" ^ (String.concat "\n" (List.map pr1 l))
	| CertC.Cert.Empty f -> "Empty " ^ (fragPr f)

let rec certCPr: CertC.Cert.t -> string
= function
	| CertC.Cert.Final f -> finalPr f
	| CertC.Cert.Bind (i, ce1, ce2) ->
		Printf.sprintf "Bind %i: %s\nin\n%s" i (cstrPr ce1) (certCPr ce2)

let joinEltPr: CertC.Cert.joinEltT -> string
= fun c ->
	Printf.sprintf "{id = %i;\narg1 = %s;\narg2 = %s}"
		(CertC.Cert.id c) (cstrPr (CertC.Cert.arg1 c)) (cstrPr (CertC.Cert.arg2 c))

let jToJoinElt: Cert.joinEltT -> CertC.Cert.joinEltT
= fun j -> {
	CertC.Cert.id = Cert.getId j;
	CertC.Cert.arg1 = cToCertCstr (Cert.getArg1 j);
	CertC.Cert.arg2 = cToCertCstr (Cert.getArg2 j)
}

let jToJoin: Cert.joinT -> CertC.Cert.joinT
= fun j -> List.map jToJoinElt j

let joinPr: CertC.Cert.joinT -> string
= fun j ->
	let sl = List.map joinEltPr j in
	String.concat "\n" sl

let bToPolBnd: Itv.QItv.bndT -> Pol.bndT
= function
	| Itv.QItv.Infty -> Pol.Infty
	| Itv.QItv.Open n -> Pol.Open (nToNb n)
	| Itv.QItv.Closed n -> Pol.Closed (nToNb n)

let iToItv: Itv.QItv.t -> Pol.itvT
= function
	| Itv.QItv.Bot -> failwith "Support.iToItv"
	| Itv.QItv.NotBot (low, up) ->
		{Pol.low = bToPolBnd low; Pol.up = bToPolBnd up}

let isEmpty: Pol.t -> CertC.Cert.t option 
= fun p -> None

let isIncl: Pol.t * Pol.t -> CertC.Cert.t option 
= fun (p1,p2) ->
	match Pol.isIncl p1 p2 with
	| Pol.NoIncl -> None
	| Pol.Incl cert -> Some (cToCertC cert)

let add: Pol.t * CstrC.Cstr.t -> Pol.t option * CertC.Cert.t
= fun (p,c) ->
	match Pol.add p (cToCstr c) with
	| Pol.Added (p', (ce, _)) -> (Some p', cToCertC ce)
	| Pol.Contrad ce -> (None, cToCertC ce)

let addp: Pol.t * CstrC.Cstr.t -> Backend.addPrecT
= fun (p,cc) ->
	match Pol.add p (cToCstr cc) with
	| Pol.Contrad ce -> Backend.Contrad (cToCertC ce)
	| Pol.Added (p', (fwd, bwd)) -> Backend.Added (p', (cToCertC fwd, cToCertC bwd))

let meet: Pol.t * Pol.t -> Pol.t option * CertC.Cert.meetT
= fun (p1,p2) ->
	let tr = fun ce -> {
		CertC.Cert.sh = (Pol.get_nxt p1);
		CertC.Cert.meetCert = (cToCertC ce)
	} in
	match Pol.meet p1 p2 with
	| Pol.Added (p, (fwd, _)) -> Some p, tr fwd
	| Pol.Contrad ce -> None, tr ce

let join: Pol.t * Pol.t -> Pol.t * CertC.Cert.joinT
= fun (p1,p2) ->
	let (p, cert) = Pol.join p1 p2 in
	(p, jToJoin cert)

let widen: Pol.t * Pol.t -> Pol.t * ConsSet.Cs.t
= fun (p1,p2) ->
	let p = Pol.widen p1 p2 in
	(p, pToConsSet p)

let bToBnd: Pol.bndT -> Cert.frag_t option -> Backend.bndT
= fun b opt ->
	match b with
	| Pol.Infty -> Backend.Infty
	| Pol.Open n ->
		(match opt with
		| None -> Backend.Infty
		| Some f -> Backend.Open (nToNumC n, fToFrag f))
	| Pol.Closed n -> 
		match opt with
		| None -> Backend.Infty
		| Some f -> Backend.Closed (nToNumC n, fToFrag f)

let getItv: Pol.t * LinTerm.LinQ.t -> Backend.itvT
= fun (p,lt) ->
	let (itv, lpf, upf) = Pol.itvize p (ltToVec lt) in
	{
		Backend.low = bToBnd (Pol.get_low itv) lpf;
		Backend.up = bToBnd (Pol.get_up itv) upf
	}

let getUpperBound : Pol.t * LinTerm.LinQ.t -> Backend.bndT
  = fun (p, l) ->
  let (b, pf) = ltToVec l |> Pol.getUpperBound p in
  bToBnd b pf

let getLowerBound : Pol.t * LinTerm.LinQ.t -> Backend.bndT
  = fun (p, l) ->
  let (b, pf) = ltToVec l |> Pol.getLowerBound p in
  bToBnd b pf

(* no more used ! *)
(*
let assign: Pol.t -> Backend.assignT -> Pol.t * CertC.Cert.asgT
=	let asgTr: Backend.assignT -> Pol.assign_t
	= fun a -> {
		Pol.var = progVarToVar a.Backend.var;
		Pol.lin = List.map (fun (x, n) -> (nToNb n, progVarToVar x)) a.Backend.lin;
		Pol.cst = Pol.S (nToNb a.Backend.cst)}
	in
	let asgCeTr: Cert.asgT -> CertC.Cert.asgT
	=	let tr = fun (i, ce) -> (i, cToCertC ce) in
	function
	| Cert.AsgInv (ce1, ce2) -> CertC.Cert.AsgInv (List.map tr ce1, cToCertC ce2)
	| Cert.AsgNInv (ce1, ce2) -> CertC.Cert.AsgNInv (cToCertC ce1, List.map tr ce2)
	in
	fun p a ->
		let (res,  ce) = Pol.assign p (asgTr a) in
		(res, asgCeTr ce)
*)

let project: Pol.t * ProgVar.PVar.t -> Pol.t * CertC.Cert.t
= fun (p,x) ->
	let x' = progVarToVar x in
	let (p', c) = Pol.project p x' in
	(p', cToCertC c)

let rename : (ProgVar.PVar.t * ProgVar.PVar.t) * Pol.t -> Pol.t
= fun ((x,y),p) -> Pol.rename (progVarToVar x) (progVarToVar y) p
