module Ident
= struct
	type t
	= int

	let compare: t -> t -> int
	= fun id1 id2 -> Pervasives.compare id1 id2

	let toVar: t -> Var.t
	= fun i -> Var.fromInt i

	let fromVar: Var.t -> t
	= fun x -> Var.toInt x

	let pr: t -> string
	= fun i -> "v" ^ (Pervasives.string_of_int i)
end

module Expr
= struct
	type affT
	=
		| Var of Ident.t
		| Cst of Nb.t
		| Prod of Nb.t * Ident.t
		| Sum of affT * affT

	type cmpT
	= Eq | Le | Lt

	type t
	= {
		aff: affT;
		cmp: cmpT;
		cst: Nb.t
	}

	let getAff: t -> affT
	= fun l -> l.aff

	let getCmp: t -> cmpT
	= fun l -> l.cmp

	let getCst: t -> Nb.t
	= fun l -> l.cst

	let toCstr: t -> Cstr.t
	= fun expr ->
		let affToVec: affT -> Nb.t * Vec.t
		= fun aff0 ->
			let rec build: Nb.t -> Vec.t -> affT -> Nb.t * Vec.t
			= fun cst vec -> function
				| Var id ->
					let var = Ident.toVar id in
					let coef = Vec.get vec var in
					(cst, Vec.set vec var (Nb.add coef Nb.u))
				| Cst n -> (Nb.add cst n, vec)
				| Prod (n, id) ->
					let var = Ident.toVar id in
					let coef = Vec.get vec var in
					(cst, Vec.set vec var (Nb.add coef n))
				| Sum (l1, l2) ->
					let (cst1, vec1) = build cst vec l1 in
					build cst1 vec1 l2
			in
			build Nb.z Vec.nil aff0
		in
		let typ =
			match getCmp expr with
			| Eq -> Cstr.Eq
			| Le -> Cstr.Le
			| Lt -> Cstr.Lt
		in
		let (cst, lin) = affToVec (getAff expr) in
		({Cstr.typ = typ; Cstr.v = lin; Cstr.c = Nb.sub (getCst expr) cst})
	
	let toAssign: Ident.t * affT -> Pol.assign_t
	= fun (id0, aff0) ->
		let rec build:Nb.t -> (Var.t * Nb.t) list -> affT -> Nb.t * (Var.t * Nb.t) list
		= fun cst coefs -> function
			| Var id ->
				let var = Ident.toVar id in
				if List.mem_assoc var coefs then
					let coef = List.assoc var coefs in
					(cst, (var, Nb.add coef Nb.u)::(List.remove_assoc var coefs))
				else
					(cst, (var, Nb.u)::coefs)
			| Cst n -> (Nb.add cst n, coefs)
			| Prod (n, id) ->
				let var = Ident.toVar id in
				if List.mem_assoc var coefs then
					let coef = List.assoc var coefs in
					(cst, (var, Nb.add coef n)::(List.remove_assoc var coefs))
				else
					(cst, (var, n)::coefs)
			| Sum (a1, a2) ->
				let (cst1, coefs1) = build cst coefs a1 in
				build cst1 coefs1 a2
		in
		let var = Ident.toVar id0 in
		let (cst, coefs) = build Nb.z [] aff0 in
		{
			Pol.var = var;
			Pol.lin = List.map (fun (a, b) -> (b, a)) coefs;
			Pol.cst = Pol.S cst
		}
end

type t
=
	| Bottom
	| NonBot of Pol.t

let top: t
= NonBot Pol.top

let meet: t -> Expr.t -> t
= fun p0 expr ->
	match p0 with
	| Bottom -> Bottom
	| NonBot pol ->
		let c = Expr.toCstr expr in
		match Pol.add pol c with
		| Pol.Contrad _ -> Bottom
		| Pol.Added (pol1, _) -> NonBot pol1

let join: t -> t -> t
= fun p1 p2 ->
	match p1, p2 with
	| Bottom, Bottom -> Bottom
	| Bottom, NonBot _ -> p2
	| NonBot _, Bottom -> p1
	| NonBot pol1, NonBot pol2 ->
		let (pol, _) = Pol.join pol1 pol2 in
		NonBot pol

let assign: t -> (Ident.t * Expr.affT) list -> t
= fun p a ->
	match p with
	| Bottom -> Bottom
	| NonBot pol ->
		let a1 =
			let f: Ident.t * Expr.affT -> Pol.assign_t list -> Pol.assign_t list
			= fun (id, aff) l ->
				let assign = Expr.toAssign (id, aff) in
				assign::l
			in
			List.fold_right f a []
		in
		NonBot (Pol.assignM pol a1)

let widen: t -> t -> t
= fun p1 p2 ->
	match p1, p2 with
	| Bottom, Bottom -> Bottom
	| Bottom, NonBot _ -> p2
	| NonBot _, Bottom -> top
	| NonBot pol1, NonBot pol2 -> NonBot (Pol.widen pol1 pol2)

let leq: t -> t -> bool
= fun p1 p2 ->
	match p1, p2 with
	| Bottom, Bottom | Bottom, NonBot _ -> true
	| NonBot _, Bottom -> false
	| NonBot pol1, NonBot pol2 ->
		match Pol.isIncl pol1 pol2 with
		| Pol.NoIncl -> false
		| Pol.Incl _ -> true

let pr: t -> string
= function
	| Bottom -> "bottom"
	| NonBot p ->
		if p = Pol.top then
			"top"
		else
			Pol.pr (fun x -> Ident.pr (Ident.fromVar x)) p
