(** A representation for program variables. *)

(** Linear terms are represented by trees (see {! Rtree}).
A value of type [t] identifies a variable and represents a path in a tree.
[XH] is the end of a path, [XO] means "take the left branch" and [XI] means "take the right branch". *)
type t = XH | XO of t | XI of t

(** [next v] gives the path to the next variable in a breadth-first search in a tree.
Next of [... XO XH] is [... XI XH],
next of [... XO (XI XH)] is [... XI (XO XH)] and
next of [XI (XI XH)] is [XO (XO (XO XH))]. *)
val next: t -> t

(** [cmp x1 x2] compares [x1] and [x2] with "less" meaning "comes before in a breadth-first search".
The result follows the same convention as [Pervasives.compare]. *)
val cmp: t -> t -> int
val compare : t -> t -> int

(** [horizon l] returns the path which immediately follows the maximum path found in [l]. *)
val horizon: t list -> t

(** [toInt p] returns the integer which [p] represents, according to the Coq definition of [positive].
If the conversion of [p] overflows the [int] representation, [Invalid_argument "Var.toInt"] is raised. *)
val toInt: t -> int

(** [fromInt i] converts a strictly positive integer into its bit string representation.
If [i] is negative or nil, [Invalid_argument "Var.fromInt"] is raised.*)
val fromInt: int -> t

(** Pretty-print a value [v] of type [t] as "vN", where N is the number
denoted by [v]. *)
val pr: t -> string

(** [pr' s v] pretty-prints [v] as "sN", where N is the number denoted by [v]. *)
val pr': string -> t -> string
