type t = Nb.t Rtree.t

type vec_t = t

let nil : t = Rtree.Nil

let cut l n r: vec_t =
	if Nb.cmpz n = 0 && l = Rtree.Nil && r = Rtree.Nil then
		Rtree.Nil
	else
		Rtree.Sub (l, n, r)

(* assume normalization *)
let rec isEq v1 v2: bool =
	match (v1, v2) with
	| (Rtree.Nil, Rtree.Nil) -> true
	| (Rtree.Sub _, Rtree.Nil) -> false
	| (Rtree.Nil, Rtree.Sub _) -> false
	| (Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2)) ->
		if Nb.cmp n1 n2 = 0 then
			isEq l1 l2 && isEq r1 r2
		else
			false

let set ve0 va0 n0 =
	let z = (Nb.cmpz n0 = 0) in
	let rec _set ve va =
		match ve, va with
		| Rtree.Nil, Var.XH when z -> Rtree.Nil
		| Rtree.Nil, Var.XH -> Rtree.Sub (Rtree.Nil, n0, Rtree.Nil)

		| Rtree.Sub (Rtree.Nil, _, Rtree.Nil), Var.XH when z -> Rtree.Nil
		| Rtree.Sub (l, _, r), Var.XH -> Rtree.Sub (l, n0, r)

		| Rtree.Nil, Var.XO t -> cut (_set Rtree.Nil t) Nb.z Rtree.Nil
		| Rtree.Nil, Var.XI t -> cut Rtree.Nil Nb.z (_set Rtree.Nil t)

		| Rtree.Sub (l, n, r), Var.XO t -> cut (_set l t) n r
		| Rtree.Sub (l, n, r), Var.XI t -> cut l n (_set r t)
	in
		_set ve0 va0

let mk (i: (Nb.t * Var.t) list) =
	List.fold_left (fun v (n, var) -> set v var n) Rtree.Nil i

let get (x : t) (v : Var.t) = Rtree.get Nb.z x v

let rec add v1 v2: vec_t =
	match (v1, v2) with
	| (Rtree.Nil, Rtree.Nil) -> Rtree.Nil
	| (Rtree.Sub _, Rtree.Nil) -> v1
	| (Rtree.Nil, Rtree.Sub _) -> v2
	| (Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2)) ->
		cut (add l1 l2) (Nb.add n1 n2) (add r1 r2)

let map f v0 =
	let rec _map v =
		match v with
		| Rtree.Nil -> Rtree.Nil
		| Rtree.Sub (l, n, r) -> cut (_map l) (f n) (_map r)
	in
	_map v0

let mult n v = map (Nb.mult n) v

let div n0 v = map (fun n -> Nb.div n n0) v

let elim (v: Var.t) (using: vec_t) (from: vec_t): vec_t =
	let n1 = get from v in
	match Nb.cmpz n1 with
	| 0 -> from
	| _ ->
		let n2 = get using v in
		match Nb.cmpz n2 with
		| 0 -> invalid_arg "Vec.elim"
		| _ -> add (mult (Nb.neg n1) using) (mult n2 from)

let isomorph v1 v2 =
	let rec _iso optr v1 v2 =
		match v1, v2 with
		| Rtree.Nil, Rtree.Nil -> (optr, true)
		| Rtree.Nil, _ | _, Rtree.Nil -> (None, false)
		| Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2) ->
			match Nb.cmpz n1, Nb.cmpz n2 with
			| 0, 0 ->
				let (optr1, res1) = _iso optr l1 l2 in
				if res1 then
					_iso optr1 r1 r2
				else
					(None, false)

			| 0,_ | _,0 -> (None, false)
			| _,_ ->
				let (res, r) =
					let ratio = Nb.div n1 n2 in
					match optr with
					| None -> (true, Some ratio)
					| Some r -> (Nb.cmp r ratio = 0, optr)
				in
				if res then
					let (_, res1) = _iso r l1 l2 in
					if res1 then
						_iso r r1 r2
					else
						(None, false)
				else
					(None, false)
	in
	let (ratio, _) = _iso None v1 v2 in ratio

let shift: Var.t -> t -> Var.t option Rtree.t -> Var.t * t * Var.t option Rtree.t
= fun nxt0 vec0 relocTbl0 ->
	let rec _shift nxt wip ve relocTbl =
		match ve, relocTbl with
		| Rtree.Nil, _ -> (nxt, wip, relocTbl)
		| Rtree.Sub (l, n, r), Rtree.Nil ->
			let (nxt1, wip1, reloc) =
				if Nb.cmpz n = 0 then
					(nxt, wip, None)
				else
					(Var.next nxt, set wip nxt n, Some nxt)
			in
			let (nxt2, wip2, lReloc) = _shift nxt1 wip1 l Rtree.Nil in
			let (nxt3, wip3, rReloc) = _shift nxt2 wip2 r Rtree.Nil in
			(nxt3, wip3, Rtree.Sub (lReloc, reloc, rReloc))

		| Rtree.Sub (l, n, r), Rtree.Sub (lReloc, reloc, rReloc) ->
			let (nxt1, wip1, reloc1) =
				if Nb.cmpz n = 0 then
					(nxt, wip, reloc)
				else
					match reloc with
					| Some x -> (nxt, set wip x n, reloc)
					| None -> (Var.next nxt, set wip nxt n, Some nxt)
			in
			let (nxt2, wip2, lReloc1) = _shift nxt1 wip1 l lReloc in
			let (nxt3, wip3, rReloc1) = _shift nxt2 wip2 r rReloc in
			(nxt3, wip3, Rtree.Sub (lReloc1, reloc1, rReloc1))
	in
	_shift nxt0 Rtree.Nil vec0 relocTbl0

let rec mult_t v1 v2: vec_t =
	match (v1, v2) with
	| (Rtree.Nil, Rtree.Nil) -> Rtree.Nil
	| (Rtree.Sub _, Rtree.Nil) -> Rtree.Nil
	| (Rtree.Nil, Rtree.Sub _) -> Rtree.Nil
	| (Rtree.Sub (l1, n1, r1), Rtree.Sub (l2, n2, r2)) ->
		cut (mult_t l1 l2) (Nb.mult n1 n2) (mult_t r1 r2)

let neg (x : t) = map Nb.neg x

let gcd v =
	let gcd = Rtree.fold (fun g a ->
		if Nb.cmpz a = 0 then
			g
		else
			match g with
			| None -> Some (Nb.toZ (Nb.abs a))
			| Some g -> Some (Nb.gcd g a)) None v
	in
	match gcd with
	| None -> Nb.u
	| Some (nGcd, dGcd) -> Nb.ofZ dGcd nGcd

let listVars: t list -> Var.t list
= fun l -> Rtree.pathsGet (Rtree.mskBuild (fun n -> Nb.cmpz n <> 0) l)

let toList : t -> (Nb.t * Var.t) list
=	let rmZeroes : (Nb.t * Var.t) list -> (Nb.t * Var.t) list
	= fun l -> List.filter (fun (n, _) -> Nb.cmpz n <> 0) l
	in
	fun v -> rmZeroes (Rtree.toList v)

let pr: (Var.t -> string) -> t -> string
= fun varPr v ->
	let nodePr a x =
		if Nb.cmpz a = 0 then
			""
		else
			(Nb.pr a) ^ "." ^ x
	in
	Rtree.pr " + " nodePr varPr v
