(** Vectors of rationals numbers and operations on them. *)

(** The type of vectors of rational numbers.
Zero coefficients are left implicit.
The internal representation is canonical. *)
type t = Nb.t Rtree.t

(** A vector with coefficients all set to zero. *)
val nil: t

(** Equality of two vectors as point-wise equality of their coefficients. *)
val isEq: t -> t -> bool

(** [set v x n] sets the coefficient of the variable [x] to [n] in vector [v]. *)
val set: t -> Var.t -> Nb.t -> t

(** Build a vector from a list of coefficients.
If a variable appears multiple times, later coefficients override earlier ones. *)
val mk: (Nb.t * Var.t) list -> t

(** [get v x] returns the coefficient of variable [x] in vector [v].
Remember that all variables, unless otherwise assigned, have a nil coefficient. *)
val get: t -> Var.t -> Nb.t

(** Add two vectors. *)
val add: t -> t -> t

(** [elim x v1 v2] eliminates the [x] from [v2] using [v1].
If [v2] has coefficient zero for variable [x], it is returned untouched.
Otherwise, if variable [x] has coefficient zero in [v1], [Invalid_argument "Vec.elim"] is raised. *)
val elim: Var.t -> t -> t -> t

(** Multiply a vector by a rational number. *)
val mult: Nb.t -> t -> t

(** Negate all the coefficients of a vector. *)
val neg: t -> t

(** [isomorph v1 v2] returns [Some r] if [v1] is equal to [mult r v2].
If there is no such [r], [None] is returned. *)
val isomorph: t -> t -> Nb.t option

(**/**)
val shift: Var.t -> t -> Var.t option Rtree.t -> Var.t * t * Var.t option Rtree.t
(**/**)

(** Pointwise multiplication of two vectors. *)
val mult_t: t -> t -> t

(** Divide a vector by a rational number. *)
val div: Nb.t -> t -> t

(** Compute the greatest common divisor of the coefficients of a vector. *)
val gcd: t -> Nb.t

(** Return the list of variables which have non-zero coefficient in the list.
There can be duplicates. *)
val listVars: t list -> Var.t list

(** Convert a vector of type [t] into a list of pairs (coefficient, variable). *)
val toList : t -> (Nb.t * Var.t) list

(** [pr b v] pretty-prints [v] using names for variables from [b]. *)
val pr: (Var.t -> string) -> t -> string
