(*
Polynomial type
Important : polynomials given as string must be EXPANDED and SIMPLIFIED!

The null polynomial is represented as [[VariablesInt.null], ScalarQ.zero] which is [[0], Q.zero].
*)

module Misc = 
	struct
	(** sublist l i j returns l[i:j] (j exluded) *)
	let rec(sublist : 'a list -> int -> int -> 'a list)
		= fun l i j->
		match (l,i,j) with
		| ([],_,_) -> []
		| (_,_,0) -> []
		| (p :: tail,_,_) -> if i > 0 
		then sublist tail (i-1) (j-1)
		else p :: (sublist tail i (j-1))
	
	(** index s c returns the index of the character c in the string s if it exists, -1 otherwise *)
	let index s c = try (String.index s c) with
		| Not_found -> -1;;

	(** substring s i1 returns the substring of s starting at index i *)
	let rec(substring : string -> int -> string)
		= fun s i ->
		(String.sub s i ((String.length s) - i));;
		
	let (findi : ('a -> bool) -> 'a list -> int)
	= let rec(findi_rec : ('a -> bool) -> 'a list -> int -> int)
		= fun prop l k ->
		match l with
		| [] -> -1
		| v :: tail -> if prop v then k else findi_rec prop tail (k+1)
	in fun prop l -> 
	findi_rec prop l 0;;
		
	let (popi : 'a list -> int -> 'a list)
		= fun l i ->
		if i >= 0 && i <= List.length l 
		then (sublist l 0 i) @ (sublist l (i+1) (List.length l))
		else l;;
		
	let (pop : ('a -> 'a -> bool) -> 'a list -> 'a -> 'a list)
		= fun eq l x ->
		let i = findi (fun y -> eq x y) l in
		popi l i;;
end;;

module type Scalar = 
	sig
	type t
	val zero : t
	val add : t -> t -> t
	val mul: t -> t -> t
	val le: t -> t -> bool
	val lt: t -> t -> bool
	val to_string: t -> string
	val of_string: string -> t
	val equal: t -> t -> bool
	val of_int: int -> t
	val to_int: t -> int
	val well_formed: t -> bool
	val minus_one: t
end;;

module type Variables =
	sig
	type t
	val null: t
	val equal: t -> t -> bool
	val compare: t -> t -> int
	val to_string: t -> string
	val of_string: string -> t
	val to_int: t -> int
	val of_int: int -> t	
end;;

module Polynomial (Coeff : Scalar)(Var : Variables) =
	struct
	type coeff = Coeff.t
	type var = Var.t
	
	module V = Var;;
	module S = Coeff;;
	
	module MonomialBasis = 
		struct 
		(* monomial_basis represents a MONIC monomial (i.e. leading coeff = 1) *)
		type t = var list
		
		let rec(to_string : t -> string)
			= fun vlist ->
			let rec(to_string_rec : var list -> var -> int -> string)
				= fun vlist prevar power ->
					match vlist with
					| [] -> let str = String.concat "" ["x";Var.to_string prevar] in 
						if power = 1 
						then str 
						else String.concat "" [str ; "^" ; string_of_int power]
					| var :: tail -> 
					if Var.equal var prevar 
					then to_string_rec tail prevar (power+1)
					else let str = String.concat "" ["x";Var.to_string prevar] in 
						if power = 1 
						then String.concat "" [str ; to_string_rec tail var 1]
						else String.concat "" [str ; "^" ; string_of_int power ; to_string_rec tail var 1]
			in match vlist with
			| [] -> ""
			| _ -> to_string_rec (List.tl vlist) (List.nth vlist 0) 1
	
		let rec (compare : t -> t -> int)
			= fun m1 m2 ->
			match m1, m2 with
			| [], [] -> 0
			| [], _ :: _ -> -1
 		 	| _ :: _, [] -> 1
 		 	| x :: m1', x' :: m2' ->
 		   	let i = Pervasives.compare x x' in
 		    	if i = 0
		     	then compare m1' m2'
		     	else i
		     	
		let (is_linear : t -> bool)
			= fun m -> 
			List.length m <= 1;;
			
		let (get_vars : t -> var list)
			= fun m -> 
			List.fold_left
			(fun l v -> if List.mem v l then l else v::l)
			[] m
			
		let (get_multiplicity : t -> var -> int)
			= fun m v ->
			List.length (List.filter (fun v' -> Var.equal v v') m)
					
	end

	module Monomial = 
		struct

		type t = (MonomialBasis.t * coeff)

		let (to_string : t -> string)
			= fun m -> let (vlist, c) = m in 
			match vlist with
			| [] -> ""
			| [var] when Var.equal var Var.null -> Coeff.to_string c
			| _ -> if Coeff.equal c (Coeff.of_int 1)
				then MonomialBasis.to_string vlist
				else if Coeff.lt c (Coeff.of_int 0) 
					then String.concat "" ["(";Coeff.to_string c;")*";MonomialBasis.to_string vlist]
					else String.concat "" [Coeff.to_string c ; "*" ; MonomialBasis.to_string vlist]
		
		(** compare_monomial x y returns 0 if x is equal to y, a negative integer if x is less than y
		Use this function to sort the monomials in a polynomial
		This function do NOT compare the monomial coefficients
		*)
		let (compare : t -> t -> int)
			= fun (m1,_) (m2,_) ->
			MonomialBasis.compare m1 m2 ;;
		
		(** monomial_equal (m1,c1) (m2,c2) returns true if the monomials (m1,c1) and (m2,c2) are equal*)
		let (equal : t -> t -> bool)
			= fun (m1,c1) (m2,c2) ->
			MonomialBasis.compare m1 m2 = 0 && Coeff.equal c1 c2;;				

		let canon : t -> t option	
			= fun (m, a) ->
  			if not (Coeff.well_formed a)
  			then None
 			else match m with
  	  			| [] -> Some ([Var.null], a)
   				| _ :: _ ->
       			if List.exists (fun x -> Var.compare Var.null x = 0) m
       			then if List.length m > 1 then None else Some ([Var.null], a)
       			else Some (List.sort Var.compare m, a)
	
		let mkMonomial : MonomialBasis.t -> coeff -> t option
	  		= fun m a -> canon (m, a)
	
		(** of_string s vlist returns a monomial corresponding to s with variables in vlist 
		The generated monomial is not sorted yet 
		if s is "", returns (["1"],0)*)
		let rec(of_string : string -> MonomialBasis.t -> t) 
		= fun s vlist-> 
		let i = (Misc.index s '*') in 
			if i <> -1 
			then let s2 = (String.sub s 0 i) and (s3,c) = of_string(Misc.substring s (i+1)) vlist in 
				let j = Misc.index s2 'x' in
					if j <> -1 
					then let v = Var.of_string (String.sub s2 (j+1) ((String.length s2) - (j+1))) in 
						if List.exists (fun x -> Var.equal x v) vlist
						then (v :: s3,c)
						else Pervasives.failwith "Polynomial.Monomial.of_string"
					else (s3, Coeff.mul c (Coeff.of_string s2))
			else let j = Misc.index s 'x' in
				if j <> -1 
				then let v = Var.of_string (String.sub s (j+1) ((String.length s) - (j+1))) in 
					if List.exists (fun x -> Var.equal x v) vlist
					then ([v],Coeff.of_int 1)
					else Pervasives.failwith "Polynomial.Monomial.of_string"
				else ([], Coeff.of_string s)
		
		let rec(mul : t -> t -> t)
		= fun (m1,c1) (m2,c2) -> 
		if Coeff.equal (Coeff.mul c1 c2) Coeff.zero 
		then ([Var.null],Coeff.zero)
		else
		let new_m = match (m1,m2) with
		| ([x],_) when Var.equal x Var.null -> m2
		| (_,[x]) when Var.equal x Var.null -> m1
		| _ -> m1 @ m2
		in (List.sort Var.compare new_m, Coeff.mul c1 c2)
		
		let (is_linear : t -> bool)
		= fun (m,c) -> 
		MonomialBasis.is_linear m;;
	end

	type t = Monomial.t list

	let to_string : t -> string
	  = fun p ->
	  List.map Monomial.to_string p
	  |> String.concat " + "
	  |> fun s -> if String.length s = 0 then "0" else s;;
	
	let canon : t -> t
  		= let rec (collapseDups : t -> t)
      		= function
      			| [] | _ :: [] as p -> p
      			| m :: (m' :: p' as p) ->
	 			if Monomial.compare m m' = 0
	 			then collapseDups ((Pervasives.fst m, Coeff.add (Pervasives.snd m) (Pervasives.snd m')) :: p')
	 			else m :: collapseDups p
    		in
    		let fixConstant
     		 = fun p ->
      			let (cst, m) = List.partition (fun (m, _) -> m = [] || m = [Var.null]) p in
     	 			([Var.null], List.fold_left (fun n (_, a) -> Coeff.add n a) Coeff.zero cst) :: m
    		in
    		fun p ->
    		fixConstant p
    		|> List.filter (fun (_, a) -> Coeff.well_formed a)
    		|> List.map Monomial.canon
    		|> List.map (function Some m -> m | None -> Pervasives.failwith "Polynomial.canon")
    		|> List.sort Monomial.compare
    		|> collapseDups
    		|> List.filter (fun (_, a) ->
				if Coeff.equal a Coeff.zero 
				then false
				else if Coeff.well_formed a 
					then true
					else Pervasives.failwith "Polynomial.canon")
    		|> function _ :: _ as p' -> p' | [] -> [[Var.null], Coeff.zero]

	let mkPolynomial : t -> coeff -> t
  		= fun p cst -> ([Var.null], cst) :: p |> canon
	
	let (get_vars : t -> var list)
		= fun p ->
		List.fold_left 
		(fun l (m,c) -> List.fold_left (fun l v -> if List.mem v l then l else v::l) l (MonomialBasis.get_vars m)) 
		[] p 
	
	let (get_multiplicity : t -> var -> int)
		= fun p v ->
		List.fold_left
		(fun i (m,c) -> i + (MonomialBasis.get_multiplicity m v))
		0
		p
			
	(* POLYNOMIAL FROM STRING *)
	(* Variables of a monomial are sorted depending on the String.compare function *)
	(* Monomials of a polynomial are sorted depending on the compare_monomial function *)

	(** string_to_polynomial s vlist returns a polynomial corresponding to s with variables in vlist *)
	let rec(of_string : string -> var list -> t)
		= let rec(of_string_rec : string -> var list -> t)
			= fun s vlist ->
			let i = (Misc.index s '+') in 
			if i <> -1 
			then let (vlist',c) = Monomial.of_string (String.sub s 0 i) vlist in 
				(List.sort Var.compare vlist', c) :: (of_string_rec(Misc.substring s (i+1)) vlist)
			else let (mlist,c) = Monomial.of_string s vlist in
				[(List.sort Var.compare mlist, c)]
		in fun s vlist -> of_string_rec s vlist |> canon
	
	(** of_cste f returns the constant polynomial f*)
	let (of_cste : coeff -> t)
		= fun i ->
		[([Var.null],i)];;
	
	let (of_var : var -> t)
		= fun v ->
		[([v],Coeff.of_int 1)];;
		
	let rec(is_constant : t -> bool)
		= fun p ->
		match p with
		| [] -> true
		| (m,coeff) :: tail -> String.compare (Var.null |> Var.to_string) (String.concat "" (List.map (Var.to_string) m)) = 0 && (is_constant tail)

	let (zero : t) = [([Var.null], Coeff.zero)];;
	
	let rec(is_zero : t -> bool)
		= fun p ->
		if p = [] then true (* nécessaire? *)
		else if List.length p = 1 
			then let (mono,coeff) = (List.nth p 0) in
				MonomialBasis.compare mono [Var.null] = 0 && Coeff.equal coeff Coeff.zero
			else false;;
	
	let (is_linear : t -> bool)
		= fun p ->
		List.for_all 
			(fun m -> Monomial.is_linear m) 
			p;;
			
	let rec(add : t -> t -> t)
		= let rec(add_rec : t -> t -> t)
			= fun p1 p2 ->
			match (p1,p2) with
			| ([],poly2) -> poly2
			| (poly1,[]) -> poly1
			| ((m1,c1) :: tail1, (m2,c2) :: tail2) -> let comp = MonomialBasis.compare m1 m2 in 
			if comp = 0
				then if Coeff.equal (Coeff.add c1 c2) Coeff.zero 
					then add_rec tail1 tail2
					else (m1,Coeff.add c1 c2)::(add_rec tail1 tail2)
				else if comp < 0 (*m1 < m2*) 
					then (m1,c1)::(add_rec tail1 ((m2,c2)::tail2))
					else (m2,c2)::(add_rec ((m1,c1)::tail1) tail2)
		in fun p1 p2 ->
		add_rec p1 p2 |> canon

	(* utile de retrier? Je pense que non*)
	let rec(mul : t -> t -> t) 
		= let rec(mul_rec : t -> t -> t)
			= fun p1 p2 ->
			match (p1,p2) with
			| ([],_) -> []
			| (m::tail1,p2) -> List.fold_left 
				add (mul_rec tail1 p2) (List.map (fun m2 -> [Monomial.mul m m2]) p2)
		in fun p1 p2 -> 
		mul_rec p1 p2 |> canon

	let rec (equal : t -> t -> bool)
		= fun p1 p2 ->
		match (p1,p2) with
		| ([],[]) -> true
		| (_,[]) | ([],_) -> false
		| (m1::tail1 , m2::tail2) -> Monomial.equal m1 m2 && equal tail1 tail2
	
	let rec (compare : t -> t -> int)
		= fun p1 p2 ->
		match (p1,p2) with
		| ([],[]) -> 0
		| (_,[]) -> -1
		| ([],_) -> 1
		| (m1 :: tl1, m2 :: tl2) -> let r =  Monomial.compare m1 m2 in 
		if r = 0 then compare tl1 tl2
		else r;;
	
	(* COEFFICIENT ACCESS *)

	(** monomial_coefficient p m returns the coefficient of the monomial_basis m in p *)
	let rec(monomial_coefficient : t -> MonomialBasis.t -> coeff)
		= fun p m ->
		match (p,m) with
		| ([],_) | (_,[]) -> Coeff.zero
		| ((m1,c)::tail, m2) -> if MonomialBasis.compare m1 m2 = 0
			then c
			else if MonomialBasis.compare m1 m2 < 0
				then monomial_coefficient tail m
				else Coeff.zero ;;
	
	let rec(sub_monomial : MonomialBasis.t -> MonomialBasis.t -> (MonomialBasis.t * bool))
	= fun m1 m2 ->
	match (m1,m2) with
	| (_,[]) -> (m1,true)
	| ([],_) -> ([],false)
	| (v1::tail1,v2::tail2) -> if Var.compare v1 v2 = 0
		then sub_monomial tail1 tail2
		else let (l,b) = sub_monomial tail1 m2 in (v1::l,b);;
	
	(** monomial_coefficient_other p monomial returns the "coefficient" of monomial in p in the following way:
	monomial_coefficient_other 3xyz - 2yz ["y","z"] = 3x - 2*)
	let (monomial_coefficient_other : t -> MonomialBasis.t -> t)
		= let rec(monomial_coefficient_other_rec : t -> MonomialBasis.t -> t)
			= fun p m ->
			match (p,m) with
			| ([],_) | (_,[]) -> []
			| ((m1,c)::tail, m2) -> if List.length m1 >= List.length m2 && MonomialBasis.compare (Misc.sublist m1 0 (List.length m2)) m2 > 0 (* m1 > m2 *)
				then []
				else let (l,b) = sub_monomial m1 m in if b
					then add [(l,c)] (monomial_coefficient_other_rec tail m2)
					else monomial_coefficient_other_rec tail m2
		in fun p m -> 
		monomial_coefficient_other_rec p m |> canon

	let (get_constant : t -> coeff)
		= fun p ->
		monomial_coefficient p [Var.null]

	let (remove_monomial : t -> MonomialBasis.t -> t)
		= fun p m ->
		Misc.pop (fun (m1,_) (m2,_) -> MonomialBasis.compare m1 m2 = 0) p (m,Coeff.zero) |> canon;;
	
	module Invariant
 		= struct
  	(** This module contains an executable specification of what it means to be
	a well-formed polynomial. *)

  		let rec helper_sorted : ('a -> 'a -> bool) -> 'a list -> bool
	    = fun f ->
    	function
   		| [] | _ :: [] -> true
   		| x :: ((x' :: _) as l') -> f x x' && helper_sorted f l'

  		module Monom
    	= struct

    		let check_sorted : Monomial.t -> bool
      		= fun (l, _) -> helper_sorted (fun x x' -> Var.compare x x' <= 0) l

    		(** Variable Var.null designates the constant.  It cannot be found in a monomial
			along with other variables. *)
    		let check_constant : Monomial.t -> bool
      		= fun (l, _) ->
      		match l with
		    | [] -> true
			| [x] when Var.equal x Var.null -> true
      		| _ :: _ ->
			try List.find (fun x -> Var.compare Var.null x = 0) l |> Pervasives.ignore; false
			with Not_found -> true

    		let check : Monomial.t -> bool
      		= fun (vlist,c) -> Coeff.well_formed c && check_sorted (vlist,c) && check_constant (vlist,c)
  		end

  		(* strict negativity enforces no duplicates *)
		let check_sorted : t -> bool
    	= helper_sorted (fun m m' -> Monomial.compare m m' < 0)

  		(** The constant monomial is [[]] or [["1"]], but only one of them can appear. *)
 		let check_constant : t -> bool
    		= let has
			= fun m p ->
			try List.find (fun (m', _) -> MonomialBasis.compare m m' = 0) p |> Pervasives.ignore; true
			with Not_found -> false
      		in
      		fun p -> not (has [] p) || not (has [Var.null] p)
      
  		let check : t -> bool
    		= fun p ->
    		match p with
    		| [] -> false
    		| [[x], a] when Var.equal Var.null x && Coeff.equal a Coeff.zero -> true
    		| _ -> List.for_all Monom.check p && check_sorted p && check_constant p

  		let checkOrFail : t -> unit
    		= fun p ->
    		if not (check p)
    		then p
	 		|> to_string
	 		|> Printf.sprintf "Polynomial.Invariant.checkOrFail: %s"
	 		|> Pervasives.failwith
	end;;
end;;

module ScalarQ = 
	struct
	type t = Q.t
	let zero = Q.of_int 0
	let minus_one = Q.of_int (-1)
	let add = Q.add
	let mul = Q.mul
	let le = Q.leq
	let lt = Q.lt
	let to_string = Q.to_string
	let of_string = Q.of_string
	let equal = Q.equal
	let of_int = Q.of_int
	let to_int = Q.to_int
	let well_formed x = Q.classify x = Q.NZERO
end;;

module ScalarZ = 
	struct
	type t = Z.t
	let zero = Z.of_int 0
	let minus_one = Z.of_int (-1)
	let add = Z.add
	let mul = Z.mul
	let le = Z.leq
	let lt = Z.lt
	let to_string = Z.to_string
	let of_string = Z.of_string
	let equal = Z.equal
	let of_int = Z.of_int
	let well_formed x = not (Z.equal x Z.zero)
	let to_int = Z.to_int
end;;

module VariablesInt = 
	struct
	type t = int
	let null = 0
	let equal x y = x = y
	let compare x y = if x < y then -1 else if x = y then 0 else 1 
	let to_string = string_of_int
	let of_string = int_of_string
	let to_int x = x
	let of_int x = x	
end;;

module PolyQ = Polynomial(ScalarQ)(VariablesInt);;
module PolyZ = Polynomial(ScalarZ)(VariablesInt);;
