(*
Polynomial type
Important : polynomials given as string must be EXPANDED and SIMPLIFIED!

The null polynomial is represented as [[VariablesInt.null], ScalarQ.zero] which is [[0], Q.zero].
*)

module Misc :	sig
	(** sublist l i j returns l[i:j] (j exluded) *)
	val sublist : 'a list -> int -> int -> 'a list
	
	(** index s c returns the index of the character c in the string s if it exists, -1 otherwise *)
	val index : string -> char -> int

	(** substring s i1 returns the substring of s starting at index i *)
	val substring : string -> int -> string
	
	val findi : ('a -> bool) -> 'a list -> int
		
	val popi : 'a list -> int -> 'a list
		
	val pop : ('a -> 'a -> bool) -> 'a list -> 'a -> 'a list
end

module type Scalar =
	sig
	type t
	val zero : t
	val add : t -> t -> t
	val mul: t -> t -> t
	val le: t -> t -> bool
	val lt: t -> t -> bool
	val to_string: t -> string
	val of_string: string -> t
	val equal: t -> t -> bool
	val of_int: int -> t
	val to_int: t -> int
	val well_formed: t -> bool
	val minus_one: t
end

module type Variables =
	sig
	type t
	val null: t
	val equal: t -> t -> bool
	val compare: t -> t -> int
	val to_string: t -> string
	val of_string: string -> t
	val to_int: t -> int
	val of_int: int -> t	
end

module ScalarZ : Scalar
module VariablesInt : Variables

module PolyZ : 
	sig
	type coeff = ScalarZ.t
	type var = VariablesInt.t
	
	module V : Variables
	module S : Scalar
	
	module MonomialBasis : sig 
		(* monomial_basis represents a MONIC monomial (i.e. leading coeff = 1) *)
		type t = var list
		
		val to_string : t -> string
	
		val compare : t -> t -> int
		
		val is_linear : t -> bool
		
		val get_multiplicity : t -> var -> int
	end

	module Monomial : sig
		
		type t = (MonomialBasis.t * coeff)

		val to_string : t -> string
		
		(** compare_monomial x y returns 0 if x is equal to y, a negative integer if x is less than y
		Use this function to sort the monomials in a polynomial
		This function do NOT compare the monomial coefficients
		*)
		val compare : t -> t -> int
		
		(** monomial_equal (m1,c1) (m2,c2) returns true if the monomials (m1,c1) and (m2,c2) are equal*)
		val equal : t -> t -> bool			

		val canon : t -> t option	
	
		val mkMonomial : MonomialBasis.t -> coeff -> t option
	
		(** of_string s vlist returns a monomial corresponding to s with variables in vlist 
		The generated monomial is not sorted yet 
		if s is "", returns (["1"],0)*)
		val of_string : string -> MonomialBasis.t -> t
		
		val mul : t -> t -> t
	end

	type t = Monomial.t list

	val to_string : t -> string
	
	val canon : t -> t

	val mkPolynomial : t -> coeff -> t

	val get_vars : t -> var list
	
	val get_multiplicity : t -> var -> int
		
	(* POLYNOMIAL FROM STRING *)
	(* Variables of a monomial are sorted depending on the String.compare function *)
	(* Monomials of a polynomial are sorted depending on the compare_monomial function *)

	(** string_to_polynomial s vlist returns a polynomial corresponding to s with variables in vlist *)
	val of_string : string -> var list -> t
	
	(** poly f returns the constant polynomial f*)
	val of_cste : coeff -> t
	
	val of_var : var -> t
	
	val is_constant : t -> bool
	
	val zero : t	
	
	val is_zero : t -> bool

	val is_linear : t -> bool
	
	val add : t -> t -> t

	(* utile de retrier? Je pense que non*)
	val mul : t -> t -> t

	val equal : t -> t -> bool
	
	val compare : t -> t -> int
	
	(* COEFFICIENT ACCESS *)

	(** monomial_coefficient p m returns the coefficient of the monomial_basis m in p *)
	val monomial_coefficient : t -> MonomialBasis.t -> coeff
	
	val sub_monomial : MonomialBasis.t -> MonomialBasis.t -> (MonomialBasis.t * bool)
	
	(** monomial_coefficient_other p monomial returns the "coefficient" of monomial in p in the following way:
	monomial_coefficient_other 3xyz - 2yz ["y","z"] = 3x - 2*)
	val monomial_coefficient_other : t -> MonomialBasis.t -> t

	val get_constant : t -> coeff
	
	val remove_monomial : t -> MonomialBasis.t -> t
	
	module Invariant : sig
  	(** This module contains an executable specification of what it means to be
	a well-formed polynomial. *)

  		val helper_sorted : ('a -> 'a -> bool) -> 'a list -> bool

  		module Monom : sig

    		val check_sorted : Monomial.t -> bool

    		(** Variable Var.null designates the constant.  It cannot be found in a monomial
			along with other variables. *)
    		val check_constant : Monomial.t -> bool

    		val check : Monomial.t -> bool
  		end

  		(* strict negativity enforces no duplicates *)
		val check_sorted : t -> bool

  		(** The constant monomial is [[]] or [["1"]], but only one of them can appear. *)
 		val check_constant : t -> bool
      
  	val check : t -> bool
  	
  	val checkOrFail : t -> unit
	end
end
