type 'a mayFailT =
| Ok of 'a
| Failed of string

module P = Map.Make (String)

module IntOrdered
= struct
	type t = int

	let compare: int -> int -> int
	= fun i1 i2 -> Pervasives.compare i1 i2
end

module M = Map.Make (IntOrdered)

type hashT = int P.t

let hashPr: hashT -> string
= fun h ->
	let pr: string -> int -> string -> string
	= fun m id str -> Printf.sprintf "%s%s -> %i\n" str m id
	in
	P.fold pr h ""

let hashGet: string -> int -> hashT -> int * int * hashT
= fun m nxt h ->
	if P.mem m h then
		(P.find m h, nxt, h)
	else
		(nxt, nxt + 1, P.add m nxt h)

type pathT = {
	m: string;
	path: string
}

let pathPr: pathT -> string
= fun p -> Printf.sprintf "module %s is %s" p.m p.path

type t = {
	paths: pathT M.t;
	depBy: int list M.t;
	depOn: int M.t;
	roots: int list; (** the list of modules which have no dependencies *)
	hash: hashT;
	nxt: int (** the next unallocated identifier *)
}

let nil: t
= {
	paths = M.empty;
	depBy = M.empty;
	depOn = M.empty;
	roots = [];
	hash = P.empty;
	nxt = 1
}

let pr: t -> string
= fun dat ->
	let prList: int list -> string
	= function
	| [] -> "nothing"
	| a::tail -> List.fold_left (fun str a -> Printf.sprintf "%s, %i" str a) (string_of_int a) tail
	in
	let prPath: int -> pathT -> string -> string
 	= fun id p str -> Printf.sprintf "%s%i: %s\n" str id (pathPr p)
	in
	let prDepBy: int -> int list -> string -> string
	= fun id deps str -> Printf.sprintf "%s%s depend on %i\n" str (prList deps) id
	in
	let prDepOn: int -> int -> string -> string
	= fun id cnt str -> Printf.sprintf "%s%i depends on %i modules\n" str id cnt
	in
	let prRoots: int list -> string
	= fun roots -> Printf.sprintf "%s depend on nothing\n" (prList roots)
	in
	Printf.sprintf "paths:\n%sdepBy:\n%sdepOn:\n%sroots:\n%shash:\n%snxt: %i\n"
		(M.fold prPath dat.paths "")
		(M.fold prDepBy dat.depBy "")
		(M.fold prDepOn dat.depOn "")
		(prRoots dat.roots)
		(hashPr dat.hash)
		dat.nxt

let pathAssoc: int -> string -> string -> pathT M.t -> pathT M.t mayFailT
= fun id m path paths ->
	if M.mem id paths then
                let old = (M.find id paths) in
                Failed ("ERROR: identifier " ^ old.m ^ " already associated with " ^ old.path ^ " (" ^ m ^ ":" ^ path ^ ")")
	else
		Ok (M.add id {m = m; path = path} paths)

let addEdges: (int * int) list -> int list M.t -> int list M.t
= fun edges depBy ->
	let add: int list M.t -> int * int -> int list M.t
	= fun d (idOn, idBy) ->
		if idOn = idBy then
			d
		else
			if M.mem idOn d then
				let deps = M.find idOn d in
				if List.exists (fun i -> idOn = i) deps then
					d
				else
					M.add idOn (idBy::deps) d
			else
				M.add idOn [idBy] d
	in
	List.fold_left add depBy edges

let depends: string -> string list -> t -> t mayFailT
= fun path deps d ->
	let name = U.modName path in
	let (id, nxt, h) = hashGet name d.nxt d.hash in
	match pathAssoc id name path d.paths with
	| Failed _ as err -> err
	| Ok paths ->
		match List.length deps with
		| 0 -> Ok {d with paths = paths; roots = id::d.roots; hash = h; nxt = nxt}
		| n ->
			let (edges, nxt, h1) =
				let f: (int * int) list * int * hashT -> string -> (int * int) list * int * hashT
				= fun (edges, nxt, h) dep ->
					let name = U.modName dep in
					let (id1, nxt, h) = hashGet name nxt h in
					((id1, id)::edges, nxt, h)
				in
				List.fold_left f ([], nxt, h) deps
			in
			let depOn = M.add id n d.depOn in
			Ok {d with
				paths = paths;
				depBy = addEdges edges d.depBy;
				depOn = depOn;
				hash = h1;
				nxt = nxt}

let rec tsort: t -> int list -> int list mayFailT
= fun d order ->
	match d.roots with
	| [] ->
		if M.is_empty d.depBy && M.is_empty d.depOn then
			Ok (List.rev order)
		else
			Failed (pr d)
	| n::tail ->
		let d = {d with roots = tail} in
		if M.mem n d.depBy then
			let nOut = M.find n d.depBy in
			let d = {d with depBy = M.remove n d.depBy} in
			match nOut with
			| [] -> Failed (Printf.sprintf "%i depends on nothing" n)
			| _::_ as l ->
				let fn d n =
					let cnt = M.find n d.depOn in
					if cnt = 1 then
						{d with roots = n::d.roots; depOn = M.remove n d.depOn}
					else
						{d with depOn = M.add n (cnt -1) d.depOn}
				in
				let d1 = List.fold_left fn d l in
				tsort d1 (n::order)
		else
			tsort d (n::order)

let order: t -> string list mayFailT
= fun d ->
	match tsort d [] with
	| Failed _ as err -> err
	| Ok order -> Ok (List.map (fun i -> (M.find i d.paths).path) order)
