(** This module suggests a simple connection interface to the VPL.
It is meant to be customized by the user of the library if needed.
It makes the initial assumption that each variable has a unique identifier
which can be converted to bit string (a [Var.t]).
The VPL is sensitive to the length of the bit strings (the shorter the better).
This assumption may only hold satisfactorily for a rather small number of variables.
This interface does not make it possible to exploit the certificate generation the VPL offers. *)

module Ident
: sig
(** Module [Ident] describes the type of identifiers used to reference variables and operations of them.
Internally the VPL represents variables using type [Var.t], the elements of which are paths inside trees. *)

(** The type of variables is essentially up to the user.
The only restriction is that there should be a one-to-one correspondence between elements of type [t]
and elements of [Var.t]. *)
	type t
	= int

(** Compare two elements of type [t], with a result convention compatible with [Pervasives.compare]. *)
	val compare: t -> t -> int

(** Convert a [t] to a [Var.t]. *)
	val toVar: t -> Var.t

(** Convert a [Var.t] to a [t]. *)
	val fromVar: Var.t -> t

(** Return a textual representation of a [t]. *)
	val pr: t -> string
end

module Expr
: sig
(** Module [Expr] describes a format for linear constraints that are given to the VPL.
The format used internally is more basic: see {! Cstr}. *)

(** A simple type for affine expressions. *)
	type affT
	=
		| Var of Ident.t
		| Cst of Nb.t
		| Prod of Nb.t * Ident.t
		| Sum of affT * affT

(** A comparison operator. *)
	type cmpT
	= Eq | Le | Lt

(** The type of a linear constraint.
An element [c] of type [t] represents the constraint [c.aff] [c.cmp] [c.cst]. *)
	type t
	= {
		aff: affT;
		cmp: cmpT;
		cst: Nb.t
	}

(** Accessor for the [aff] field of a [t]. *)
	val getAff: t -> affT

(** Accessor for the [cmp] field of a [t]. *)
	val getCmp: t -> cmpT

(** Accessor for the [cst] field of a [t]. *)
	val getCst: t -> Nb.t

(** Convert a [t] to VPL internal representation of constraints: [Cstr.t]. *)
	val toCstr: t -> Cstr.t

(** Convert an assignment [x = e] (represented by [(x, e)]) to an equivalent [Pol.assign_t]. *)
	val toAssign: Ident.t * affT -> Pol.assign_t
end

(** The type of an abstract value (i.e. a polyhedron). *)
type t
=
	| Bottom
	| NonBot of Pol.t

(** The polyhedron representing the whole space of variable values. *)
val top: t

(** Compute the intersection between a polyhedron and a linear constraint. *)
val meet: t -> Expr.t -> t

(** Compute the convex hull of two polyhedra. *)
val join: t -> t -> t

val assign: t -> (Ident.t * Expr.affT) list -> t

val widen: t -> t -> t

(** Test the inclusion of two polyhedra. *)
val leq: t -> t -> bool

val pr: t -> string
